/******************************************************************************
 *
 * Filename: engine.h
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Execution engine interface and runtime datatypes.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#ifndef _engine_h

#define _engine_h


#include "classfile.h"


typedef enum _ExceptionKind
{
	/* Exception represented by enum value and message */
	VerifyError,			/* Invalid classfile */
	NoClassDefFoundError,	/* Couldn't load classfile */
	IllegalAccessError,		/* Can't access class/method/field */
	ClassFormatError,		/* Invalid classfile */
	NoSuchFieldError,		/* Field not found */
	NoSuchMethodError,		/* Method not found */
	InternalError,			/* VM error */
	UnsatisfiedLinkError,   /* Method not linked */
	ExceptionInInitializerError,	/* An exception occurred in a <clinit> method */

	NullPointerException,	/* null object reference */
	ArrayIndexOutOfBoundsException, /* Array access exception */
	ArithmeticException,    /* divide by zero, etc. */
	ClassCastException,		/* checkcast */
	ClassNotFoundException,	/* Couldn't load classfile */
	IOException,			/* input/output problems */
	NegativeArraySizeException, /* array size is invalid */
	CloneNotSupportedException, /* class does not implement Cloneable */
	ArrayStoreException,	/* Error in System.arrayCopy() */

#if JDK_VERSION < 140
	NoSuchFieldException,	/* Field does not exist */
	NoSuchMethodException,	/* Method does not exist */
#endif /* JDK 1.4 */

#if JDK_VERSION >= 140
	PrivilegedActionException, /* Exception raised by AccessController.doPrivileged() */
#endif /* JDK 1.4 */

	/* Exception represented by ObjRef */
	JavaException			/* All Java exceptions */
} ExceptionKind;


#if JDK_VERSION >= 140
/**
 * A single stack trace entry.  This is used to represent a position in the
 * execution of a method, giving the class, the method, the program counter and
 * where possible the calculated line number (or -1 for no line number).
 */
typedef struct _StackTrace_Entry
{
	struct _ClassFile*	 cf;		/* ClassFile of owning class */
	struct _method_info* mi;		/* method_info of entry */
	int					 pc;		/* Program counter */
	int					 linenum;	/* Best fit line number (calculated) */
} StackTrace_Entry;

/**
 * A complete stack trace.  This is a holder for a set of stack trace entries
 * which together represent the execution position of a particular thread and
 * in particular which methods called which other methods.
 */
typedef struct _StackTrace
{
	int				  entriesc;		/* Number of entries */
	StackTrace_Entry* entriesv;		/* Array of individial entries as an array */
} StackTrace;
#endif /* JDK 1.4 */


/*
 * class java.lang.String
 * {
 *   private char value[];
 *   private int offset;
 *   private int count;
 * }
 */
typedef struct _javaLangString
{
	u4 value;
	u4 offset;
	u4 count;
} javaLangString;


/*
 * class java.lang.Thread
 * {
 *   private char		name[];
 *   private int        priority;
 *   private Thread		threadQ;
 *   private int 		PrivateInfo;
 *   private int		eetop;
 *   private boolean	single_step;
 *   private boolean	daemon;
 *   private boolean	stillborn;
 *   private Runnable	target;
 *   private ThreadGroup	group;
 *   private int		initial_stack_memory;
 * }
 */
typedef struct _javaLangThread
{
	u4 name;
	u4 priority;
	u4 threadQ;
	u4 PrivateInfo;
	u4 eetop;
	u4 single_step;
	u4 daemon;
	u4 stillborn;
	u4 target;
	u4 group;
	u4 initial_stack_memory;
} javaLangThread;


/*
 * class javaIoObjectStreamField
 * {
 *   String name;		// the name of the field
 *   char type;			// type first byte of the type signature
 *   int  offset;		// Offset into the object of the field
 *   String typeString;		// iff object, typename
 * }
 */
typedef struct _javaIoObjectStreamField
{
	u4 name;
	u4 type;
	u4 offset;
	u4 typeString;
} javaIoObjectStreamField;


typedef struct _ObjRef
{
	ClassFile* cf;			/* Object Type */
	u1 type;				/* Built-in Type */
	u4* data;				/* Data block */
	u4 length;				/* Array dimension */
} ObjRef;


typedef struct _ObjRefList
{
	ObjRef* obj;
	struct _ObjRefList* next;
} ObjRefList;


typedef struct _MemoryBlockList
{
	void*  block;
	size_t block_size;
	struct _MemoryBlockList* next;
} MemoryBlockList;

typedef struct _Exception
{
	ExceptionKind kind;
	char* message;
	ObjRef* ex;
} Exception;


typedef struct _Frame
{
	ClassFile* cf;
	method_info* mi;
	u1* pc;
	u1* min_pc;
	u1* max_pc;
	u4* local;
	u4* sp;
	u1* tlocal;
	u1* tsp;
	Exception* ex;
	struct _Frame* next;
	struct _Frame* prev;
	struct _Frame* init;
	int stopOnPop;

	/* data follows */
	u4* data;
} Frame;


typedef struct _Stack
{
	Exception* ex;
	Frame *fp;
	Frame *tail;
	int retcode;
} Stack;


/* ----- pushes ----- */
#define PUSH_ANY_TYPE(fp,v,t)     *((fp)->sp) = (u4) v; ((fp)->sp)++; *((fp)->tsp) = (u1) t; ((fp)->tsp)++;
#define PUSH_DORJ_TYPE(fp,v,n,t)  *((fp)->sp) = v[n];   ((fp)->sp)++; *((fp)->tsp) = (u1) t; ((fp)->tsp)++;
#define PUSH_LOCAL(fp,n)		  *((fp)->sp) = (fp)->local[n]; ((fp)->sp)++; *((fp)->tsp) = (fp)->tlocal[n]; ((fp)->tsp)++;

#define PUSH_OBJREF(fp,v)   PUSH_ANY_TYPE(fp,v,T_OBJREF);
#define PUSH_RETADDR(fp,v)  PUSH_ANY_TYPE(fp,v,T_RETADDR);
#define PUSH_INT(fp,v)      PUSH_ANY_TYPE(fp,v,T_INT);
#define PUSH_LONG(fp,v)     PUSH_DORJ_TYPE(fp,v,1,T_LONG1); PUSH_DORJ_TYPE(fp,v,0,T_LONG);
#define PUSH_FLOAT(fp,v)    PUSH_ANY_TYPE(fp,v,T_FLOAT);

/* optimization fix for GCC -O2/3 */
#ifdef GCC_OPTIMIZER_CLUDGE
#define PUSH_DOUBLE(fp,v)   printf( "v[1] = %08x\n", v[1] ); *((fp)->sp) = v[1]; printf( "PUSH_DOUBLE1( [%p], %08x ) => %08x\n", fp->sp, v[1], *(fp->sp) ); ((fp)->sp)++; *((fp)->tsp) = (u1) T_DOUBLE1; ((fp)->tsp)++; \
							printf( "v[0] = %08x\n", v[0] ); *((fp)->sp) = v[0]; printf( "PUSH_DOUBLE0( [%p], %08x ) => %08x\n", fp->sp, v[0], *(fp->sp) ); ((fp)->sp)++; *((fp)->tsp) = (u1) T_DOUBLE; ((fp)->tsp)++;
#else
#define PUSH_DOUBLE(fp,v)   PUSH_DORJ_TYPE(fp,v,1,T_DOUBLE1); PUSH_DORJ_TYPE(fp,v,0,T_DOUBLE);
#endif /* GCC_OPTIMIZER_CLUDGE */

#define PUSH_LOBJREF(fp,n)  pushCheckLocal( fp, n, T_OBJREF );
#define PUSH_LINT(fp,n)     pushCheckLocal( fp, n, T_INT );
#define PUSH_LLONG(fp,n)    pushCheckLocal2( fp, n, T_LONG, T_LONG1 );
#define PUSH_LFLOAT(fp,n)   pushCheckLocal( fp, n, T_FLOAT );
#define PUSH_LDOUBLE(fp,n)  pushCheckLocal2( fp, n, T_DOUBLE, T_DOUBLE1 );

/* ----- pops ----- */
#define POP_OBJREF(fp,v)   v = (ObjRef*) popCheck( fp, T_OBJREF );
#define POP_INT(fp,v)      v = (int32) popCheck( fp, T_INT );
#define POP_LONG(fp,v)     ZERO_LONG(v); v[0] = popCheck( fp, T_LONG ); v[1] = popCheck1( fp, T_LONG1 );
#define POP_FLOAT(fp,v)    v = popCheck( fp, T_FLOAT );

/* optimization fix for GCC -O2/3 */
#ifdef GCC_OPTIMIZER_CLUDGE
#define POP_DOUBLE(fp,v)   ZERO_DOUBLE(v); v[0] = popCheck( fp, T_DOUBLE ); printf( "v[0] = %08x\n", v[0] ); \
                           v[1] = popCheck1( fp, T_DOUBLE1 ); printf( "v[1] = %08x\n", v[1] );
#else
#define POP_DOUBLE(fp,v)   ZERO_DOUBLE(v); v[0] = popCheck( fp, T_DOUBLE ); \
                           v[1] = popCheck1( fp, T_DOUBLE1 );
#endif /* GCC_OPTIMIZER_CLUDGE */

#define POP_ANY(fp,v)      v = popCheckAny( fp );
#define POP_ANY_TYPE(fp,v,t)   ((fp)->sp)--; v = *((fp)->sp); ((fp)->tsp)--; t = *((fp)->tsp);
#define POP_LOBJREF(fp,n)  popCheckLocal( fp, n, T_OBJREF );
#define POP_LINT(fp,n)     popCheckLocal( fp, n, T_INT );
#define POP_LLONG(fp,n)    popCheckLocal2( fp, n, T_LONG, T_LONG1 );
#define POP_LFLOAT(fp,n)   popCheckLocal( fp, n, T_FLOAT );
#define POP_LDOUBLE(fp,n)  popCheckLocal2( fp, n, T_DOUBLE, T_DOUBLE1 );
#define POP_LOCAL(fp,n)    ((fp)->sp)--; (fp)->local[n] = *((fp)->sp); ((fp)->tsp)--; (fp)->tlocal[n] = *((fp)->tsp);

/* ----- local variable manipulation ----- */
#define PUT_LOBJREF(fp,n,v)   (fp)->local[n] = (u4) v; (fp)->tlocal[n] = (u1) T_OBJREF;
#define GET_LRETADDR(fp,n,v)  v = getLocalRetAddr( fp, n );
#define PUT_LINT(fp,n,v)   (fp)->local[n] = (u4) v; (fp)->tlocal[n] = (u1) T_INT;
#define GET_LINT(fp,n,v)   v = getLocalInt( fp, n );
#define PUT_LANY_TYPE(fp,n,v,t)   (fp)->local[n] = (u4) v; (fp)->tlocal[n] = (u1) t;

/* ----- misc ----- */
#define DUP(fp)	           *((fp)->sp) = *((fp)->sp-1); ((fp)->sp)++; *((fp)->tsp) = *((fp)->tsp-1); ((fp)->tsp)++;

/* optimization fix for GCC -O2/3 */
#ifdef GCC_OPTIMIZER_CLUDGE
#define ZERO_LONG(v)       v[1] = v[0] = 0;
#define ZERO_DOUBLE(v)     v[1] = v[0] = 0;
#else
#define ZERO_LONG(v)       ;
#define ZERO_DOUBLE(v)     ;
#endif  /* GCC_OPTIMIZER_CLUDGE */

/* ----- debug ----- */
#ifdef _DEBUG
#define DEBUG(x) x
#define ARITHMETIC_DEBUG(x)    if( arithmetic_debug ) { x }
#define ARITHMETIC_DEBUGN(n,x) if( arithmetic_debug & n ) { x }
#else
#define DEBUG(x) ;
#define ARITHMETIC_DEBUG(x)    ;
#define ARITHMETIC_DEBUGN(n,x) ;
#endif

#ifdef __cplusplus
extern "C" {
#endif


extern int subclassOf( const ClassFile* cf, const ClassFile* compare );

extern int newCompleteException( Stack *stp, int kind, const char* mess );
extern int newCompleteExceptionWithThrowable( Stack* stp, int kind, ObjRef* throwable );

extern int forceExecuteJavaMethod( ObjRef* objref, Stack* stp, const char* name, const char* desc );
extern int forceExecuteJavaMethodWithArgs( ObjRef* objref, Stack* stp, const char* name, const char* desc, int argc, u4 argv[], u1 argvt[] );
extern int forceExecuteVirtualJavaMethod( ObjRef* objref, Stack* stp, const char* name, const char* desc, int stopOnPop );
extern int forceExecuteStaticJavaMethod( ClassFile* cf, Stack* stp, const char* name, const char* desc );
extern int executeCLInit( ClassFile* cf, Stack* stp );
extern int execute( ClassFile* cf, Stack* stp, int argc, char** argv );
extern int executeInstruction( Stack* stp );

extern inline u4 popCheck( Frame* fp, u1 type );
extern inline u4 popCheck1( Frame* fp, u1 type );
extern inline void pushCheckLocal( Frame* fp, u2 index, u1 type );
extern inline void pushCheckLocal2( Frame* fp, u2 index, u1 type0, u1 type1 );
extern inline void popCheckLocal( Frame* fp, u2 index, u1 type );
extern inline void popCheckLocal2( Frame* fp, u2 index, u1 type0, u1 type1 );
extern inline ObjRef* getLocalObjRef( Frame* fp, u2 index );
extern inline int32 getLocalInt( Frame* fp, u2 index );

#ifdef __cplusplus
}
#endif


#endif
