/******************************************************************************
 *
 * Filename: hashtable.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  16th April 2003
 *
 * Purpose:  Java-style hashtable for convenience.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/


#include <stdio.h>
#include <string.h>

#include "config.h"
#include "jstring.h"
#include "memory.h"
#include "globals.h"

#include "hashtable.h"



/**
 * Create a new instance of a hashtable with the given number of buckets.
 */
Hashtable* Hashtable_new( u4 buckets )
{
	u4 n;

	Hashtable* ret = (Hashtable*) javavm_malloc( sizeof(Hashtable) );
	ret->bucketc = buckets;
	ret->bucketv = (Hashtable_Entry**) javavm_malloc( buckets * sizeof(Hashtable_Entry*) );
	ret->count   = 0;

	for( n = 0; n < buckets; n++ )
		ret->bucketv[n] = NULL;

	return ret;
}


/**
 * Destroy the given hashtable.  Note that this does not free any keys or
 * values, as it is assumed the caller owns the storage.
 */
void Hashtable_delete( Hashtable* table, size_t valuesize, int keys, int values, DestructorFnPtr destructor )
{
	Hashtable_removeAll( table, valuesize, keys, values, destructor );

	javavm_free( table->bucketv, table->bucketc * sizeof(Hashtable_Entry*) );
	table->bucketc = table->count = 0;
	javavm_free( table, sizeof(Hashtable) );
}


/**
 * Removes the contents of the given hashtable.  If 'keys' is set, then it
 * frees the keys.  If 'values' is sent, then it frees the values.  If
 * 'destructor' is passed then this is called for each value before it is
 * freed. The hashtable itself is not freed, though the entries hanging from
 * the buckets are.  The hashtable is then empty and ready to be used again.
 */
void Hashtable_removeAll( Hashtable* table, size_t valuesize, int keys, int values, DestructorFnPtr destructor )
{
	Hashtable_Entry *entry, *temp;
	u4 n;

	/* Loop through all the buckets */
	for( n = 0; n < table->bucketc; n++ )
	{
		/* Follow the linked list for each bucket */
		for( entry = table->bucketv[n]; entry; entry = temp )
		{
			/* If asked, free the key */
			if( keys )
				javavm_free( (void*) entry->key, strlen(entry->key) );

			entry->key = NULL;

			/* If given a destructor fn, call it on the value */
			if( destructor )
				(*destructor)( entry->value );

			/* If asked, free the value */
			if( values )
				javavm_free( entry->value, valuesize );

			entry->value = NULL;

			/* Remember the next entry */
			temp = entry->next;
			entry->next = NULL;

			/* Free the current entry */
			javavm_free( entry, sizeof(Hashtable_Entry) );
		}

		table->bucketv[n] = NULL;
	}

	table->count = 0;
}


/**
 * Create a new instance of a hashtable entry.
 */
Hashtable_Entry* newHashtableEntry( Hashtable_Entry* next, const char* key, void* value )
{
	Hashtable_Entry* ret = (Hashtable_Entry*) javavm_malloc( sizeof(Hashtable_Entry) );
	ret->next  = next;
	ret->key   = key;
	ret->value = value;
	return ret;
}


/**
 * Increase (or decrease) the number of buckets used by the hashtable in order
 * to reduce (or increase) the length of the linked lists used for each bucket.
 * This is normally used when the loading (ratio of entries to buckets) of the
 * Hashtable gets too high and in order to preserve performance, the table is
 * restructured accordingly.  As re-hashing is quite a slow process, it should
 * not be performed too often.
 */
void Hashtable_rehash( Hashtable* table )
{
	Hashtable_Entry **newbv, *entry, *temp;
	u4 n, newbucket;
	u4 buckets = table->bucketc * 2 + 1;

	if( verbose & 2 )
		printf( "Rehashing from %d buckets to %d buckets.\n", table->bucketc, buckets );

	/* Create a new bucket array with the new number of buckets */
	newbv = (Hashtable_Entry**) javavm_malloc( buckets * sizeof(Hashtable_Entry*) );

	for( n = 0; n < buckets; n++ )
		newbv[n] = NULL;

	/* For each entry, calculate its new bucket number and insert it using head
	   insertion.  As there are no duplicate keys we don't need to do any
	   strcmp's. */
	for( n = 0; n < table->bucketc; n++ )
	{
		for( entry = table->bucketv[n]; entry; entry = temp )
		{
			newbucket = hash( entry->key ) % buckets;

			temp = entry->next;
			entry->next = newbv[newbucket];
			newbv[newbucket] = entry;
		}
	}

	/* Free the old bucket array */
	javavm_free( table->bucketv, table->bucketc * sizeof(Hashtable_Entry*) );

	/* Put the new bucket array into the table and set the count */
	table->bucketc = buckets;
	table->bucketv = newbv;
}


/**
 * Insert a key/value pair into the given hashtable.  If the key exists, then
 * the existing value is replaced and the previous value returned, otherwise
 * a new entry is created with the given key and value and NULL is returned.
 */
void* Hashtable_put( Hashtable* table, const char* key, void* value )
{
	Hashtable_Entry* entry;
	u4 bucket;

	if( !key || !value ) {
		printf( "ERROR: Attempt to put NULL key or value to internal hashtable!\n" );
		return NULL;
	}

	/* Find the appropriate bucket */
	bucket = hash( key ) % table->bucketc;

	/* Does the hashtable contain this key? */
	for( entry = table->bucketv[bucket]; entry; entry = entry->next )
	{
		if( strcmp( key, entry->key ) == 0 )
			break;
	}

	/* Did we find the key? */
	if( entry )
	{
		/* Replace the value with our new one */
		Hashtable_Entry* temp = entry->value;
		entry->value = value;
		return temp;
	}
	else
	{
		/* Create a new entry if needed and perform head insertion */
		entry = newHashtableEntry( table->bucketv[bucket], key, value );
		table->bucketv[bucket] = entry;
		table->count++;

		/* If this insert gives l.f. greater than 1 do a re-hash */
		if( table->count > table->bucketc )
			Hashtable_rehash( table );
	}

	return NULL;
}


/**
 * Retrieve the value corresponding to the given key from the specified
 * hashtable.  Returns the value or NULL if the table does not contain the key.
 */
void* Hashtable_get( Hashtable* table, const char* key )
{
	Hashtable_Entry* entry;
	u4 bucket;

	if( !key ) {
		printf( "ERROR: Attempt to get NULL key from internal hashtable!\n" );
		return NULL;
	}

	/* Find the appropriate bucket */
	bucket = hash( key ) % table->bucketc;

	/* Does the hashtable contain this key? */
	for( entry = table->bucketv[bucket]; entry; entry = entry->next )
	{
		if( strcmp( key, entry->key ) == 0 )
			break;
	}

	if( entry )
		return entry->value;

	return NULL;
}


/**
 * Returns the key if the hashtable contains the given value, null otherwise.
 * Searches the entire set of values to see if 'value' is in the hashtable.
 * This is a sequential search and will, on average, search at least half of
 * the hashtable before finding the value.  Use with caution.
 */
const char* Hashtable_findValue( Hashtable* table, void* value )
{
	Hashtable_Entry* entry;
	u4 n;

	for( n = 0; n < table->bucketc; n++ )
	{
		for( entry = table->bucketv[n]; entry; entry = entry->next )
		{
			if( value == entry->value )
				return entry->key;
		}
	}

	return NULL;
}


#ifdef _DEBUG
/**
 * Test method to dump the contents of the hashtable for debugging purposes.
 */
void Hashtable_dump( Hashtable* table )
{
	Hashtable_Entry* entry;
	u4 n;

	printf( "Table contains %d entries\n", table->count );

	for( n = 0; n < table->bucketc; n++ )
	{
		printf( "[%3d]", n );

		for( entry = table->bucketv[n]; entry; entry = entry->next )
		{
			printf( " (\"%s\",%p)", entry->key, entry->value );
		}

		printf( "\n" );
	}
}
#endif /* _DEBUG */


#if 0
int Hashtable_test()
{
	Hashtable* table = Hashtable_new( 4 );

	Hashtable_put( table, "one",   "ONE" );
	Hashtable_put( table, "two",   "TWO" );
	Hashtable_put( table, "three", "THREE" );
	Hashtable_put( table, "four",  "FOUR" );
	Hashtable_put( table, "five",  "FIVE" );
	Hashtable_put( table, "six",   "SIX" );
	Hashtable_put( table, "seven", "SEVEN" );
	Hashtable_put( table, "eight", "EIGHT" );
	Hashtable_put( table, "nine",  "NINE" );
	Hashtable_put( table, "ten",   "TEN" );

	printf( "[\"one\"]   = \"%s\"\n", (const char*) Hashtable_get( table, "one" ) );
	printf( "[\"three\"] = \"%s\"\n", (const char*) Hashtable_get( table, "three" ) );
	printf( "[\"five\"]  = \"%s\"\n", (const char*) Hashtable_get( table, "five" ) );

	Hashtable_put( table, "one", "1" );

	printf( "[\"one\"]   = \"%s\"\n", (const char*) Hashtable_get( table, "one" ) );

	Hashtable_dump( table );

	Hashtable_delete( table );

	return 0;
}
#endif /* _DEBUG */
