/******************************************************************************
 *
 * Filename: linker.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Static class linker.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "config.h"
#include "classfile.h"
#include "loader.h"
#include "zipload.h"
#include "verifier.h"
#include "disasm.h"
#include "native.h"
#include "memory.h"
#include "util.h"
#include "classpath.h"
#include "globals.h"

#include "linker.h"


extern void systemOutPrint( const char* buf, int len );

/*#define LOGFILE*/

#ifdef LOGFILE
FILE* log = NULL;
#endif


int compareUTF8( ClassFile* cf1, u2 index1, ClassFile* cf2, u2 index2 )
{
	if( (index1 > 0 ) && (index1 < cf1->constant_pool_count)
		&& (index2 > 0 ) && (index2 < cf2->constant_pool_count)
		&& cf1->constant_pool[index1] && (cf1->constant_pool[index1]->tag == CONSTANT_Utf8)
		&& cf2->constant_pool[index2] && (cf2->constant_pool[index2]->tag == CONSTANT_Utf8) )
	{
		CONSTANT_Utf8_info* utfp1 = (CONSTANT_Utf8_info*) cf1->constant_pool[index1];
		CONSTANT_Utf8_info* utfp2 = (CONSTANT_Utf8_info*) cf2->constant_pool[index2];

		/*printf( "Comparing \"" ); writeUTF8( utfp1->bytes, utfp1->length );
		printf( "\" to \"" ); writeUTF8( utfp2->bytes, utfp2->length ); printf( "\"\n" );*/ 

		if( utfp1->utf8_length != utfp2->utf8_length )
			return 0;

		/* ### Utf8 strings do not contain '\0' so we should be able to strcmp() them */
		return (strncmp( (const char*) utfp1->utf8_bytes, (const char*) utfp2->utf8_bytes, utfp1->utf8_length ) == 0);
	}

#ifdef _DEBUG
	/* Report why we failed */
	if( (index1 <= 0) || (index1 >= cf1->constant_pool_count) )
		printf( "compareUTF8: index1 out of range (%d outside 0..%d)\n", index1, cf1->constant_pool_count );

	if( (index2 <= 0) || (index2 >= cf2->constant_pool_count) )
		printf( "compareUTF8: index2 out of range (%d outside 0..%d)\n", index1, cf1->constant_pool_count );

	if( !cf1->constant_pool[index1] )
		printf( "compareUTF8: constant_pool entry 1 is null\n" );

	if( !cf2->constant_pool[index2] )
		printf( "compareUTF8: constant_pool entry 2 is null\n" );

	if( cf1->constant_pool[index1] && (cf1->constant_pool[index1]->tag != CONSTANT_Utf8) )
		printf( "compareUTF8: constant_pool entry 1 is not a Utf8\n" );

	if( cf2->constant_pool[index2] && (cf2->constant_pool[index2]->tag != CONSTANT_Utf8) )
		printf( "compareUTF8: constant_pool entry 2 is not a Utf8\n" );
#endif /* _DEBUG */

	return 0;
}


u1 countArguments( ClassFile* cf, u2 index )
{
	u1 args = 0;

	if( CHECKCONSTANT( cf, index, CONSTANT_Utf8) )
	{
		CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[index];
		char desc[1024];
		u2 desclen, n, array = 0;

		desclen = sprintUtf8( desc, utfp );

		if( !desclen || (desc[0] != '(') )
			return 0;

		for( n = 1; (n < desclen) && (desc[n] != ')'); n++ )
		{
			switch( desc[n] )
			{
			case 'B': args++; array = 0; break;
			case 'C': args++; array = 0; break;
			case 'D': if( array ) { args++; } else { args+=2; }; array = 0; break;
			case 'F': args++; array = 0; break;
			case 'I': args++; array = 0; break;
			case 'J': if( array ) { args++; } else { args+=2; }; array = 0; break;
			case 'L':
				args++; array = 0;
				for( n++; (n < desclen) && (desc[n] != ';'); n++ );
				break;
			case 'S': args++; array = 0; break;
			case 'Z': args++; array = 0; break;
			case '[': array++; break;
			}
		}
	}

	return args;
}


u1 countField( ClassFile* cf, u2 index )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Utf8) )
	{
		CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[index];

		/* ### No need to decode Utf8 as same as ASCII for these chars */
		switch( utfp->utf8_bytes[0] )
		{
		case 'B': return 1;
		case 'C': return 1;
		case 'D': return 2;
		case 'F': return 1;
		case 'I': return 1;
		case 'J': return 2;
		case 'L': return 1;
		case 'S': return 1;
		case 'Z': return 1;
		case '[': return 1;
		}
	}

	return 0;
}


/**
 * Hooks up pointers to attribute values for a field to its field_info
 * structure.  These include a ConstantValue attribute, containing an
 * (optional) initial value.
 */
void resolveFieldAttributes( ClassFile* cf, field_info* fi )
{
	u2 m;

	fi->cf = cf;	/* link field_info to ClassFile */

	for( m = 0; m < fi->attributes_count; m++ )
	{
		if( fi->attributes[m]
			&& checkSimpleName( fi->cf, fi->attributes[m]->attribute_name_index, CONSTANTVALUE_NAME ) )
		{
			/* link ConstantValue_attribute to field_info */
			fi->cva = (ConstantValue_attribute*) fi->attributes[m];
		}
	}
}


void resolveMethodAttributes( ClassFile* cf, method_info* mi )
{
	mi->cf = cf;	/* link method_info to ClassFile */
	mi->args = countArguments( cf, mi->descriptor_index );

	/* Process Native methods */
	if( mi->access_flags & ACC_NATIVE )
	{
		mi->nativestr = findNativeMethod( cf, mi->name_index, mi->descriptor_index );

		if( !mi->nativestr && 0 /* verbose ### */ )
		{
			char buf[255];
			sprintMethodInfo( buf, mi );
			printf( "LINKER: No native implementation of %s\n", buf );
		}

		return;
	}

	/* Process Java methods */
	if( mi->attributes_count )
	{
		u2 m;

		for( m = 0; m < mi->attributes_count; m++ )
		{
			if( mi->attributes[m] && checkSimpleName( cf, mi->attributes[m]->attribute_name_index, "Code" ) )
			{
				/* link Code_attribute and code itself */
				mi->ca = (Code_attribute*) mi->attributes[m];
				mi->nativestr = NULL;

				if( mi->ca->attributes_count )
				{
					u2 p;

					for( p = 0; p < mi->ca->attributes_count; p++ )
					{
						if( mi->ca->attributes[p] && checkSimpleName( cf, mi->ca->attributes[p]->attribute_name_index, "LineNumberTable" ) )
						{
							mi->lnta = (LineNumberTable_attribute*) mi->ca->attributes[p];
						}
						else if( mi->ca->attributes[p] && checkSimpleName( cf, mi->ca->attributes[p]->attribute_name_index, "LocalVariableTable" ) )
						{
							mi->lvta = (LocalVariableTable_attribute*) mi->ca->attributes[p];
						}
					}
				}
			}
		}

		/* Check we found some code */
		if( !mi->ca && !(mi->access_flags & ACC_ABSTRACT) )
		{
			char buf[255];
			sprintMethodInfo( buf, mi );
			printf( "LINKER: No Java code for %s\n", buf );
		}
	}
}


void linkMethodRef( ClassFile* cf, u2 index, Stack* stp )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Methodref ) )
	{
		CONSTANT_Methodref_info* mp = (CONSTANT_Methodref_info*) cf->constant_pool[index];
		CONSTANT_Class_info* cp;
		CONSTANT_NameAndType_info* natp;
		method_info* mip;
		u2 m;

		if( !(CHECKCONSTANT( cf, mp->class_index, CONSTANT_Class)) )
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			printf( "LINKER: No Class constant for method %s\n", buf );
			return;
		}

		cp = (CONSTANT_Class_info*) cf->constant_pool[mp->class_index];

		if( !(CHECKCONSTANT( cf, mp->name_and_type_index, CONSTANT_NameAndType)) )
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			printf( "LINKER: No NameAndType constant for method %s\n", buf );
			return;
		}

		natp = (CONSTANT_NameAndType_info*) cf->constant_pool[mp->name_and_type_index];
		
		if( !cp->cf )
		{
			linkClass( cf, mp->class_index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) || !cp->cf )
			{
				char buf[255];
				sprintClassName( buf, cf, mp->class_index );
				printf( "LINKER: Class \"%s\" cannot be loaded\n", buf );
				return;
			}
		}
		
		for( m = 0; (m < cp->cf->methods_count)
			&& !( compareUTF8( cp->cf, cp->cf->methods[m]->name_index, cf, natp->name_index )
				&& compareUTF8( cp->cf, cp->cf->methods[m]->descriptor_index, cf, natp->descriptor_index ) );
			m++ );

		if( m >= cp->cf->methods_count )
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			printf( "LINKER: Failed to find method %s\n", buf );
			return;
		}

		mip = cp->cf->methods[m];
		
#ifdef LOGFILE
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			fprintf( log, "Linker: Resolved method %s\n", buf );
		}
#endif
	
		mp->info = mip;		/* link Methodref to method_info */

		resolveMethodAttributes( cp->cf, cp->cf->methods[m] );
	}
}


void linkVirtualMethodRef( ClassFile* cf, u2 index, Stack* stp )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Methodref ) )
	{
		CONSTANT_Methodref_info* mp = (CONSTANT_Methodref_info*) cf->constant_pool[index];
		CONSTANT_Class_info* cp;
		CONSTANT_NameAndType_info* natp;
		method_info* mip;
		u2 m;

		if( !(CHECKCONSTANT( cf, mp->class_index, CONSTANT_Class)) )
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			printf( "LINKER: No Class constant for method %s\n", buf );
			return;
		}

		cp = (CONSTANT_Class_info*) cf->constant_pool[mp->class_index];

		if( !(CHECKCONSTANT( cf, mp->name_and_type_index, CONSTANT_NameAndType)) )
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			printf( "LINKER: No NameAndType constant for method %s\n", buf );
			return;
		}

		natp = (CONSTANT_NameAndType_info*) cf->constant_pool[mp->name_and_type_index];
		
		if( !cp->cf )
		{
			linkClass( cf, mp->class_index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) || !cp->cf )
			{
				char buf[255];
				sprintClassName( buf, cf, mp->class_index );
				printf( "LINKER: Class \"%s\" cannot be loaded\n", buf );
				return;
			}
		}
		
		for( m = 0; (m < cp->cf->virtuals_count)
			&& !( compareUTF8( cp->cf->virtuals[m]->parent, cp->cf->virtuals[m]->name_index, cf, natp->name_index )
				&& compareUTF8( cp->cf->virtuals[m]->parent, cp->cf->virtuals[m]->descriptor_index, cf, natp->descriptor_index ) );
			m++ );

		if( m >= cp->cf->virtuals_count )
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			printf( "LINKER: Failed to find virtual method %s\n", buf );
			return;
		}

		mip = cp->cf->virtuals[m];
		
#ifdef LOGFILE
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			fprintf( log, "Linker: Resolved virtual method %s\n", buf );
		}
#endif
	
		mp->info = mip;		/* link Methodref to method_info */

		resolveMethodAttributes( mip->parent, mip );
	}
}


void linkInterfaceMethodRef( ClassFile* cf, u2 index, Stack* stp )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_InterfaceMethodref ) )
	{
		CONSTANT_InterfaceMethodref_info* mp = (CONSTANT_InterfaceMethodref_info*) cf->constant_pool[index];
		CONSTANT_Class_info* cp;
		CONSTANT_NameAndType_info* natp;
		method_info* mip;
		u2 m;

		if( !(CHECKCONSTANT( cf, mp->class_index, CONSTANT_Class)) )
		{
			char buf[255];
			sprintInterfaceMethodRef( buf, cf, index );
			printf( "LINKER: No Class constant for method %s\n", buf );
			return;
		}

		cp = (CONSTANT_Class_info*) cf->constant_pool[mp->class_index];

		if( !(CHECKCONSTANT( cf, mp->name_and_type_index, CONSTANT_NameAndType)) )
		{
			char buf[255];
			sprintInterfaceMethodRef( buf, cf, index );
			printf( "LINKER: No NameAndType constant for method %s\n", buf );
			return;
		}

		natp = (CONSTANT_NameAndType_info*) cf->constant_pool[mp->name_and_type_index];
		
		if( !cp->cf )
		{
			linkClass( cf, mp->class_index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) || !cp->cf )
			{
				char buf[255];
				sprintClassName( buf, cf, mp->class_index );
				printf( "LINKER: Class \"%s\" cannot be loaded\n", buf );
				return;
			}
		}
		
		for( m = 0; (m < cp->cf->methods_count)
			&& !( compareUTF8( cp->cf, cp->cf->methods[m]->name_index, cf, natp->name_index )
				&& compareUTF8( cp->cf, cp->cf->methods[m]->descriptor_index, cf, natp->descriptor_index ) );
			m++ );

		if( m >= cp->cf->methods_count )
		{
			char buf[255];
			sprintInterfaceMethodRef( buf, cf, index );
			printf( "LINKER: Failed to find method %s\n", buf );
			return;
		}

		mip = cp->cf->methods[m];
		
#ifdef LOGFILE
		{
			char buf[255];
			sprintInterfaceMethodRef( buf, cf, index );
			fprintf( log, "Linker: Resolved method %s\n", buf );
		}
#endif
	
		mp->info = mip;		/* link Methodref to method_info */

		resolveMethodAttributes( cp->cf, cp->cf->methods[m] );
	}
}


void linkFieldRef( ClassFile* cf, u2 index, Stack* stp )
{
	/* Make sure that the constant pool entry at 'index' is a field reference */
	if( CHECKCONSTANT( cf, index, CONSTANT_Fieldref ) )
	{
		/* Cast the field entry so we can use it */
		CONSTANT_Fieldref_info* fp = (CONSTANT_Fieldref_info*) cf->constant_pool[index];

		CONSTANT_Class_info* cp;
		CONSTANT_NameAndType_info* natp;
		ClassFile* thisClass;
		char buf[255];
		u2 m;

		/* Make sure that the entry for the field's class is a class reference */
		if( !(CHECKCONSTANT( cf, fp->class_index, CONSTANT_Class)) )
		{
			sprintFieldRef( buf, cf, index );
			printf( "LINKER: No Class constant for field %s\n", buf );
			return;
		}

		/* Cast the class entry so we can use it */
		cp = (CONSTANT_Class_info*) cf->constant_pool[fp->class_index];

		/* Make sure that the entry for the field's type is a name and type reference */
		if( !(CHECKCONSTANT( cf, fp->name_and_type_index, CONSTANT_NameAndType)) )
		{
			sprintFieldRef( buf, cf, index );
			printf( "LINKER: No NameAndType constant for field %s\n", buf );
			return;
		}

		/* Cast the name and type entry so we can use it */
		natp = (CONSTANT_NameAndType_info*) cf->constant_pool[fp->name_and_type_index];
		
		/* Has the class specified in the class reference been linked? */
		if( !cp->cf )
		{
			/* Link it */
			linkClass( cf, fp->class_index, stp );

			/* Abort if an exception occurred during the loading of the class */
			if( stp->ex || (stp->fp && stp->fp->ex) || !cp->cf )
			{
				sprintClassName( buf, cf, fp->class_index );
				printf( "LINKER: Class \"%s\" cannot be loaded\n", buf );
				return;
			}
		}
		
		/* Work upwards from this class through each of its superclasses */
		for( thisClass = cp->cf; thisClass; thisClass = thisClass->super )
		{
			/* Check each field in turn */
			for( m = 0; m < thisClass->fields_count; m++ )
			{
				/* Do the field names match? */
				if( compareUTF8( thisClass, thisClass->fields[m]->name_index, cf, natp->name_index ) )
				{
					field_info* fip = thisClass->fields[m];

					/* Do the types match?  If not we have an incompatible class change */
					if( !compareUTF8( thisClass, thisClass->fields[m]->descriptor_index, cf, natp->descriptor_index ) )
					{
						char cbuf[256];
						sprintClassName( cbuf, thisClass, thisClass->this_class );
						sprintFieldRef( buf, cf, index );
						printf( "ERROR: FieldRef %s has conflicting type with its declaration in %s.\n", buf, cbuf );
						return;
					}

#ifdef LOGFILE
					sprintFieldRef( buf, cf, index );
					fprintf( log, "Linker: Resolved field %s\n", buf );
#endif
					fp->info = fip;			/* link Fieldref to field_info */
					fip->cf  = thisClass;	/* link field_info to ClassFile */
					return;
				}
			} /* for m */
		} /* for thisClass */

		/* Didn't find it - report an error */
		sprintFieldRef( buf, cf, index );
		printf( "LINKER: Failed to find field %s\n", buf );
	}
}


void linkClass( ClassFile* cf, u2 index, Stack* stp )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Class ) )
	{
		CONSTANT_Class_info* cp = (CONSTANT_Class_info*) cf->constant_pool[index];

		if( CHECKCONSTANT( cf, cp->name_index, CONSTANT_Utf8 ) )
		{
			CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[cp->name_index];
			char buf[255];
		
			sprintUtf8( buf, utfp );

			cp->cf = load( buf, stp );
		}
	}
}


/* resolve this class and all its super-classes */
void resolveSuperClasses( ClassFile* cf )
{
	size_t currsize, oldsize;
	u2 n, m;

#ifdef LOGFILE
	char classname[255];
	sprintClassName( classname, cf, cf->this_class );
	fprintf( log, "Resolving class \"%s\"...\n", classname );
#endif

	cf->resolved = 1;

	/* Make each of the field_info's refer to their parent ClassFile */
	for( n = 0; n < cf->fields_count; n++ )
	{
		if( cf->fields[n] && !cf->fields[n]->cf )
			resolveFieldAttributes( cf, cf->fields[n] );
	}

	/* Call resolveMethodAttributes() for each method_info */
	for( n = 0; n < cf->methods_count; n++ )
	{
		if( cf->methods[n] && !cf->methods[n]->cf )
			resolveMethodAttributes( cf, cf->methods[n] );
	}

	/* Resolve source file attribute (if any) */
	for( n = 0; n < cf->attributes_count; n++ )
	{
		if( cf->attributes[n] && checkSimpleName( cf, cf->attributes[n]->attribute_name_index, "SourceFile" ) )
		{
			cf->sfa = (SourceFile_attribute*) cf->attributes[n];
			break;
		}
	}

	/* Force base clase to be resolved first */
	if( cf->super )
	{
		if( !cf->super->resolved )
			resolveSuperClasses( cf->super );

		cf->object_size = cf->super->object_size;
	}
	else
		cf->object_size = 0;

	/* Calculate sizes of static and object data */
	for( n = 0; n < cf->fields_count; n++ )
	{
		if( cf->fields[n] )
		{
			if( cf->fields[n]->access_flags & ACC_STATIC )
			{
				cf->fields[n]->offset = cf->static_size;
				cf->static_size += countField( cf, cf->fields[n]->descriptor_index );
			}
			else
			{
				cf->fields[n]->offset = cf->object_size;
				cf->object_size += countField( cf, cf->fields[n]->descriptor_index );
			}
		}
	}

#ifdef LOGFILE
	fprintf( log, "Class \"%s\", object size = %d, static size = %d\n", classname, cf->object_size, cf->static_size );
#endif

	/* do virtual method resolution */
	if( !cf->super && !cf->interfaces_count )
	{
#ifdef LOGFILE
		fprintf( log, "Class \"%s\" has no super-classes.\n", classname );

		for( m = 0; m < cf->methods_count; m++ )
		{
			char mbuf[255];
			sprintMethodInfo( mbuf, cf->methods[m] );
			fprintf( log, "(%d) Added method \"%s\"\n", m, mbuf );
		}
#endif

		cf->virtuals = cf->methods;
		cf->virtuals_count = cf->methods_count;
		return;
	}

	/* Allocate enough space for this just class' methods */
	oldsize = currsize =  cf->methods_count * sizeof( method_info* );
	cf->virtuals = javavm_malloc( currsize );
	cf->virtuals_count = 0;

	if( cf->super )
	{
#ifdef LOGFILE
		char sbuf[255];
		sprintClassName( sbuf, cf->super, cf->super->this_class );
		fprintf( log, "Class \"%s\" has super-class \"%s\".\n", classname, sbuf );
#endif

		if( cf->super->virtuals )
		{
			/* Make room for additional methods from superclass */
			currsize = (cf->virtuals_count + cf->super->virtuals_count) * sizeof( method_info* );
			cf->virtuals = javavm_realloc( cf->virtuals, currsize, oldsize );
			oldsize = currsize;

			/* Copy superclass' virtual method table */
			memcpy( &cf->virtuals[cf->virtuals_count], cf->super->virtuals, cf->super->virtuals_count * sizeof( method_info* ) );
#ifdef LOGFILE
			for( m = cf->virtuals_count; m < cf->virtuals_count + cf->super->virtuals_count; m++ )
			{
				char mbuf[255];
				sprintMethodInfo( mbuf, cf->virtuals[m] );
				fprintf( log, "(%d) Inherited method \"%s\"\n", m, mbuf );
			}
#endif
			/* Update the count */
			cf->virtuals_count += cf->super->virtuals_count;
		}
		else
		{
			char cbuf[255], sbuf[255];
			sprintClassName( cbuf, cf, cf->this_class );
			sprintClassName( sbuf, cf, cf->super_class );
			printf( "LINKER: super-class \"%s\" of \"%s\" has not been resolved properly\n", sbuf, cbuf );
		}
	}

	if( cf->interfaces_count )
	{
		for( n = 0; n < cf->interfaces_count; n++ )
		{
			if( cf->interfaces[n] )
			{
#ifdef LOGFILE
				char sbuf[255];
				sprintClassName( sbuf, cf, cf->interfaces[n]->class_index );
				fprintf( log, "Class \"%s\" has super-interface \"%s\".\n", classname, sbuf );
#endif

				if( cf->interfaces[n]->cf && cf->interfaces[n]->cf->virtuals )
				{
					/* copy super-interface's virtual method table */
					currsize = (cf->virtuals_count + cf->interfaces[n]->cf->virtuals_count) * sizeof( method_info* );
					cf->virtuals = javavm_realloc( cf->virtuals, currsize, oldsize );
					oldsize = currsize;

					if( cf->virtuals_count > 0 )
					{
						u2 p;
						for( p = 0; p < cf->interfaces[n]->cf->virtuals_count; p++ )
						{
							for( m = 0; (m < cf->virtuals_count) && (cf->virtuals[m] != cf->interfaces[n]->cf->virtuals[p]); m++ );

							if( m >= cf->virtuals_count )
							{
#ifdef LOGFILE
								char mbuf[255];
								sprintMethodInfo( mbuf, cf->interfaces[n]->cf->virtuals[p] );
								fprintf( log, "(%d) Inherited method \"%s\"\n", cf->virtuals_count, mbuf );
#endif
								cf->virtuals[cf->virtuals_count] = cf->interfaces[n]->cf->virtuals[p];
								cf->virtuals_count++;
							}
						}
					}
					else
					{
						memcpy( &cf->virtuals[cf->virtuals_count], cf->interfaces[n]->cf->virtuals, cf->interfaces[n]->cf->virtuals_count * sizeof( method_info* ) );
#ifdef LOGFILE
						for( m = cf->virtuals_count; m < cf->virtuals_count + cf->interfaces[n]->cf->virtuals_count; m++ )
						{
							char mbuf[255];
							sprintMethodInfo( mbuf, cf->virtuals[m] );
							fprintf( log, "(%d) Inherited method \"%s\"\n", m, mbuf );
						}
#endif
						cf->virtuals_count += cf->interfaces[n]->cf->virtuals_count;
					}
				}
				else
				{
					char cbuf[255], ibuf[255];
					sprintClassName( cbuf, cf, cf->this_class );
					sprintClassName( ibuf, cf, cf->interfaces[n]->class_index );
					printf( "LINKER: super-interface \"%s\" of \"%s\" has not been resolved properly\n", ibuf, cbuf );
				}
			}
		}
	}

#ifdef LOGFILE
	fprintf( log, "Class \"%s\" has %d virtuals before overrides.\n", classname, cf->virtuals_count );
#endif

	/* process overrides */
	for( n = 0; n < cf->virtuals_count; n++ )
	{
		ClassFile *vcf = cf->virtuals[n]->cf;
		method_info *vmi = cf->virtuals[n];

		for( m = 0; (m < cf->methods_count)
			&& !( compareUTF8( cf, cf->methods[m]->name_index, vcf, vmi->name_index )
				&& compareUTF8( cf, cf->methods[m]->descriptor_index, vcf, vmi->descriptor_index ) );
			m++ );

		if( m < cf->methods_count )
		{
#ifdef LOGFILE
			char mbuf[255], vbuf[255];
			sprintMethodInfo( mbuf, cf->methods[m] );
			sprintMethodInfo( vbuf, cf->virtuals[n] );
			fprintf( log, "(%d) Resolved virtual method \"%s\" over \"%s\"\n", n, mbuf, vbuf );
#endif
			cf->virtuals[n] = cf->methods[m];
			cf->methods[m]->index = n;
		}
	}

	/* Make room for class' own methods */
	currsize = (cf->virtuals_count + cf->methods_count) * sizeof( method_info* );
	cf->virtuals = javavm_realloc( cf->virtuals, currsize, oldsize );
	oldsize = currsize;

	/* Add this class' methods to end of virtual table */
	for( n = 0; n < cf->methods_count; n++ )
	{
		for( m = 0; (m < cf->virtuals_count) && (cf->virtuals[m] != cf->methods[n]); m++ );

		if( m >= cf->virtuals_count )
		{
#ifdef LOGFILE
			char mbuf[255];
			sprintMethodInfo( mbuf, cf->methods[n] );
			fprintf( log, "(%d) Added method \"%s\"\n", cf->virtuals_count, mbuf );
#endif
			cf->virtuals[cf->virtuals_count] = cf->methods[n];
			cf->methods[n]->index = cf->virtuals_count;
			cf->virtuals_count++;
		}
	}

	/* Shrink virtuals table to minimum necessary */
	currsize = cf->virtuals_count * sizeof( method_info* );
	cf->virtuals = javavm_realloc( cf->virtuals, currsize, oldsize );

#ifdef LOGFILE
	fprintf( log, "Class \"%s\" has %d virtuals after overrides.\n", classname, cf->virtuals_count );
#endif
}


ClassFile* loadClass( const char* classname, ClassFile* cf, u2 index, Stack* stp )
{
	char buf[255];

	if( CHECKCONSTANT( cf, index, CONSTANT_Class) )
	{
		CONSTANT_Class_info* cip = (CONSTANT_Class_info*) cf->constant_pool[index];

		if( CHECKCONSTANT( cf, cip->name_index, CONSTANT_Utf8 ) )
		{
			CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[cip->name_index];

			sprintUtf8( buf, utfp );

			if( buf[0] == '[' )
			{
				stp->retcode = newCompleteException( stp, NoClassDefFoundError, buf );
				return NULL;
			}

			if( !(cip->cf = load( buf, stp )) )
			{
				stp->retcode = newCompleteException( stp, NoClassDefFoundError, buf );
				return NULL;
			}

			return cip->cf;
		}
	}

	sprintf( buf, "Super-interface constant pool entry corrupt in class \"%s\"", classname );
	stp->retcode = newCompleteException( stp, InternalError, buf );
	return NULL;
}


/**
 * The complete load, link, verify and initialise process for bringing a class
 * into the VM.  The classpath is encapsulated here, delegating to the loader
 * to actually check for and load the class file.  Linking is then performed to
 * link just the superclass and super-interfaces of this class (it also hooks
 * up some internal pointers), but methods and fields are not resolved until
 * they are used at runtime.  The loaded and linked class is then verified with
 * the pass 2 verifier (pass 1 is part of the loader).  Finally the class is
 * initialised (see initialise() for more info).
 */
ClassFile* load( const char* classname, Stack* stp )
{
	int retcode = 0, found = 0, n, cpindex = -1;
	ClassFile* cf;
	char hostfname[255], zipfname[255];
	char *hbuf = hostfname, *zbuf = zipfname;

    cf = checkLoaded( classname );
    
    if( cf )
        return cf;

	/** Create an host dependant filename: Add ".class" and replace '/'s and '$'s */

	/* Copy class name */
	strcpy( hbuf, classname ); hbuf += strlen( classname );
#if USES_FILE_EXTENSIONS
	/* Append ".class" onto filename to give host filename */
	strcpy( hbuf, CLASS_FILE_EXTN ); hbuf += strlen( CLASS_FILE_EXTN );
#endif /* USES_FILE_EXTENSIONS */
	*hbuf = '\0';
	
#ifdef USES_BACKSLASHED_FILENAMES
	/* Replace '/'s with '\'s */
    for( hbuf = hostfname; *hbuf; hbuf++ )
    {
        if( *hbuf == '/' ) *hbuf = '\\';
    }
#endif /* USES_BACKSLASHED_FILENAMES */
	
#ifdef USES_NONDOLLAR_FILENAMES
	/* Replace '$'s with '_'s */
    for( hbuf = hostfname; *hbuf; hbuf++ )
    {
        if( *hbuf == '$' ) *hbuf = '_';
    }
#endif /* USES_NONDOLLAR_FILENAMES */

	/** Create a zip format filename: Add ".class" and leave '/'s alone */

	/* Copy class name */
	strcpy( zbuf, classname ); zbuf += strlen( classname );
#if USES_ZIP_FILE_EXTENSIONS
	/* Append ".class" onto filename to give zip filename */
	strcpy( zbuf, ZIP_CLASS_FILE_EXTN ); zbuf += strlen( ZIP_CLASS_FILE_EXTN );
#endif /* USES_ZIP_FILE_EXTENSIONS */
	*zbuf = '\0';

	/* Implementation of CLASSPATH */
	for( n = 0; !found && (n < classpathc); n++ )
	{
		cf = NULL;

		if( classpathv[n].kind == cpt_Directory )
		{
			/* Prepend classpath entry onto host filename */
			char prehfname[512];
			char* pbuf = prehfname;
			strcpy( pbuf, classpathv[n].path ); pbuf += strlen( classpathv[n].path );
			strcpy( pbuf, FILE_SEPARATOR ); pbuf += strlen( FILE_SEPARATOR );
			strcpy( pbuf, hostfname ); pbuf += strlen( hostfname );
			*pbuf = '\0';

			cpindex = n;

			/* Try normal class loader - using prepended host filename  */
			cf = readClassFile( prehfname, classname, stp, &found );
		}
#ifndef USES_UNZIPPED_CLASSES_ONLY
		else if( classpathv[n].kind == cpt_Archive )
		{
			cpindex = n;

			/* Try ZIP loader - using raw filename */
			cf = readZippedClassFile( &classpathv[n], zipfname, classname, stp, &found );
		}
#endif /* USES_UNZIPPED_CLASSES_ONLY */

		if( !cf && found )
			return NULL; /* load failed - exception will be on stack */
	}

	/* If not yet found then raise an exception */
	if( !found )
	{
		stp->retcode = newCompleteException( stp, NoClassDefFoundError, classname );
		return NULL;
	}
	/* End of CLASSPATH implementation */

	/* Load any interface classfiles */
	if( cf->interfaces_count )
	{
		u2 i;

		for( i = 0; i < cf->interfaces_count; i++ )
		{
			if( cf->interfaces[i] )
			{
				if( !(cf->interfaces[i]->cf = loadClass( classname, cf, cf->interfaces[i]->class_index, stp )) )
					return NULL; /* load failed - exception will be on stack */
			}
		}
	}

	/* Load the superclass classfile */
	if( cf->super_class )
	{
		ClassFile* scp;

		if( !(scp = loadClass( classname, cf, cf->super_class, stp )) )
			return NULL; /* load failed - exception will be on stack */

		cf->super = scp;
	}

#ifdef LOGFILE
	log = fopen( "./javavm.log", "aw" );
#endif

	resolveSuperClasses( cf );

#ifdef LOGFILE
	fclose( log );
#endif

	if( !verifyPass2( cf, stp ) )
		return NULL; /* verify puts exception on stack */

	if( !cf->initialised )
		retcode = initialise( cf, stp );

	/* If an exception was rasied, convert it to an ExceptionInInitializerError */
	if( retcode == 2 )
	{
		Exception *ex = NULL;
		
		if( stp->ex )
			ex = stp->ex;
		else if( stp->fp && stp->fp->ex )
			ex = stp->fp->ex;


		stp->ex = NULL;

		if( ex && ex->ex )
		{
			if( verbose )
				printf( "EXCEPTION: Exception in initialiser: %s\n", "JavaException" );

			stp->retcode = newCompleteExceptionWithThrowable( stp, ExceptionInInitializerError, ex->ex );
		}
		else
		{
			if( verbose )
				printf( "EXCEPTION: Exception in initialiser: %s\n", ex ? exceptionKindToString( ex->kind ) : "No Exception" );

			stp->retcode = newCompleteException( stp, ExceptionInInitializerError, ex ? ex->message : NULL );
		}

		return NULL;
	}

	/* JDK-style [Loaded] message */
	if( verbose )
	{
		char buf[255];

		if( (cpindex >= 0) && (classpathv[cpindex].kind == cpt_Archive) )
			sprintf( buf, "[Loaded %s from %s]\n", classname, classpathv[cpindex].path );
		else
			sprintf( buf, "[Loaded %s]\n", classname );

		systemOutPrint( buf, strlen( buf ) );
	}

	/* ### should really be exception in initialiser */
	return (retcode == 2 ? NULL : cf);
}


/***************************** initialise code *****************************/


/**
 * Initialises a loaded and linked class.  If the class has 1 or more static
 * fields, a static data area is created.  The static fields are then checked
 * for ConstantValue attributes and initialised to that value (if there is
 * one).  <clinit> is then invoked to complete the initialisation.  Note that
 * we mark the fact the the class is initialised before we enter <clinit>.
 * This prevents <clinit> being called again before the first call has finished
 * executing.
 */
int initialise( ClassFile* cf, Stack* stp )
{
	int retcode = 0;

	if( cf->static_size )
	{
		u2 n;

		if( verbose & 2 )
		{
			char cbuf[255], buf[512];
			sprintClassName( cbuf, cf, cf->this_class );
			sprintf( buf, "Allocating %d words of static data to class \"%s\"\n", (int) cf->static_size, cbuf );
			systemOutPrint( buf, strlen( buf ) );
		}

		/* Allocate a block of memory for the class' static data */
		cf->static_data = (u4*) javavm_malloc( 4 * cf->static_size );
		memset( cf->static_data, 0, 4 * cf->static_size );

		/* Check through all the fields */
		for( n = 0; n < cf->fields_count; n++ )
		{
			/* Initialise any static fields with ConstantValue attibutes */
			if( ((cf->fields[n]->access_flags & ACC_STATIC) != 0) && cf->fields[n]->cva )
			{
				field_info* fi = cf->fields[n];
				cp_info* cpe = cf->constant_pool[fi->cva->constantvalue_index];

				/* See what kind of constant the constant value attribute refers to */
				switch( cpe->tag )
				{
				case CONSTANT_String:
					{
					CONSTANT_String_info* sp = (CONSTANT_String_info*) cpe;

					if( !CHECKCONSTANT( cf, sp->string_index, CONSTANT_Utf8 ) )
						return newCompleteException( stp, InternalError, "String constant does not refer to Utf8 value" );

					if( !sp->objref )
					{
						CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[sp->string_index];
						sp->objref = newJavaLangStringFromUtf8( utfp );

						if( !sp->objref )
							return newCompleteException( stp, InternalError, "ldc failed - failed to create string" );

						/* Let the object pool clean up the String object */
						addToList( sp->objref, &objectPool );
					}

					cf->static_data[fi->offset] = (u4) sp->objref;
					}
					break;
				case CONSTANT_Integer:
					cf->static_data[fi->offset] = (u4) ((CONSTANT_Integer_info*) cpe)->bytes;
					break;
				case CONSTANT_Float:
					cf->static_data[fi->offset] = (u4) ((CONSTANT_Float_info*) cpe)->bytes;
					break;
				case CONSTANT_Long:
					cf->static_data[fi->offset+1] = ((CONSTANT_Long_info*) cpe)->high_bytes;
					cf->static_data[fi->offset]   = ((CONSTANT_Long_info*) cpe)->low_bytes;
					break;
				case CONSTANT_Double:
					cf->static_data[fi->offset+1] = ((CONSTANT_Double_info*) cpe)->high_bytes;
					cf->static_data[fi->offset]   = ((CONSTANT_Double_info*) cpe)->low_bytes;
					break;
				default:
					{
					char buf[255];
					sprintf( buf, "Unsupported constant type (%d)", cpe->tag );
					return newCompleteException( stp, InternalError, buf );
					}
				} /* End switch */ 
			} /* End static with constant value */
		} /* End for */
	} /* End has static data */

	cf->initialised = 1;

	retcode = executeCLInit( cf, stp );

	return retcode;
}
