/******************************************************************************
 * 
 * Filename: math-wins.h
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  4th December 1999
 *
 * Purpose:  Missing maths routines in WINS libraries.  These routines are not
 *			 called directly, but are a results of maths operations and casts.
 *
 * (c) Copyright R.H.Thornburrow, 1999-2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#include "config.h"
#include "types.h"

#ifndef PLATFORM_EPOC32_WINS
#error "This file is only intended for WINS (the Epoc32 Emulator)"
#endif /* PLATFORM_EPOC32_WINS */

#ifndef FLT_MAX
#define FLT_MAX         3.402823466e+38F        /* max value */
#endif /* FLT_MAX */

extern int32 _ftol( float a );


/* int64 multiplication */
__int64 _allmul( __int64 a, __int64 b )
{
    __int64 r = (long) a * (long) b;
    
    int signa = (a & 0x8000000000000000) ? -1 : 1;
    int signb = (b & 0x8000000000000000) ? -1 : 1;
    int signr = (r & 0x8000000000000000) ? -1 : 1;
    
    if( (a & 0xffffffff00000000) || (b & 0xffffffff00000000) || (signa * signb != signr) )
    {
        printf( "WARNING: overflow in allmul()\n" );
        return (__int64) 1;
    }

    return r;
}


/* int64 shift left */
__int64 _allshl( __int64 a, __int64 b )
{
    __int64 r;
    long bl = (long) b;
    long *aw = (long*) &a;
    long *rw = (long*) &r;
    
    if( b <= 32 )
    {
        long ol;
        
        rw[0] = aw[0] << bl;
        rw[1] = aw[1] << bl;
        ol = aw[0] >> (32 - bl);
        rw[1] |= ol;
        return r;
    }

    if( b <= 64 )
    {
        rw[0] = 0;
        rw[1] = aw[0] << (bl - 32);
        return r;
    }

    printf( "WARNING: overflow in allshl(0x%08x%08x, 0x%08x )\n", aw[1], aw[0], bl );        
    return (long) a << (long) b;
}


/* int64 shift right */
__int64 _allshr( __int64 a, __int64 b )
{
    __int64 r;
    long bl = (long) b;
    long *aw = (long*) &a;
    long *bw = (long*) &b;
    long *rw = (long*) &r;
    
    if( b <= 32 )
    {
        long ol;
        
        rw[0] = aw[0] >> bl;
        rw[1] = aw[1] >> bl;
        ol = (aw[1] << (32 - bl)) >> (32 - bl);
        rw[1] |= ol;
        return r;
    }

    if( b <= 64 )
    {
        rw[0] = aw[1] >> (bl - 32);
        rw[1] = aw[1] >> 32;
        return r;
    }

    printf( "WARNING: overflow in allshr( 0x%08x%08x, 0x%08x%08x )\n", aw[1], aw[0], bw[1], bw[0] );
    return (long) a >> (long) b;
}


/* int64 division */
__int64 _alldiv( __int64 a, __int64 b )
{
    if( (a & 0xffffffff00000000) || (b & 0xffffffff00000000) )
        printf( "WARNING: overflow in alldiv()\n" );

    return (long) a / (long) b;
}


/* int64 remainder */
__int64 _allrem( __int64 a, __int64 b )
{
    if( (a & 0xffffffff00000000) || (b & 0xffffffff00000000) )
        printf( "WARNING: overflow in allrem()\n" );

    return (long) a % (long) b;
}


/* Floating point double to integer long cast */
int32 _ftol2( double a )
{
	if( (a > FLT_MAX) || (a < - FLT_MAX) )
	{
        printf( "WARNING: overflow in ftol2( %f )\n", a );
		return 0;
	}

	return _ftol( (float) a );
}