/******************************************************************************
 *
 * Filename: native.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for command line runtime.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "config.h"
#include "types.h"
#include "disasm.h"
#include "util.h"
#include "hashtable.h"
#include "openfiles.h"
#include "memory.h"

#include "../native/java_io_File.h"
#include "../native/java_io_FileDescriptor.h"
#include "../native/java_io_FileInputStream.h"
#include "../native/java_io_FileOutputStream.h"
#include "../native/java_io_ObjectInputStream.h"
#include "../native/java_io_ObjectOutputStream.h"
#include "../native/java_io_ObjectStreamClass.h"
#include "../native/java_io_RandomAccessFile.h"
#include "../native/java_lang_Class.h"
#include "../native/java_lang_ClassLoader.h"
#include "../native/java_lang_Object.h"
#include "../native/java_lang_Runtime.h"
#include "../native/java_lang_SecurityManager.h"
#include "../native/java_lang_String.h"
#include "../native/java_lang_System.h"
#include "../native/java_lang_Thread.h"
#include "../native/java_lang_Throwable.h"
#include "../native/java_util_ResourceBundle.h"

#if JDK_VERSION < 140
#include "../native/java_lang_Math.h"
#include "../native/java_lang_reflect_Constructor.h"
#include "../native/java_lang_reflect_Field.h"
#endif /* JDK 1.1 */

#if JDK_VERSION >= 140
#if USES_UNIXFILESYSTEM
#include "../native/java_io_UnixFileSystem.h"
#else
#include "../native/java_io_Win32FileSystem.h"
#endif /* USES_UNIXFILESYSTEM */
#include "../native/java_lang_StrictMath.h"
#include "../native/java_security_AccessController.h"
#include "../native/sun_misc_Signal.h"
#include "../native/sun_misc_Unsafe.h"
#include "../native/sun_reflect_NativeConstructorAccessorImpl.h"
#include "../native/sun_reflect_Reflection.h"
#endif /* JDK 1.4 */

/* From Zip sub-project */
#if JDK_VERSION >= 140
#include "../zip/java_util_jar_JarFile.h"
#include "../zip/java_util_zip_ZipEntry.h"
#include "../zip/java_util_zip_ZipFile.h"
#endif /* JDK 1.4 */

#include "native.h"


Hashtable* nativesTable = NULL;


OpenFile_Entry* openfilesv[255];
int32 openfilesc = 3;


#define STRING_MAX 256 /* 1024 */

#define CHECK_OBJREF_NOT_NULL( objref, stp )  	\
	{ if( !objref ) return newCompleteException( stp, NullPointerException, "" ); }

#define CHECK_OBJREF_CF_AND_DATA( objref, stp )  \
	{ if( !objref->cf || (objref->type != 0) ) return newCompleteException( stp, InternalError, "corrupt object reference" ); }

#define CHECK_OBJREF_TYPE( objref, type, stp )	\
	{ if( !checkType(objref->cf,type) ) return newCompleteException( stp, InternalError, "object reference not of type \"" type "\"" ); }

#define CHECK_OBJREF_DATA( objref, stp ) \
	{ if( !objref->data ) return newCompleteException( stp, InternalError, "bad object reference" ); }

#define CHECK_OBJREF( objref, fp )  { CHECK_OBJREF_NOT_NULL(objref,fp); CHECK_OBJREF_CF_AND_DATA(objref,fp); }

#define CHECK_OBJREF_AND_TYPE( objref, type, fp )  { \
	CHECK_OBJREF_NOT_NULL(objref,fp); CHECK_OBJREF_CF_AND_DATA(objref,fp); \
	CHECK_OBJREF_TYPE(objref,type,fp); CHECK_OBJREF_DATA(objref,fp); }


/******************************************************************************
 *
 * Wrapper functions
 *
 *****************************************************************************/

typedef void     (*fn_V)( JNIEnv* env, jobject self );
typedef jboolean (*fn_Z)( JNIEnv* env, jobject self );
typedef jint     (*fn_I)( JNIEnv* env, jobject self );
typedef jlong    (*fn_J)( JNIEnv* env, jobject self );
typedef jobject  (*fn_L)( JNIEnv* env, jobject self );


/* void fn() - ()V */
int wrapper_V( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_V fn = (fn_V) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* boolean fn() - ()Z */
int wrapper_Z( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_Z fn = (fn_Z) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	int32 ret;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* int fn() - ()I */
int wrapper_I( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_I fn = (fn_I) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	int32 ret;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* long fn() - ()J */
int wrapper_J( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_J fn = (fn_J) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	u4 retw[2];
	int64* ret = (int64*) retw;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	*ret = (*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_LONG( fp, retw );
	return 0;
}


/* object fn() - ()Lxxx; */
int wrapper_L( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_L fn = (fn_L) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *ret;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/******************************** 1 arg **************************************/


typedef void     (*fn_ZV)( JNIEnv* env, jobject self, jboolean arg1 );
typedef void     (*fn_IV)( JNIEnv* env, jobject self, jint arg1 );
typedef void     (*fn_JV)( JNIEnv* env, jobject self, jlong arg1 );
typedef void     (*fn_LV)( JNIEnv* env, jobject self, jobject arg1 );

typedef jboolean (*fn_LZ)( JNIEnv* env, jobject self, jobject arg1 );

typedef jbyte    (*fn_JB)( JNIEnv* env, jobject self, jlong arg1 );

typedef jint     (*fn_LI)( JNIEnv* env, jobject self, jobject arg1 );
typedef jlong    (*fn_LJ)( JNIEnv* env, jobject self, jobject arg1 );

typedef jlong    (*fn_JJ)( JNIEnv* env, jobject self, jlong arg1 );

typedef jobject  (*fn_ZL)( JNIEnv* env, jobject self, jboolean arg1 );
typedef jobject  (*fn_IL)( JNIEnv* env, jobject self, jint arg1 );
typedef jobject  (*fn_LL)( JNIEnv* env, jobject self, jobject arg1 );


/* void fn(boolean) - (Z)V */
int wrapper_ZV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_ZV fn = (fn_ZV) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	int arg1;
	
	POP_INT( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* void fn(int) - (I)V */
int wrapper_IV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_IV fn = (fn_IV) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	int32 arg1;
	
	POP_INT( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* void fn(long) - (J)V */
int wrapper_JV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_JV fn = (fn_JV) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	u4 value[2];
	int64* arg1 = (int64*) value;

	POP_LONG( fp, value );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, *arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* void fn(object) - (Lxxx;)V */
int wrapper_LV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LV fn = (fn_LV) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );
	CHECK_OBJREF( arg1, stp );

	(*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* boolean fn(object) - (Lxxx;)Z */
int wrapper_LZ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LZ fn = (fn_LZ) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	int ret;
	
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


#if JDK_VERSION >= 140
/* byte fn(long) - (J)B */
int wrapper_JB( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_JB fn = (fn_JB) gfn;
	Frame* fp = stp->fp;
	ObjRef *self;
	u4 value[2];
	int64* l = (int64*) value;
	int64 arg1;
	jbyte ret;
	
	POP_LONG( fp, value );
	arg1 = *l;
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}
#endif /* JDK 1.4 */


/* int fn(object) - (Lxxx;)I */
int wrapper_LI( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LI fn = (fn_LI) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	int32 ret;
	
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


#if JDK_VERSION >= 140
/* long fn(object) - (Lxxx;)J */
int wrapper_LJ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LJ fn = (fn_LJ) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	u4 value[2];
	int64* l = (int64*) value;
	int64 ret;
	
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	*l = ret;
	PUSH_LONG( fp, value );
	return 0;
}
#endif /* JDK 1.4 */


/* long fn(long) - (J)J */
int wrapper_JJ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_JJ fn = (fn_JJ) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	u4 value[2];
	int64* l = (int64*) value;
	int64 arg1, ret;

	POP_LONG( fp, value );
	arg1 = *l;
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	*l = ret;
	PUSH_LONG( fp, value );
	return 0;
}


#if JDK_VERSION >= 140
/* object fn(boolean) - (Z)Lxxx; */
int wrapper_ZL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_ZL fn = (fn_ZL) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *ret;
	int32 arg1;
	
	POP_INT( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}
#endif /* JDK 1.4 */


/* object fn(int) - (I)Lxxx; */
int wrapper_IL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_IL fn = (fn_IL) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *ret;
	int32 arg1;
	
	POP_INT( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/* object fn(object) - (Lxxx;)Lxxx; */
int wrapper_LL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LL fn = (fn_LL) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1, *ret;
	
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/******************************** 2 args *************************************/


typedef void     (*fn_JJV)( JNIEnv* env, jobject self, jlong arg1, jlong arg2 );
typedef void     (*fn_LZV)( JNIEnv* env, jobject self, jobject arg1, jboolean arg2 );
typedef jobject  (*fn_LIL)( JNIEnv* env, jobject self, jobject arg1, jint arg2 );
typedef jobject  (*fn_LLL)( JNIEnv* env, jobject self, jobject arg1, jobject arg2 );


#if JDK_VERSION >= 140
/* void fn(long,long) - (JJ)V */
int wrapper_JJV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_JJV fn = (fn_JJV) gfn;
	Frame* fp = stp->fp;
	ObjRef *self;
	u4 value[2];
	int64* l = (int64*) value;
	int64 arg1, arg2;

	POP_LONG( fp, value );
	arg2 = *l;

	POP_LONG( fp, value );
	arg1 = *l;
	
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}
#endif /* JDK 1.4 */


/* void fn(object,object) - (Lxxx;Z)V; */
int wrapper_LZV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LZV fn = (fn_LZV) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	jboolean arg2;
	
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* object fn(object,int) - (Lxxx;I)Lxxx; */
int wrapper_LIL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LIL fn = (fn_LIL) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1, *ret;
	int32 arg2;
	
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/* object fn(object,object) - (Lxxx;Lxxx;)Lxxx; */
int wrapper_LLL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LLL fn = (fn_LLL) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1, *arg2, *ret;
	
	POP_OBJREF( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/******************************** 3 args *************************************/


typedef void     (*fn_LLLV)( JNIEnv* env, jobject self, jobject arg1, jobject arg2, jobject arg3 );
typedef void     (*fn_LIIV)( JNIEnv* env, jobject self, jobject arg1, jint arg2, jint arg3 );
typedef jint     (*fn_LIII)( JNIEnv* env, jobject self, jobject arg1, jint arg2, jint arg3 );


/* void fn(object,object,object) - (Lxxx;Lxxx;Lxxx;)V */
int wrapper_LLLV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LLLV fn = (fn_LLLV) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1, *arg2, *arg3;
	
	POP_OBJREF( fp, arg3 );
	POP_OBJREF( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1, arg2, arg3 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* void fn(object,int,int) - (Lxxx;II)V */
int wrapper_LIIV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LIII fn = (fn_LIII) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	int32 arg2, arg3;
	
	POP_INT( fp, arg3 );
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1, arg2, arg3 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* int fn(object,int,int) - (Lxxx;II)I */
int wrapper_LIII( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LIII fn = (fn_LIII) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	int32 arg2, arg3, ret;
	
	POP_INT( fp, arg3 );
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1, arg2, arg3 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/**************************** static no args ********************************/


typedef void     (*fn_static_V)( JNIEnv* env, jclass staticself );
typedef jlong    (*fn_static_J)( JNIEnv* env, jclass staticself );
typedef jobject  (*fn_static_L)( JNIEnv* env, jclass staticself );


#if JDK_VERSION >= 140
/* static void fn() - ()V */
int wrapper_static_V( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_V fn = (fn_static_V) gfn;
	Frame* fp = stp->fp;

	(*fn)( stp, staticself );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}
#endif /* JDK_VERSION 1.4.0 */


/* static long fn() - ()J */
int wrapper_static_J( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_J fn = (fn_static_J) gfn;
	Frame* fp = stp->fp;
	u4 retw[2];
	int64* ret = (int64*) retw;

	*ret = (*fn)( stp, staticself );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_LONG( fp, retw );
	return 0;
}


/* static object fn() - ()L */
int wrapper_static_L( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_L fn = (fn_static_L) gfn;
	Frame* fp = stp->fp;
	ObjRef* ret;

	ret = (*fn)( stp, staticself );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/************************** static 1 arg void ********************************/


typedef void     (*fn_static_IV)( JNIEnv* env, jclass staticself, jint arg1 );
typedef void     (*fn_static_JV)( JNIEnv* env, jclass staticself, jlong arg1 );
typedef void     (*fn_static_LV)( JNIEnv* env, jclass staticself, jobject arg1 );


/* static void fn(int) - (I)V */
int wrapper_static_IV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_IV fn = (fn_static_IV) gfn;
	Frame* fp = stp->fp;
	int32 arg1;
	
	POP_INT( fp, arg1 );

	(*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


#if JDK_VERSION >= 140
/* static void fn(long) - (J)V */
int wrapper_static_JV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_JV fn = (fn_static_JV) gfn;
	Frame* fp = stp->fp;
	int64 arg1;
	u4 value[2];
	int64 *l = (int64*) value;
	
	POP_LONG( fp, value );
	arg1 = *l;

	(*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}
#endif /* JDK 1.4 */


/* static void fn(object) - (Lxxx;)V */
int wrapper_static_LV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LV fn = (fn_static_LV) gfn;
	Frame* fp = stp->fp;
	ObjRef* arg1;
	
	POP_OBJREF( fp, arg1 );

	(*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/************************** static 1 arg retval ******************************/


typedef jboolean (*fn_static_LZ)( JNIEnv* env, jclass staticself, jobject arg1 );
typedef jint     (*fn_static_JI)( JNIEnv* env, jclass staticself, jlong arg1 );
typedef jint     (*fn_static_LI)( JNIEnv* env, jclass staticself, jobject arg1 );
typedef jlong    (*fn_static_LJ)( JNIEnv* env, jclass staticself, jobject arg1 );
typedef jdouble  (*fn_static_DD)( JNIEnv* env, jclass staticself, jdouble arg1 );
typedef jobject  (*fn_static_IL)( JNIEnv* env, jclass staticself, jint arg1 );
typedef jobject  (*fn_static_JL)( JNIEnv* env, jclass staticself, jlong arg1 );
typedef jobject  (*fn_static_LL)( JNIEnv* env, jclass staticself, jobject arg1 );


/* static boolean fn(object) - (Lxxx;)Z */
int wrapper_static_LZ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LZ fn = (fn_static_LZ) gfn;
	Frame* fp = stp->fp;
	ObjRef* arg1;
	int ret;
	
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


#if JDK_VERSION >= 140
/* static int fn(long) - (J)I */
int wrapper_static_JI( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_JI fn = (fn_static_JI) gfn;
	Frame* fp = stp->fp;
	int64 arg1;
	u4 value[2];
	int64 *l = (int64*) value;
	int32 ret;
	
	POP_LONG( fp, value );
	arg1 = *l;

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}
#endif /* JDK 1.4 */


/* static int fn(object) - (Lxxx;)I */
int wrapper_static_LI( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LI fn = (fn_static_LI) gfn;
	Frame* fp = stp->fp;
	ObjRef* arg1;
	int32 ret;
	
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* static long fn(object) - (Lxxx;)J */
int wrapper_static_LJ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LJ fn = (fn_static_LJ) gfn;
	Frame* fp = stp->fp;
	ObjRef* arg1;
	u4 value[2];
	int64* ret = (int64*) value;
	
	POP_OBJREF( fp, arg1 );

	*ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_LONG( fp, value );
	return 0;
}


/* static double fn(double) - (D)D */
int wrapper_static_DD( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_DD fn = (fn_static_DD) gfn;
	Frame* fp = stp->fp;
	u4 value[2];
	double* d = (double*) value;
	double arg1, ret;

	POP_DOUBLE( fp, value );
	arg1 = *d;

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	*d = ret;
	PUSH_DOUBLE( fp, value );
	return 0;
}


#if JDK_VERSION >= 140
/* static object fn(int) - (I)Lxxx; */
int wrapper_static_IL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_IL fn = (fn_static_IL) gfn;
	Frame* fp = stp->fp;
	ObjRef* ret;
	int32 arg1;
	
	POP_INT( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}
#endif /* JDK_VERSION 1.4.0 */


#if JDK_VERSION >= 140
/* static object fn(long) - (J)Lxxx; */
int wrapper_static_JL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_JL fn = (fn_static_JL) gfn;
	Frame* fp = stp->fp;
	ObjRef* ret;
	int64 arg1;
	u4 value[2];
	int64 *l = (int64*) value;
	
	POP_LONG( fp, value );
	arg1 = *l;

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}
#endif /* JDK_VERSION 1.4.0 */


/* static object fn(object) - (Lxxx;)Lxxx; */
int wrapper_static_LL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LL fn = (fn_static_LL) gfn;
	Frame* fp = stp->fp;
	ObjRef *arg1, *ret;
	
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/************************** static 2 args ************************************/

typedef void     (*fn_static_JJV)( JNIEnv* env, jclass staticself, jlong arg1, jlong arg2 );
typedef jint     (*fn_static_LLI)( JNIEnv* env, jclass staticself, jobject arg1, jobject arg2 );
typedef jlong    (*fn_static_IJJ)( JNIEnv* env, jclass staticself, jint arg1, jlong arg2 );
typedef jlong    (*fn_static_JIJ)( JNIEnv* env, jclass staticself, jlong arg1, jint arg2 );
typedef jlong    (*fn_static_JLJ)( JNIEnv* env, jclass staticself, jlong arg1, jobject arg2 );
typedef jobject  (*fn_static_LIL)( JNIEnv* env, jclass staticself, jobject arg1, jint arg2 );
typedef jobject  (*fn_static_LLL)( JNIEnv* env, jclass staticself, jobject arg1, jobject arg2 );


#if JDK_VERSION >= 140
/* static void fn(long,long) - (JJ)V */
int wrapper_static_JJV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_JJV fn = (fn_static_JJV) gfn;
	Frame* fp = stp->fp;
	int64 arg1, arg2;
	u4 value[2];
	int64* l = (int64*) value;
	
	POP_LONG( fp, value );
	arg2 = *l;
	POP_LONG( fp, value );
	arg1 = *l;

	(*fn)( stp, staticself, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}
#endif /* JDK 1.4 */


/* static int fn(object,object) - (Lxxx;Lxxx;)I */
int wrapper_static_LLI( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LLI fn = (fn_static_LLI) gfn;
	Frame* fp = stp->fp;
	ObjRef *arg1, *arg2;
	int32 ret;
	
	POP_OBJREF( fp, arg2 );
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


#if JDK_VERSION >= 140
/* static long fn(int,long) - (IJ)J */
int wrapper_static_IJJ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_IJJ fn = (fn_static_IJJ) gfn;
	Frame* fp = stp->fp;
	int32 arg1;
	int64 ret, arg2;
	u4 value[2];
	int64* l = (int64*) value;
	
	POP_LONG( fp, value );
	arg2 = *l;
	POP_INT( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	*l = ret;
	PUSH_LONG( fp, value );
	return 0;
}
#endif /* JDK 1.4 */


#if JDK_VERSION >= 140
/* static long fn(long,int) - (JI)J */
int wrapper_static_JIJ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_JIJ fn = (fn_static_JIJ) gfn;
	Frame* fp = stp->fp;
	int32 arg2;
	int64 ret, arg1;
	u4 value[2];
	int64* l = (int64*) value;
	
	POP_INT( fp, arg2 );
	POP_LONG( fp, value );
	arg1 = *l;

	ret = (*fn)( stp, staticself, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	*l = ret;
	PUSH_LONG( fp, value );
	return 0;
}
#endif /* JDK 1.4 */


#if JDK_VERSION >= 140
/* static long fn(long,object) - (JL)J */
int wrapper_static_JLJ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_JLJ fn = (fn_static_JLJ) gfn;
	Frame* fp = stp->fp;
	int64 ret, arg1;
	ObjRef *arg2;
	u4 value[2];
	int64* l = (int64*) value;
	
	POP_OBJREF( fp, arg2 );
	POP_LONG( fp, value );
	arg1 = *l;

	ret = (*fn)( stp, staticself, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	*l = ret;
	PUSH_LONG( fp, value );
	return 0;
}
#endif /* JDK 1.4 */


/* static object fn(object,int) - (Lxxx;I)Lxxx; */
int wrapper_static_LIL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LIL fn = (fn_static_LIL) gfn;
	Frame* fp = stp->fp;
	ObjRef *arg1, *ret;
	int32 arg2;
	
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


#if JDK_VERSION >= 140
/* static object fn(object,object) - (Lxxx;Lxxx;)Lxxx; */
int wrapper_static_LLL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LLL fn = (fn_static_LLL) gfn;
	Frame* fp = stp->fp;
	ObjRef *arg1, *arg2, *ret;
	
	POP_OBJREF( fp, arg2 );
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}
#endif /* JDK 1.4 */


/************************** static 3 args ************************************/


typedef jlong    (*fn_static_LIJJ)( JNIEnv* env, jclass staticself, jobject arg1, jint arg2, jlong arg3 );
typedef jobject  (*fn_static_LZLL)( JNIEnv* env, jclass staticself, jobject arg1, jint arg2, jobject arg3 );


#if JDK_VERSION >= 140
/* static long fn(object,int,long) - (Lxxx;I;J)J */
int wrapper_static_LIJJ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LIJJ fn = (fn_static_LIJJ) gfn;
	Frame* fp = stp->fp;
	ObjRef *arg1;
	int32 arg2;
	int64 arg3, ret;
	u4 value[2];
	int64 *l = (int64*) value;
	
	POP_LONG( fp, value );
	arg3 = *l;
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1, arg2, arg3 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	*l = ret;
	PUSH_LONG( fp, value );
	return 0;
}
#endif /* JDK_VERSION 1.4.0 */


#if JDK_VERSION >= 140
/* static object fn(object,int,object) - (Lxxx;I;Lxxx)Lxxx; */
int wrapper_static_LZLL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LZLL fn = (fn_static_LZLL) gfn;
	Frame* fp = stp->fp;
	ObjRef *arg1, *arg3, *ret;
	int32 arg2;
	
	POP_OBJREF( fp, arg3 );
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1, arg2, arg3 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}
#endif /* JDK_VERSION 1.4.0 */


/************************** static 6 args ************************************/


typedef jint    (*fn_static_JJILIII)( JNIEnv* env, jclass staticself, jlong arg1, jlong arg2, jint arg3, jobject arg4, jint arg5, jint arg6 );


#if JDK_VERSION >= 140
/* static long fn(object,int,long) - (Lxxx;I;J)J */
int wrapper_static_JJILIII( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_JJILIII fn = (fn_static_JJILIII) gfn;
	Frame* fp = stp->fp;
	int64 arg1, arg2;
	int32 arg3, arg5, arg6, ret;
	ObjRef *arg4;
	u4 value[2];
	int64 *l = (int64*) value;
	
	POP_INT( fp, arg6 );
	POP_INT( fp, arg5 );
	POP_OBJREF( fp, arg4 );
	POP_INT( fp, arg3 );
	POP_LONG( fp, value );
	arg2 = *l;
	POP_LONG( fp, value );
	arg1 = *l;

	ret = (*fn)( stp, staticself, arg1, arg2, arg3, arg4, arg5, arg6 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}
#endif /* JDK_VERSION 1.4.0 */


/******************************************************************************
 *
 * Special methods which can't be handled by wrapper fn's
 *
 *****************************************************************************/


/******************************************************************************
 *
 * class java.lang.Class
 *
 *****************************************************************************/

#if JDK_VERSION < 140

/* java.lang.Object newInstance() 
 *     throws InstantiationException, IllegalAccessException; */
int Class_newInstance( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	ObjRef* self;
	ClassFile* cf;
	ObjRef* newobj;
	int retcode;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	cf = (ClassFile*) self->data[self->cf->object_size];	/* Last field is cf */
	newobj = newObjectRefAndObject( cf );
	
	/* need to leave newobj on the stack AFTER <init> has been called */
	PUSH_OBJREF( fp, newobj );

	retcode = forceExecuteJavaMethod( newobj, stp, "<init>", "()V" );
	/* whereAreWe( fp ); */

	return retcode;
}

#endif /* JDK 1.1 */


/******************************************************************************
 *
 * class java.lang.Double
 *
 *****************************************************************************/

/* static long doubleToLongBits( double d ); */
int Double_doubleToLongBits( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	u4 value[2];

	POP_DOUBLE( fp, value );
	PUSH_LONG( fp, value );
	return 0;
}


/* static double longBitsToDouble( long bits ); */
int Double_longBitsToDouble( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	u4 value[2];

	POP_LONG( fp, value );
	PUSH_DOUBLE( fp, value );
	return 0;
}


#if JDK_VERSION >= 140
/******************************************************************************
 *
 * class java.io.FileSystem
 *
 *****************************************************************************/

/**
 * @since  JDK1.4 (at least JDK1.2 or later)
 *
 * public static native java.io.FileSystem getFileSystem();
 */
int java_io_FileSystem_getFileSystem( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	ClassFile* fsyscf;
	ObjRef* newobj;
	int retcode;

#if USES_UNIXFILESYSTEM
	if( !(fsyscf = getClass( stp, "java/io/UnixFileSystem" )) )
#else
	if( !(fsyscf = getClass( stp, "java/io/Win32FileSystem" )) )
#endif /* USES_UNIXFILESYSTEM */
		return stp->retcode;	/* Exception on stack */
	newobj = newObjectRefAndObject( fsyscf );
	
	/* need to leave newobj on the stack AFTER <init> has been called */
	PUSH_OBJREF( fp, newobj );

	retcode = forceExecuteJavaMethod( newobj, stp, "<init>", "()V" );

	return retcode;
}
#endif /* JDK 1.4 */


/******************************************************************************
 *
 * class java.lang.Float
 *
 *****************************************************************************/

/* static int floatToIntBits( float f ); */
int Float_floatToIntBits( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	u4 value;

	POP_FLOAT( fp, value );
	PUSH_INT( fp, value );
	return 0;
}


/* static float intBitsToFloat( int bits ); */
int Float_intBitsToFloat( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	u4 value;

	POP_INT( fp, value );
	PUSH_FLOAT( fp, value );
	return 0;
}


/******************************************************************************
 *
 * class java.lang.Object
 *
 *****************************************************************************/

/* Object clone() */
int java_lang_Object_clone( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	ObjRef *self, *ret;

	POP_OBJREF( fp, self );

	/* Is this an array?  Object.clone() is overridden by arrays: */
	if( self && !self->cf && (self->type != 0) )
	{
		ObjRef* clone = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
		clone->cf     = self->cf;
		clone->type   = self->type;
		clone->data	  = NULL;
		clone->length = self->length;

		if( self->data )
		{
			size_t size = sizeOfJavaType(self->type) * self->length;
			/*printf( "Cloning array (%d bytes).\n", size );*/

			clone->data = (u4*) javavm_malloc( size );
			memcpy( clone->data, self->data, size );
		}

		PUSH_OBJREF( fp, clone );
		return 0;
	}

	CHECK_OBJREF( self, stp );

	ret = Java_java_lang_Object_clone( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/* Object getClass() */
int java_lang_Object_getClass( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	ObjRef *self, *ret;

	POP_OBJREF( fp, self );

	/* Is this null or NOT an array? If so, validate the objectref. */
	if( !self || self->cf || (self->type == 0) )
	{
		CHECK_OBJREF( self, stp );
	}

	ret = Java_java_lang_Object_getClass( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/* int hashCode() */
int java_lang_Object_hashCode( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	ObjRef* self;
	int32 ret;

	POP_OBJREF( fp, self );

	/* Is this null or NOT an array? If so, validate the objectref. */
	if( !self || self->cf || (self->type == 0) )
	{
		CHECK_OBJREF( self, stp );
	}

	ret = Java_java_lang_Object_hashCode( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* void notifyAll() */
int java_lang_Object_notifyAll( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	ObjRef* self;

	POP_OBJREF( fp, self );

	/* Is this null or NOT an array? If so, validate the objectref. */
	if( !self || self->cf || (self->type == 0) )
	{
		CHECK_OBJREF( self, stp );
	}

	Java_java_lang_Object_notifyAll( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


#if JDK_VERSION < 140
/******************************************************************************
 *
 * class java.lang.Runtime
 *
 *****************************************************************************/

/* void exitInternal( int status ) */
int Runtime_exitInternal( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	wrapper_IV( stp, staticself, (generic_fn) &Java_java_lang_Runtime_exitInternal );
	return 1; /* exit */
}
#endif /* JDK 1.1 */


#if JDK_VERSION >= 140
/******************************************************************************
 *
 * class java.lang.Shutdown
 *
 *****************************************************************************/

/* void halt( int exitCode ) */
int Shutdown_halt( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	stp->retcode = 1; /* exit */
	return 1; /* exit */
}
#endif /* JDK 1.4 */


/******************************************************************************
 *
 * class java.lang.System
 *
 *****************************************************************************/

/* static void arraycopy( java.lang.Object src, int src_position,
 *                        java.lang.Object dst, int dst_position,
 *                        int length );
 */
int System_arraycopy( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	ObjRef *src, *dst;
	u4 src_position, dst_position, length;

	POP_INT( fp, length  );
	POP_INT( fp, dst_position );
	POP_OBJREF( fp, dst );       /* ### POP_ARRAY */
	POP_INT( fp, src_position );
	POP_OBJREF( fp, src );       /* ### POP_ARRAY */

	if( !src || !dst )
		return newCompleteException( stp, NullPointerException, "" );

	if( ((src->type == 0) && !src->cf ) || ((dst->type == 0) && !dst->cf ) )
		return newCompleteException( stp, InternalError, "Corrupt array reference" );

	Java_java_lang_System_arraycopy( stp, staticself, src, src_position, dst, dst_position, length );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


#if JDK_VERSION >= 140
/******************************************************************************
 *
 * class sun.misc.AtomicLong
 *
 *****************************************************************************/

/**
 * @since  JDK1.4 (at least JDK1.2 or later)
 *
 * private static native boolean VMSupportsCS8();
 */
int sun_misc_AtomicLong_VMSupportsCS8( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	PUSH_INT( stp->fp, 0 );		/* Always return false */
	return 0;
}
#endif /* JDK 1.4 */


/*****************************************************************************/


typedef struct _NativeMethodInitStr
{
	const char* sig;
	NativeMethodStr fn;
} NativeMethodInitStr;


NativeMethodInitStr methods[] =
{
#if JDK_VERSION >= 140
	{ "java/lang/Class" "."			    "desiredAssertionStatus0" "(Ljava/lang/Class;)Z",				 { &wrapper_static_LZ, (generic_fn) &Java_java_lang_Class_desiredAssertionStatus0 } },
#endif /* JDK 1.4 */
#if JDK_VERSION < 140
	{ "java/lang/Class" "."			    "forName"				"(Ljava/lang/String;)Ljava/lang/Class;", { &wrapper_static_LL, (generic_fn) &Java_java_lang_Class_forName } },
	{ "java/lang/Class" "."			    "getClassLoader"		"()Ljava/lang/ClassLoader;",			 { &wrapper_L, (generic_fn) &Java_java_lang_Class_getClassLoader } },
#else
	{ "java/lang/Class" "."			    "forName0"				"(Ljava/lang/String;ZLjava/lang/ClassLoader;)Ljava/lang/Class;", { &wrapper_static_LZLL, (generic_fn) &Java_java_lang_Class_forName0 } },
	{ "java/lang/Class" "."			    "getClassLoader0"		"()Ljava/lang/ClassLoader;",			 { &wrapper_L, (generic_fn) &Java_java_lang_Class_getClassLoader0 } },
#endif /* JDK 1.4 */
#if JDK_VERSION < 140
	{ "java/lang/Class" "."			    "getConstructors0"		"(I)[Ljava/lang/reflect/Constructor;",   { &wrapper_IL, (generic_fn) &Java_java_lang_Class_getConstructors0 } },
	{ "java/lang/Class" "."			    "getConstructor0"		"([Ljava/lang/Class;I)Ljava/lang/reflect/Constructor;", { &wrapper_LIL, (generic_fn) &Java_java_lang_Class_getConstructor0 } },
	{ "java/lang/Class" "."			    "getFields0"			"(I)[Ljava/lang/reflect/Field;",		 { &wrapper_IL, (generic_fn) &Java_java_lang_Class_getFields0 } },
	{ "java/lang/Class" "."			    "getField0"				"(Ljava/lang/String;I)Ljava/lang/reflect/Field;", { &wrapper_LIL, (generic_fn) &Java_java_lang_Class_getField0 } },
#else
	{ "java/lang/Class" "."			    "getDeclaredConstructors0" "(Z)[Ljava/lang/reflect/Constructor;", { &wrapper_ZL, (generic_fn) &Java_java_lang_Class_getDeclaredConstructors0 } },
	{ "java/lang/Class" "."			    "getDeclaredFields0"	"(Z)[Ljava/lang/reflect/Field;",		 { &wrapper_ZL, (generic_fn) &Java_java_lang_Class_getDeclaredFields0 } },
#endif /* JDK 1.4 */
	{ "java/lang/Class" "."			    "getInterfaces"		    "()[Ljava/lang/Class;",					 { &wrapper_L, (generic_fn) &Java_java_lang_Class_getInterfaces } },
	{ "java/lang/Class" "."				"getModifiers"			"()I",									 { &wrapper_I, (generic_fn) &Java_java_lang_Class_getModifiers } },
	{ "java/lang/Class" "."			    "getName"				"()Ljava/lang/String;",					 { &wrapper_L, (generic_fn) &Java_java_lang_Class_getName } },
	{ "java/lang/Class" "."			    "getPrimitiveClass" 	"(Ljava/lang/String;)Ljava/lang/Class;", { &wrapper_static_LL, (generic_fn) &Java_java_lang_Class_getPrimitiveClass } },
	{ "java/lang/Class" "."			    "getSuperclass" 		"()Ljava/lang/Class;",					 { &wrapper_L, (generic_fn) &Java_java_lang_Class_getSuperclass } },
	{ "java/lang/Class" "."			    "isArray"				"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_lang_Class_isArray } },
	{ "java/lang/Class" "."			    "isAssignableFrom"		"(Ljava/lang/Class;)Z",					 { &wrapper_LZ, (generic_fn) &Java_java_lang_Class_isAssignableFrom } },
	{ "java/lang/Class" "."			    "isInterface"			"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_lang_Class_isInterface } },
	{ "java/lang/Class" "."			    "isPrimitive"			"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_lang_Class_isPrimitive } },
#if JDK_VERSION < 140
	{ "java/lang/Class" "."			    "newInstance"			"()Ljava/lang/Object;",					 { &Class_newInstance, NULL } },
#endif /* JDK 1.1 */
#if JDK_VERSION >= 140
	{ "java/lang/Class"	"." 		    "registerNatives"	    "()V",									 { &wrapper_static_V, (generic_fn) &Java_java_lang_Class_registerNatives } },
#endif /* JDK 1.4 */
#if JDK_VERSION < 140
	{ "java/lang/ClassLoader" "."		"init"					"()V",									 { &wrapper_V, (generic_fn) &Java_java_lang_ClassLoader_init } },
	{ "java/lang/ClassLoader" "."		"getSystemResourceAsStream0" "(Ljava/lang/String;)Ljava/io/InputStream;", { &wrapper_static_LL, (generic_fn) &Java_java_lang_ClassLoader_getSystemResourceAsStream0 } },
	{ "java/lang/ClassLoader" "."		"findSystemClass0"		"(Ljava/lang/String;)Ljava/lang/Class;", { &wrapper_LL, (generic_fn) &Java_java_lang_ClassLoader_findSystemClass0 } },
#endif /* JDK 1.1 */
#if JDK_VERSION >= 140
	{ "java/lang/ClassLoader" "."		"findBootstrapClass"	"(Ljava/lang/String;)Ljava/lang/Class;", { &wrapper_LL, (generic_fn) &Java_java_lang_ClassLoader_findBootstrapClass } },
	{ "java/lang/ClassLoader" "."		"findLoadedClass"		"(Ljava/lang/String;)Ljava/lang/Class;", { &wrapper_LL, (generic_fn) &Java_java_lang_ClassLoader_findLoadedClass } },
	{ "java/lang/ClassLoader" "."		"registerNatives"	    "()V",									 { &wrapper_static_V, (generic_fn) &Java_java_lang_ClassLoader_registerNatives } },
	{ "java/lang/ClassLoader$NativeLibrary" "."		"load"	    "(Ljava/lang/String;)V",				 { &wrapper_LV, (generic_fn) &Java_java_lang_ClassLoader_NativeLibrary_load } },
#endif /* JDK 1.4 */
	{ "java/lang/Double" "."			"doubleToLongBits"		"(D)J",									 { &Double_doubleToLongBits, NULL } },
	{ "java/lang/Double" "."			"longBitsToDouble"		"(J)D",									 { &Double_longBitsToDouble, NULL } },
	{ "java/lang/Float" "."			    "intBitsToFloat"		"(I)F",									 { &Float_intBitsToFloat, NULL } },
	{ "java/lang/Float" "."			    "floatToIntBits"		"(F)I",									 { &Float_floatToIntBits, NULL } },
#if JDK_VERSION < 140
	{ "java/lang/Math" "."				"floor"				    "(D)D",									 { &wrapper_static_DD, (generic_fn) &Java_java_lang_Math_floor } },
	{ "java/lang/Math" "."				"log"				    "(D)D",									 { &wrapper_static_DD, (generic_fn) &Java_java_lang_Math_log } },
#endif /* JDK 1.1 */
	{ "java/lang/Object" "."			"clone"				    "()Ljava/lang/Object;",					 { &java_lang_Object_clone, NULL } },
	{ "java/lang/Object" "."			"getClass"				"()Ljava/lang/Class;",					 { &java_lang_Object_getClass, NULL } },
	{ "java/lang/Object" "."			"hashCode"				"()I",									 { &java_lang_Object_hashCode, NULL } },
	{ "java/lang/Object" "."			"notifyAll"			    "()V",									 { &java_lang_Object_notifyAll, NULL } },
#if JDK_VERSION >= 140
	{ "java/lang/Object" "."			"registerNatives"	    "()V",									 { &wrapper_static_V, (generic_fn) &Java_java_lang_Object_registerNatives } },
#endif /* JDK 1.4 */
	{ "java/lang/Object" "."			"wait"					"(J)V",									 { &wrapper_JV, (generic_fn) &Java_java_lang_Object_wait } },
#if JDK_VERSION < 140
	{ "java/lang/reflect/Constructor" "." "newInstance"			"([Ljava/lang/Object;)Ljava/lang/Object;", { &wrapper_LL, (generic_fn) &Java_java_lang_reflect_Constructor_newInstance } },
	{ "java/lang/reflect/Field"	"."		"getModifiers"			"()I",									 { &wrapper_I, (generic_fn) &Java_java_lang_reflect_Field_getModifiers } },
	{ "java/lang/Runtime" "."			"buildLibName"			"(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;", { &wrapper_LLL, (generic_fn) &Java_java_lang_Runtime_buildLibName } },
	{ "java/lang/Runtime" "."			"exitInternal"			"(I)V",									 { &Runtime_exitInternal, NULL } },
	{ "java/lang/Runtime" "."			"initializeLinkerInternal" "()Ljava/lang/String;",				 { &wrapper_L, (generic_fn) &Java_java_lang_Runtime_initializeLinkerInternal } },
	{ "java/lang/Runtime" "."			"loadFileInternal"		"(Ljava/lang/String;)I",				 { &wrapper_LI, (generic_fn) &Java_java_lang_Runtime_loadFileInternal } },
#endif /* JDK 1.1 */
	{ "java/lang/Runtime" "."			"traceInstructions" 	"(Z)V",									 { &wrapper_ZV, (generic_fn) &Java_java_lang_Runtime_traceInstructions } },
	{ "java/lang/Runtime" "."			"traceMethodCalls"		"(Z)V",									 { &wrapper_ZV, (generic_fn) &Java_java_lang_Runtime_traceMethodCalls } },
	{ "java/lang/SecurityManager" "."   "classLoaderDepth"		"()I",									 { &wrapper_I, (generic_fn) &Java_java_lang_SecurityManager_classLoaderDepth } },
	{ "java/lang/SecurityManager" "."   "currentClassLoader"	"()Ljava/lang/ClassLoader;",			 { &wrapper_L, (generic_fn) &Java_java_lang_SecurityManager_currentClassLoader } },
	{ "java/lang/SecurityManager" "."   "getClassContext"		"()[Ljava/lang/Class;",					 { &wrapper_L, (generic_fn) &Java_java_lang_SecurityManager_getClassContext } },
#if JDK_VERSION >= 140
	{ "java/lang/Shutdown" "."			"halt"					"(I)V",									 { &Shutdown_halt, NULL } },
	{ "java/lang/StrictMath" "."		"floor"				    "(D)D",									 { &wrapper_static_DD, (generic_fn) &Java_java_lang_StrictMath_floor } },
	{ "java/lang/StrictMath" "."		"log"				    "(D)D",									 { &wrapper_static_DD, (generic_fn) &Java_java_lang_StrictMath_log } },
#endif /* JDK 1.4 */
	{ "java/lang/String" "."			"intern" 				"()Ljava/lang/String;",					 { &wrapper_L, (generic_fn) &Java_java_lang_String_intern } },
	{ "java/lang/System" "."			"arraycopy" 			"(Ljava/lang/Object;ILjava/lang/Object;II)V", { &System_arraycopy, NULL } },
	{ "java/lang/System" "."			"currentTimeMillis" 	"()J",									 { &wrapper_static_J, (generic_fn) &Java_java_lang_System_currentTimeMillis } },
	{ "java/lang/System" "."			"identityHashCode"	 	"(Ljava/lang/Object;)I",				 { &wrapper_static_LI, (generic_fn) &Java_java_lang_System_identityHashCode } },
	{ "java/lang/System" "."			"initProperties"		"(Ljava/util/Properties;)Ljava/util/Properties;", { &wrapper_static_LL, (generic_fn) &Java_java_lang_System_initProperties } },
#if JDK_VERSION >= 140
	{ "java/lang/System" "."			"mapLibraryName"	    "(Ljava/lang/String;)Ljava/lang/String;", { &wrapper_static_LL, (generic_fn) &Java_java_lang_System_mapLibraryName } },
	{ "java/lang/System" "."			"registerNatives"	    "()V",									 { &wrapper_static_V, (generic_fn) &Java_java_lang_System_registerNatives } },
#endif /* JDK 1.4 */
	{ "java/lang/System" "."			"setIn0"				"(Ljava/io/InputStream;)V",				 { &wrapper_static_LV, (generic_fn) &Java_java_lang_System_setIn0 } },
	{ "java/lang/System" "."			"setOut0"				"(Ljava/io/PrintStream;)V",				 { &wrapper_static_LV, (generic_fn) &Java_java_lang_System_setOut0 } },
	{ "java/lang/System" "."			"setErr0"				"(Ljava/io/PrintStream;)V",				 { &wrapper_static_LV, (generic_fn) &Java_java_lang_System_setErr0 } },
	{ "java/lang/Thread" "."			"currentThread" 		"()Ljava/lang/Thread;",					 { &wrapper_static_L, (generic_fn) &Java_java_lang_Thread_currentThread } },
	{ "java/lang/Thread" "."			"isAlive" 				"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_lang_Thread_isAlive } },
#if JDK_VERSION >= 140
	{ "java/lang/Thread" "."			"registerNatives"	    "()V",									 { &wrapper_static_V, (generic_fn) &Java_java_lang_Thread_registerNatives } },
#endif /* JDK 1.4 */
	{ "java/lang/Thread" "."			"setPriority0"			"(I)V",									 { &wrapper_static_IV, (generic_fn) &Java_java_lang_Thread_setPriority0 } },
	{ "java/lang/Thread" "."			"start" 				"()V",									 { &wrapper_V, (generic_fn) &Java_java_lang_Thread_start } },
	{ "java/lang/Throwable" "."		    "fillInStackTrace"		"()Ljava/lang/Throwable;",				 { &wrapper_L, (generic_fn) &Java_java_lang_Throwable_fillInStackTrace } },
#if JDK_VERSION >= 140
	{ "java/lang/Throwable" "."		    "getStackTraceDepth"	"()I",									 { &wrapper_I, (generic_fn) &Java_java_lang_Throwable_getStackTraceDepth } },
	{ "java/lang/Throwable" "."		    "getStackTraceElement"	"(I)Ljava/lang/StackTraceElement;",		 { &wrapper_IL, (generic_fn) &Java_java_lang_Throwable_getStackTraceElement } },
#endif /* JDK 1.4 */
#if JDK_VERSION < 140
	{ "java/lang/Throwable" "."		    "printStackTrace0"		"(Ljava/lang/Object;)V",				 { &wrapper_LV, (generic_fn) &Java_java_lang_Throwable_printStackTrace0 } },
	{ "java/io/File" "."				"canRead0"				"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_io_File_canRead0 } },
	{ "java/io/File" "."				"delete0"				"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_io_File_delete0 } },
	{ "java/io/File" "."				"exists0"				"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_io_File_exists0 } },
	{ "java/io/File" "."				"isAbsolute"			"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_io_File_isAbsolute } },
	{ "java/io/File" "."				"isDirectory0"			"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_io_File_isDirectory0 } },
	{ "java/io/File" "."				"isFile0"				"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_io_File_isFile0 } },
	{ "java/io/File" "."				"lastModified0" 		"()J",									 { &wrapper_J, (generic_fn) &Java_java_io_File_lastModified0 } },
	{ "java/io/File" "."				"list0" 				"()[Ljava/lang/String;",				 { &wrapper_L, (generic_fn) &Java_java_io_File_list0 } },
	{ "java/io/File" "."				"mkdir0"				"()Z",									 { &wrapper_Z, (generic_fn) &Java_java_io_File_mkdir0 } },
#endif /* JDK 1.1 */
#if JDK_VERSION < 140
	{ "java/io/FileDescriptor" "."		"initSystemFD"			"(Ljava/io/FileDescriptor;I)Ljava/io/FileDescriptor;", { &wrapper_static_LIL, (generic_fn) &Java_java_io_FileDescriptor_initSystemFD } },
#else
	{ "java/io/FileDescriptor" "."		"initIDs"				"()V",									 { &wrapper_static_V, (generic_fn) &Java_java_io_FileDescriptor_initIDs } },
#endif /* JDK 1.4 */
#if JDK_VERSION >= 140
	{ "java/io/FileInputStream" "."	    "initIDs"				"()V",									 { &wrapper_static_V, (generic_fn) &Java_java_io_FileInputStream_initIDs } },
#endif /* JDK 1.4 */
	{ "java/io/FileInputStream" "."	    "available"			    "()I",									 { &wrapper_I, (generic_fn) &Java_java_io_FileInputStream_available } },
	{ "java/io/FileInputStream" "."	    "open"					"(Ljava/lang/String;)V",				 { &wrapper_LV, (generic_fn) &Java_java_io_FileInputStream_open } },
	{ "java/io/FileInputStream" "."	    "read"					"()I",									 { &wrapper_I, (generic_fn) &Java_java_io_FileInputStream_read } },
	{ "java/io/FileInputStream" "."	    "readBytes"			    "([BII)I",								 { &wrapper_LIII, (generic_fn) &Java_java_io_FileInputStream_readBytes } },
	{ "java/io/FileInputStream" "."	    "skip"					"(J)J",									 { &wrapper_JJ, (generic_fn) &Java_java_io_FileInputStream_skip } },
#if JDK_VERSION < 140
	{ "java/io/FileInputStream" "."	    "close"				    "()V",									 { &wrapper_V, (generic_fn) &Java_java_io_FileInputStream_close } },
#else
	{ "java/io/FileInputStream" "."	    "close0"			    "()V",									 { &wrapper_V, (generic_fn) &Java_java_io_FileInputStream_close0 } },
#endif /* JDK 1.4 */
#if JDK_VERSION >= 140
	{ "java/io/FileOutputStream" "."	"initIDs"				"()V",									 { &wrapper_static_V, (generic_fn) &Java_java_io_FileOutputStream_initIDs } },
#endif /* JDK 1.4 */
	{ "java/io/FileOutputStream" "."	"open"					"(Ljava/lang/String;)V",				 { &wrapper_LV, (generic_fn) &Java_java_io_FileOutputStream_open } },
	{ "java/io/FileOutputStream" "."	"openAppend"			"(Ljava/lang/String;)V",				 { &wrapper_LV, (generic_fn) &Java_java_io_FileOutputStream_openAppend } },
	{ "java/io/FileOutputStream" "."	"write"				    "(I)V",									 { &wrapper_IV, (generic_fn) &Java_java_io_FileOutputStream_write } },
	{ "java/io/FileOutputStream" "."	"writeBytes"			"([BII)V",								 { &wrapper_LIIV, (generic_fn) &Java_java_io_FileOutputStream_writeBytes } },
#if JDK_VERSION < 140
	{ "java/io/FileOutputStream" "."	"close"				    "()V",									 { &wrapper_V, (generic_fn) &Java_java_io_FileOutputStream_close } },
#else
	{ "java/io/FileOutputStream" "."	"close0"			    "()V",									 { &wrapper_V, (generic_fn) &Java_java_io_FileOutputStream_close0 } },
#endif /* JDK 1.4 */
#if JDK_VERSION >= 140
	{ "java/io/FileSystem" "."			"getFileSystem"			"()Ljava/io/FileSystem;",				 { &java_io_FileSystem_getFileSystem, NULL } },
#endif /* JDK 1.4 */
	{ "java/io/ObjectInputStream" "."   "loadClass0"			"(Ljava/lang/Class;Ljava/lang/String;)Ljava/lang/Class;", { &wrapper_LLL, (generic_fn) &Java_java_io_ObjectInputStream_loadClass0 } },
	{ "java/io/ObjectOutputStream" "."  "getRefHashCode"		"(Ljava/lang/Object;)I",				 { &wrapper_static_LI, (generic_fn) &Java_java_io_ObjectOutputStream_getRefHashCode } },
	{ "java/io/ObjectOutputStream" "."  "outputClassFields"	    "(Ljava/lang/Object;Ljava/lang/Class;[I)V", { &wrapper_LLLV, (generic_fn) &Java_java_io_ObjectOutputStream_outputClassFields } },
	{ "java/io/ObjectStreamClass" "."   "getClassAccess"		"(Ljava/lang/Class;)I",					 { &wrapper_static_LI, (generic_fn) &Java_java_io_ObjectStreamClass_getClassAccess } },
	{ "java/io/ObjectStreamClass" "."   "getFields0"			"(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;", { &wrapper_LL, (generic_fn) &Java_java_io_ObjectStreamClass_getFields0 } },
	{ "java/io/ObjectStreamClass" "."   "getFieldAccess"		"(Ljava/lang/Class;Ljava/lang/String;)I", { &wrapper_static_LLI, (generic_fn) &Java_java_io_ObjectStreamClass_getFieldAccess } },
	{ "java/io/ObjectStreamClass" "."   "getFieldSignatures"	"(Ljava/lang/Class;)[Ljava/lang/String;", { &wrapper_static_LL, (generic_fn) &Java_java_io_ObjectStreamClass_getFieldSignatures } },
	{ "java/io/ObjectStreamClass" "."   "getMethodAccess"		"(Ljava/lang/Class;Ljava/lang/String;)I", { &wrapper_static_LLI, (generic_fn) &Java_java_io_ObjectStreamClass_getMethodAccess } },
	{ "java/io/ObjectStreamClass" "."   "getMethodSignatures"	"(Ljava/lang/Class;)[Ljava/lang/String;", { &wrapper_static_LL, (generic_fn) &Java_java_io_ObjectStreamClass_getMethodSignatures } },
	{ "java/io/ObjectStreamClass" "."   "getSerialVersionUID"	"(Ljava/lang/Class;)J",					 { &wrapper_static_LJ, (generic_fn) &Java_java_io_ObjectStreamClass_getSerialVersionUID } },
	{ "java/io/ObjectStreamClass" "."   "hasWriteObject"		"(Ljava/lang/Class;)Z",					 { &wrapper_static_LZ, (generic_fn) &Java_java_io_ObjectStreamClass_hasWriteObject } },
#if JDK_VERSION >= 140
	{ "java/io/ObjectStreamClass" "."   "initNative"			"()V",									 { &wrapper_static_V, (generic_fn) &Java_java_io_ObjectStreamClass_initNative } },
#endif /* JDK 1.4 */
	{ "java/io/RandomAccessFile" "."	"close"					"()V",									 { &wrapper_V, (generic_fn) &Java_java_io_RandomAccessFile_close } },
	{ "java/io/RandomAccessFile" "."	"length"				"()J",									 { &wrapper_J, (generic_fn) &Java_java_io_RandomAccessFile_length } },
	{ "java/io/RandomAccessFile" "."	"open"					"(Ljava/lang/String;Z)V",				 { &wrapper_LZV, (generic_fn) &Java_java_io_RandomAccessFile_open } },
	{ "java/io/RandomAccessFile" "."	"readBytes"				"([BII)I",								 { &wrapper_LIII, (generic_fn) &Java_java_io_RandomAccessFile_readBytes } },
	{ "java/io/RandomAccessFile" "."	"seek"					"(J)V",									 { &wrapper_JV, (generic_fn) &Java_java_io_RandomAccessFile_seek } },
#if JDK_VERSION >= 140
#if USES_UNIXFILESYSTEM
	{ "java/io/UnixFileSystem" "."		"canonicalize"			"(Ljava/lang/String;)Ljava/lang/String;", { &wrapper_LL, (generic_fn) &Java_java_io_UnixFileSystem_canonicalize } },
	{ "java/io/UnixFileSystem" "."		"getBooleanAttributes0"	"(Ljava/io/File;)I",					 { &wrapper_LI, (generic_fn) &Java_java_io_UnixFileSystem_getBooleanAttributes0 } },
	{ "java/io/UnixFileSystem" "."		"getLastModifiedTime"	"(Ljava/io/File;)J",					 { &wrapper_LJ, (generic_fn) &Java_java_io_UnixFileSystem_getLastModifiedTime } },
	{ "java/io/UnixFileSystem" "."		"getLength"				"(Ljava/io/File;)J",					 { &wrapper_LJ, (generic_fn) &Java_java_io_UnixFileSystem_getLength} },
	{ "java/io/UnixFileSystem" "."		"initIDs"				"()V",									 { &wrapper_static_V, (generic_fn) &Java_java_io_UnixFileSystem_initIDs } },
	{ "java/io/UnixFileSystem" "."		"list"					"(Ljava/io/File;)[Ljava/lang/String;",	 { &wrapper_LL, (generic_fn) &Java_java_io_UnixFileSystem_list } },
#else
	{ "java/io/Win32FileSystem" "."		"canonicalize"			"(Ljava/lang/String;)Ljava/lang/String;", { &wrapper_LL, (generic_fn) &Java_java_io_Win32FileSystem_canonicalize } },
	{ "java/io/Win32FileSystem" "."		"getBooleanAttributes"	"(Ljava/io/File;)I",					 { &wrapper_LI, (generic_fn) &Java_java_io_Win32FileSystem_getBooleanAttributes } },
	{ "java/io/Win32FileSystem" "."		"getLastModifiedTime"	"(Ljava/io/File;)J",					 { &wrapper_LJ, (generic_fn) &Java_java_io_Win32FileSystem_getLastModifiedTime } },
	{ "java/io/Win32FileSystem" "."		"getLength"				"(Ljava/io/File;)J",					 { &wrapper_LJ, (generic_fn) &Java_java_io_Win32FileSystem_getLength} },
	{ "java/io/Win32FileSystem" "."		"initIDs"				"()V",									 { &wrapper_static_V, (generic_fn) &Java_java_io_Win32FileSystem_initIDs } },
	{ "java/io/Win32FileSystem" "."		"list"					"(Ljava/io/File;)[Ljava/lang/String;",	 { &wrapper_LL, (generic_fn) &Java_java_io_Win32FileSystem_list } },
#endif /* USES_UNIXFILESYSTEM */
#endif /* JDK 1.4 */
#if JDK_VERSION >= 140
	{ "java/util/jar/JarFile" "."		"getMetaInfEntryNames"  "()[Ljava/lang/String;",				 { &wrapper_L, (generic_fn) &Java_java_util_jar_JarFile_getMetaInfEntryNames } },
#endif /* JDK 1.4 */
	{ "java/util/ResourceBundle" "."    "getClassContext"		"()[Ljava/lang/Class;",					 { &wrapper_static_L, (generic_fn) &Java_java_util_ResourceBundle_getClassContext } },
#if JDK_VERSION >= 140
	{ "java/util/zip/ZipEntry" "."		"initFields"			"(J)V",									 { &wrapper_JV, (generic_fn) &Java_java_util_zip_ZipEntry_initFields } },
	{ "java/util/zip/ZipEntry" "."		"initIDs"				"()V",									 { &wrapper_static_V, (generic_fn) &Java_java_util_zip_ZipEntry_initIDs } },
	{ "java/util/zip/ZipFile" "."		"close"					"(J)V",									 { &wrapper_static_JV, (generic_fn) &Java_java_util_zip_ZipFile_close } },
	{ "java/util/zip/ZipFile" "."		"freeEntry"				"(JJ)V",								 { &wrapper_static_JJV, (generic_fn) &Java_java_util_zip_ZipFile_freeEntry } },
	{ "java/util/zip/ZipFile" "."		"getCSize"				"(J)I",									 { &wrapper_static_JI, (generic_fn) &Java_java_util_zip_ZipFile_getCSize } },
	{ "java/util/zip/ZipFile" "."		"getEntry"				"(JLjava/lang/String;)J",				 { &wrapper_static_JLJ, (generic_fn) &Java_java_util_zip_ZipFile_getEntry } },
	{ "java/util/zip/ZipFile" "."		"getMethod"				"(J)I",									 { &wrapper_static_JI, (generic_fn) &Java_java_util_zip_ZipFile_getMethod } },
	{ "java/util/zip/ZipFile" "."		"getNextEntry"			"(JI)J",								 { &wrapper_static_JIJ, (generic_fn) &Java_java_util_zip_ZipFile_getNextEntry } },
	{ "java/util/zip/ZipFile" "."		"getSize"				"(J)I",									 { &wrapper_static_JI, (generic_fn) &Java_java_util_zip_ZipFile_getSize } },
	{ "java/util/zip/ZipFile" "."		"getTotal"				"(J)I",									 { &wrapper_static_JI, (generic_fn) &Java_java_util_zip_ZipFile_getTotal } },
	{ "java/util/zip/ZipFile" "."		"getZipMessage"			"(J)Ljava/lang/String;",				 { &wrapper_static_JL, (generic_fn) &Java_java_util_zip_ZipFile_getZipMessage } },
	{ "java/util/zip/ZipFile" "."		"initIDs"				"()V",									 { &wrapper_static_V, (generic_fn) &Java_java_util_zip_ZipFile_initIDs } },
	{ "java/util/zip/ZipFile" "."		"open"					"(Ljava/lang/String;IJ)J",				 { &wrapper_static_LIJJ, (generic_fn) &Java_java_util_zip_ZipFile_open } },
	{ "java/util/zip/ZipFile" "."		"read"					"(JJI[BII)I",							 { &wrapper_static_JJILIII, (generic_fn) &Java_java_util_zip_ZipFile_read } },
#endif /* JDK 1.4 */
#if JDK_VERSION >= 140
	{ "java/security/AccessController" "." "getStackAccessControlContext" "()Ljava/security/AccessControlContext;", { &wrapper_static_L, (generic_fn) &Java_java_security_AccessController_getStackAccessControlContext } },
	{ "java/security/AccessController" "." "doPrivileged"		"(Ljava/security/PrivilegedAction;)Ljava/lang/Object;", { &wrapper_static_LL, (generic_fn) &Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedAction_2 } },
	{ "java/security/AccessController" "." "doPrivileged"		"(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;", { &wrapper_static_LLL, (generic_fn) &Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedAction_2Ljava_security_AccessControlContext_2 } },
	{ "java/security/AccessController" "." "doPrivileged"		"(Ljava/security/PrivilegedExceptionAction;)Ljava/lang/Object;", { &wrapper_static_LL, (generic_fn) &Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedExceptionAction_2 } },
	{ "java/security/AccessController" "." "doPrivileged"		"(Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;", { &wrapper_static_LLL, (generic_fn) &Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedExceptionAction_2Ljava_security_AccessControlContext_2 } },
	{ "sun/misc/AtomicLong" "."			"VMSupportsCS8"			"()Z",									 { &sun_misc_AtomicLong_VMSupportsCS8, NULL } },
	{ "sun/misc/Signal" "."				"handle0"				"(IJ)J",								 { &wrapper_static_IJJ, (generic_fn) &Java_sun_misc_Signal_handle0 } },
	{ "sun/misc/Signal" "."				"findSignal"			"(Ljava/lang/String;)I",				 { &wrapper_static_LI, (generic_fn) &Java_sun_misc_Signal_findSignal } },
	{ "sun/misc/Unsafe" "."				"allocateMemory"		"(J)J",									 { &wrapper_JJ, (generic_fn) &Java_sun_misc_Unsafe_allocateMemory } },
	{ "sun/misc/Unsafe" "."				"ensureClassInitialized" "(Ljava/lang/Class;)V",				 { &wrapper_LV, (generic_fn) &Java_sun_misc_Unsafe_ensureClassInitialized } },
	{ "sun/misc/Unsafe" "."				"fieldOffset"			"(Ljava/lang/reflect/Field;)I",			 { &wrapper_LI, (generic_fn) &Java_sun_misc_Unsafe_fieldOffset } },
	{ "sun/misc/Unsafe" "."				"freeMemory"			"(J)V",									 { &wrapper_JV, (generic_fn) &Java_sun_misc_Unsafe_freeMemory } },
	{ "sun/misc/Unsafe" "."				"getByte"				"(J)B"	,								 { &wrapper_JB, (generic_fn) &Java_sun_misc_Unsafe_getByte } },
	{ "sun/misc/Unsafe" "."				"putLong"				"(JJ)V",								 { &wrapper_JJV, (generic_fn) &Java_sun_misc_Unsafe_putLong } },
	{ "sun/misc/Unsafe" "."				"getObject"				"(Ljava/lang/Object;I)Ljava/lang/Object;", { &wrapper_LIL, (generic_fn) &Java_sun_misc_Unsafe_getObject } },
	{ "sun/misc/Unsafe" "."				"registerNatives"		"()V",									 { &wrapper_static_V, (generic_fn) &Java_sun_misc_Unsafe_registerNatives } },
	{ "sun/misc/Unsafe" "."				"staticFieldBase"		"(Ljava/lang/Class;)Ljava/lang/Object;", { &wrapper_LL, (generic_fn) &Java_sun_misc_Unsafe_staticFieldBase } },
	{ "sun/reflect/NativeConstructorAccessorImpl" "." "newInstance0" "(Ljava/lang/reflect/Constructor;[Ljava/lang/Object;)Ljava/lang/Object;", { &wrapper_static_LLL, (generic_fn) &Java_sun_reflect_NativeConstructorAccessorImpl_newInstance0 } },
	{ "sun/reflect/Reflection" "."	    "getCallerClass"		"(I)Ljava/lang/Class;",					 { &wrapper_static_IL, (generic_fn) &Java_sun_reflect_Reflection_getCallerClass } },
	{ "sun/reflect/Reflection" "."		"getClassAccessFlags"	"(Ljava/lang/Class;)I",					 { &wrapper_static_LI, (generic_fn) &Java_sun_reflect_Reflection_getClassAccessFlags } },
#endif /* JDK 1.4 */
	{ NULL, { NULL, NULL } }
};


/**
 * Put all of the native methods from the array into a hashtable.
 */
void initNativeMethods()
{
	int n;

	nativesTable = Hashtable_new( 101 );

	for( n = 0; methods[n].sig; n++ )
		Hashtable_put( nativesTable, methods[n].sig, &methods[n].fn );
}


/**
 * Frees the native methods hashtable.
 */
void freeNativeMethods()
{
	Hashtable_delete( nativesTable, 0, 0, 0, NULL );
}


/**
 * Look up a given native method in the hashtable.
 */
NativeMethodStr* findNativeMethod( ClassFile *cf, u2 name_index, u2 descriptor_index )
{
	char key[255];
	char* buf = key;

	u2 class_name = ((CONSTANT_Class_info*) cf->constant_pool[cf->this_class])->name_index;

	buf += sprintRawSimpleName( buf, cf, class_name );
	*buf = '.'; buf++;
	buf += sprintRawSimpleName( buf, cf, name_index );
	buf += sprintRawSimpleName( buf, cf, descriptor_index );

	return Hashtable_get( nativesTable, key );
}
