/******************************************************************************
 *
 * Filename: util.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Class construction utilities.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <string.h>
#include <stdio.h>

#include "config.h"
#include "classfile.h"
#include "engine.h"
#include "loader.h"
#include "linker.h"
#include "disasm.h"
#include "memory.h"
#include "globals.h"
#include "openfiles.h"

#include "util.h"


#define CLASS_CLASSNAME "java/lang/Class"
#define STRING_CLASSNAME "java/lang/String"
#define THREAD_CLASSNAME "java/lang/Thread"
#define FILEDESC_CLASSNAME "java/io/FileDescriptor"
#define OBJSTRFIELD_CLASSNAME "java/io/ObjectStreamField"


extern void systemOutPrint( const char* buf, int len );



/**
 * Count the number of characters held in a UTF8 encoded string.
 *
 * if bit7     = 0,    single byte (\u0001 - \u007f)
 * if bits765  = 110,  two bytes, second byte bits76 = 10 (\u0080 - \u07ff)
 * if bits7651 = 1110, three bytes, second and third bits76 = 10 (\u0800 - \uffff)
 * mask 11000000 = c0  value 10000000 = 80
 * mask 11100000 = e0  value 11000000 = c0
 * mask 11110000 = f0  value 11100000 = e0
 */
u2 countUtf8( const u1* utf8, u2 utf8len )
{
	u2 count = 0;
	u2 n;

	for( n = 0; n < utf8len; n++ )
	{
		if( (utf8[n] & 0x80) == 0x00 ) { count++; continue; }	  /* ordinary byte */
		if( (utf8[n] & 0xc0) == 0x80 ) { continue; }			  /* continuation byte */
		if( (utf8[n] & 0xe0) == 0xc0 ) { count++; continue; }   /* two-byte start byte */
		if( (utf8[n] & 0xf0) == 0xe0 ) { count++; continue; }   /* three-byte start byte */

		printf( "countUtf8: Corrupt Utf8 constant at position %d\n", count );
		return count;
	}

	return count;
}

/* Decode a UTF8 encoded array of bytes into an array of Java chars */
void decodeUtf8( u2* chars, u2 arraylen, const u1* utf8, u2 utf8len )
{
	u2 count = 0;
	u2 n, value;
	u1 x, y, z;

	for( n = 0; n < utf8len; n++ )
	{
		if( (utf8[n] & 0x80) == 0x00 )	/* ordinary byte */
		{
			chars[count++] = utf8[n];
			continue;
		}
		else if( (utf8[n] & 0xe0) == 0xc0 )   /* two-byte start byte */
		{
			if( (n+1) < utf8len )
			{
				x = utf8[n];
				y = utf8[++n];
				value = ((x & 0x1f) << 6) + (y & 0x3f);
				chars[count++] = value;
				continue;
			}
		}
		else if( (utf8[n] & 0xf0) == 0xe0 )   /* three-byte start byte */
		{
			if( (n+2) < utf8len )
			{
				x = utf8[n];
				y = utf8[++n];
				z = utf8[++n];
				value = ((x & 0xf) << 12) + ((y & 0x3f) << 6) + (z & 0x3f);
				chars[count++] = value;
				continue;
			}
		}

		printf( "decodeUtf8: Corrupt Utf8 constant at position %d\n", count );
		return;
	}

	if( count > arraylen )
		printf( "WARNING: Buffer overrun by %d in decodeUtf8()\n", (count - arraylen - 1) );
}


/* Copy a C char array into a Java char array */
void c2wstrncpy( u2* dst, const char* src, int len )
{
	int n;
	for( n = 0; n < len; n++ )
		dst[n] = (u2) src[n];
}


/* Copy a Java char array into a C char array */
void w2cstrncpy( char* dst, const u2* src, int len )
{
	int n;
	for( n = 0; n < len; n++ )
	{
		if( src[n] < 256 ) {
			dst[n] = (char) src[n];
		} else {
			printf( "w2cstrncpy: Truncating Java char %04x\n", src[n] );
			dst[n] = (char) ( src[n] & 0x00ff );
		}
	}
}


/* Create a new stack frame in intialise it to zeros */
Frame* newStackFrame( method_info* mi, Frame* next )
{
	Frame *frame = NULL;

	if( !mi || !mi->ca || !mi->ca->code )
		return NULL;
	
	frame = javavm_malloc( sizeof( Frame ) + 5 * (mi->ca->max_stack + mi->ca->max_locals) );
	frame->cf = mi->cf;
	frame->mi = mi;
	frame->pc = (u1*) mi->ca->code;
	frame->min_pc = frame->pc;
	frame->max_pc = frame->pc + mi->ca->code_length;
	frame->local = (u4*) &frame->data; /* start of data */
	frame->sp = &frame->local[mi->ca->max_locals]; /* end of locals */
	frame->tlocal = (u1*) &frame->sp[mi->ca->max_stack]; /* end of stack */
	frame->tsp = &frame->tlocal[mi->ca->max_locals]; /* end of tlocals */
	frame->ex = NULL;
	frame->next = next;
	frame->prev = NULL;
	frame->init = NULL;
	frame->stopOnPop = 0;

	memset( frame->local, 0, 5 * (mi->ca->max_stack + mi->ca->max_locals) );

	return frame;
}


/* Free a stack frame created with newStackFrame() */
void freeStackFrame( Frame* fp, method_info* mi )
{
	javavm_free( fp, sizeof( Frame ) + 5 * (mi->ca->max_stack + mi->ca->max_locals) );
}


/* Structure for holding queued Class object creation requests */
typedef struct _DeferredClassCreate
{
	ClassFile* cf;
	ObjRef* classobj;
} DeferredClassCreate;

/* The queue itself */
DeferredClassCreate deferredv[8];
int deferredc = 0;

/**
 * Create an instance of java.lang.Class.  There is a little trick here in that
 * we create the object 4 bytes too big and use the last field (the 4 bytes) to
 * hold the ClassFile pointer (cf).  Note that because we have to have loaded
 * both Object and Class and resolved them both before we know object_size, we
 * must queue requests until we have done so.  This means that no Class object
 * will be created for either Object or Class until the next class is
 * successfully loaded.
 */
ObjRef* newJavaLangClass( ClassFile* cf )
{
	ClassFile* classcf;
	ObjRef* classobj;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	/* Make sure Class is loaded and get a handle on it */
    if( !(classcf = getClass( &stack, CLASS_CLASSNAME )) )
		return NULL; /* ###? exception */

	/* Create a new object reference */
	classobj = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
	classobj->type = 0;
	classobj->cf   = classcf;
	classobj->data = NULL;
	classobj->length = 1;

	/* If we haven't resolved Class yet, we don't know it's object_size, queue for later */
	if( !classcf->resolved )
	{
		if( verbose & 2 ) {
			char buf[255];
			sprintClassName( buf, cf, cf->this_class );
			printf( "WARNING: Deferring creation of Class object for %s\n", buf );
		}

		deferredv[deferredc].cf       = cf;
		deferredv[deferredc].classobj = classobj;
		deferredc++;
		return classobj;
	}

	/* Have now resolved Class, so create objects for everything in the queue */
	if( deferredc > 0 )
	{
		int n;
		for( n = 0; n < deferredc; n++ )
		{
			deferredv[n].classobj->data = (u4*) javavm_malloc( 4 * classcf->object_size + 4 );
			memset( deferredv[n].classobj->data, 0, 4 * classcf->object_size + 4 );
			deferredv[n].classobj->data[classcf->object_size] = (u4) deferredv[n].cf;
		}
		deferredc = 0;
	}

	/* Fill in the object part of the Class object */
	classobj->data = (u4*) javavm_malloc( 4 * classcf->object_size + 4 );
	memset( classobj->data, 0, 4 * classcf->object_size + 4 );
	classobj->data[classcf->object_size] = (u4) cf;  /* Cheat - this is how we associate a Class with a ClassFile! */

	return classobj;
}


/* Create an instance of java.lang.String from a Java char[] */
ObjRef* newJavaLangStringInternal( const ObjRef* value, int length )
{
	ClassFile* stringcf;
	javaLangString *data;
	Stack stack;
	ObjRef* objref;
	stack.ex = NULL;
	stack.fp = NULL;

	if( !(stringcf = getClass( &stack, STRING_CLASSNAME )) )
	{
	    if( verbose ) printf( "ERROR: newJavaLangString: class java.lang.String not loaded\n" );
		return NULL; /* ### exception */
    }

	objref = newObjectRefAndObject( stringcf );
	data = (javaLangString*) objref->data;
	data->value = (u4) value;
	data->offset = 0;
	data->count = length;
	return objref;
}

/* Create an instance of java.lang.String from a C string */
ObjRef* newJavaLangStringFromBytes( const char* bytes, int length )
{
	ObjRef* value = newObjectRefAndArray( T_CHAR, length );
	addToList( value, &arrayPool );
	c2wstrncpy( (u2*) value->data, bytes, length );
	return newJavaLangStringInternal( value, length );
}

/* Create an instance of java.lang.String from a Utf8 string */
ObjRef* newJavaLangStringFromUtf8( const CONSTANT_Utf8_info* utf8 )
{
	u2 charlen = countUtf8( utf8->utf8_bytes, (u2) utf8->utf8_length );
	ObjRef* value = newObjectRefAndArray( T_CHAR, charlen );
	addToList( value, &arrayPool );
	decodeUtf8( (u2*) value->data, charlen, utf8->utf8_bytes, (u2) utf8->utf8_length );
	return newJavaLangStringInternal( value, charlen );
}


/* Create an instance of java.lang.Thread */
ObjRef* newJavaLangThread( const char* threadName, ObjRef* group )
{
	ClassFile* threadcf;
	javaLangThread *data;
	ObjRef* name;
	ObjRef* objref;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	if( !(threadcf = getClass( &stack, THREAD_CLASSNAME )) )
	{
	    if( verbose ) printf( "ERROR: newJavaLangThread: class java.lang.Thread not loaded\n" );
		return NULL; /* ### exception */
    }

	name = newObjectRefAndArray( T_CHAR, strlen(threadName) );
	c2wstrncpy( (u2*) name->data, threadName, strlen(threadName) );

	objref = newObjectRefAndObject( threadcf );
	data = (javaLangThread*) objref->data;
	data->name = (u4) name;
	data->priority = 5; /* NORM */
	data->threadQ = (u4) NULL;
	data->PrivateInfo = 0;
	data->eetop = 0;
	data->single_step = 0; /* false */
	data->daemon = 0; /* false */
	data->stillborn = 0; /* false */
	data->target = (u4) NULL;
	data->group = (u4) group;
	data->initial_stack_memory = 0;
	return objref;
}


ObjRef* newJavaIoFileDescriptor( int fd )
{
	ClassFile* filedesccf;
	ObjRef* objref;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	if( !(filedesccf = getClass( &stack, FILEDESC_CLASSNAME )) )
		return NULL; /* ### exception */

	objref = newObjectRefAndObject( filedesccf );
	objref->data[0] = (u4) fd;

	return objref;
}


ObjRef* newJavaIoObjectStreamField( ObjRef* name, u1 type, u4 offset, ObjRef* typeString )
{
	ClassFile* objstrfieldcf;
	ObjRef* objref;
	javaIoObjectStreamField* data;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	if( !(objstrfieldcf = getClass( &stack, OBJSTRFIELD_CLASSNAME )) )
		return NULL; /* ### exception */

	objref = newObjectRefAndObject( objstrfieldcf );
	data = (javaIoObjectStreamField*) objref->data;
	data->name = (u4) name;
	data->type = type;
	data->offset = offset;
	data->typeString = (u4) typeString;

	return objref;
}


/* Creates a new object reference which points to a blank object of the given class */
ObjRef* newObjectRefAndObject( ClassFile* cf )
{
	ObjRef* objref = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
	objref->cf = cf;
	objref->type = 0;
	objref->data = (u4*) javavm_malloc( 4 * cf->object_size );
	objref->length = 1;
	memset( objref->data, 0, 4 * cf->object_size );
	return objref;
}


/**
 * Returns the size of a given type, or 0 if an invalid type has been passed.
 * Note: These are the sizes for array elements - boolean, char, byte and short
 * are all length 4 when they a pushed onto the stack.
 */
u4 sizeOfJavaType( u1 type )
{
	switch( type )
	{
	case T_OBJREF:	return sizeof(ObjRef*);
	case T_BOOLEAN: return 1;
	case T_CHAR:	return 2;
	case T_FLOAT:	return 4;
	case T_DOUBLE:	return 8;
	case T_BYTE:	return 1;
	case T_SHORT:	return 2;
	case T_INT:		return 4;
	case T_LONG:	return 8;
	default:
		printf( "ERROR: sizeOfJavaType: bad type (%d)\n", type );
		return 0;
	}
}


/**
 * Creates a new object reference which points to a blank array of the given
 * type.  Will return NULL if an invalid type is passed in.
 */
ObjRef* newObjectRefAndArray( u1 type, int32 length )
{
	ObjRef* objref;
	u4 size = sizeOfJavaType( type );

	/* If we are passed an invalid type, return NULL */
	if( !size )
		return NULL;

	/* Create the object reference */
	objref = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
	objref->cf = NULL;
	objref->type = type;
	objref->data = NULL;
	objref->length = length;

	/* Arrays of length 0 have a NULL data pointer */
	if( length > 0 )
	{
		/* Create the array */
		objref->data = javavm_malloc( size * length );
		memset( objref->data, 0, size * length );
	}

	return objref;
}


ClassFile* getClass( Stack *stp, const char* classname )
{
    ClassFile *cf;
    
    cf = checkLoaded( classname );
    
	if( !cf )
		cf = load( classname, stp );

	return cf;
}


int checkType( ClassFile* cf, const char* classname )
{
	ClassFile* checkcf;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	checkcf = getClass( &stack, classname );

	if( !checkcf )
		return 0;

	return (checkcf == cf);
}


char* extractFromJavaLangString( char* buf, int buflen, ObjRef* str )
{
	if( str->cf && checkType( str->cf, STRING_CLASSNAME ) && str->data && str->data[0] )
	{
		javaLangString* data = (javaLangString*) str->data;
		ObjRef* array = (ObjRef*) data->value;

		if( array && !array->cf	&& (array->type == T_CHAR) )
		{
			if( (data->offset <= array->length)
				&& ((data->offset + data->count) <= array->length)
				&& ((int) data->count < buflen) )
			{
				u2* src = (u2*) array->data;
				w2cstrncpy( buf, &src[data->offset], data->count );
				buf[data->count] = '\0';
				return buf;
			}
		}
	}

	return NULL;
}


int newUnsatisfiedLinkError( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintMethodRef( buf, cf, index );
	printf( "ERROR: No Java code or no native implementation for method %s\n", buf );
	
	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, UnsatisfiedLinkError, buf );
}


int newUnsatisfiedLinkErrorIface( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintInterfaceMethodRef( buf, cf, index );
	printf( "ERROR: No Java code or no native implementation for method %s\n", buf );
	
	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, UnsatisfiedLinkError, buf );
}


int newNoSuchMethodError( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintMethodRef( buf, cf, index );
	printf( "ERROR: Method %s not linked\n", buf );
	
	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, NoSuchMethodError, buf );
}


int newNoSuchMethodErrorIface( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintInterfaceMethodRef( buf, cf, index );
	printf( "ERROR: Method %s not linked\n", buf );
	
	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, NoSuchMethodError, buf );
}


int newNoSuchFieldError( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintFieldRef( buf, cf, index );

	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, NoSuchFieldError, buf );
}


const char* exceptionKindToString( ExceptionKind kind )
{
	switch( kind )
	{
	case VerifyError:            return "java/lang/VerifyError";
	case NoClassDefFoundError:   return "java/lang/NoClassDefFoundError";
	case IllegalAccessError:     return "java/lang/IllegalAccessError";
	case ClassFormatError:       return "java/lang/ClassFormatError";
	case NoSuchFieldError:       return "java/lang/NoSuchFieldError";
	case NoSuchMethodError:      return "java/lang/NoSuchMethodError";
	case InternalError:          return "java/lang/InternalError";
	case UnsatisfiedLinkError:	 return "java/lang/UnsatisfiedLinkError";
	case ExceptionInInitializerError: return "java/lang/ExceptionInInitializerError";

	case NullPointerException:   return "java/lang/NullPointerException";
	case ArrayIndexOutOfBoundsException: return "java/lang/ArrayIndexOutOfBoundsException";
	case ArithmeticException:    return "java/lang/ArithmeticException";
	case ClassCastException:     return "java/lang/ClassCastException";
	case ClassNotFoundException: return "java/lang/ClassNotFoundException";
	case IOException:            return "java/io/IOException";
	case NegativeArraySizeException: return "java/lang/NegativeArraySizeException";
	case CloneNotSupportedException: return "java/lang/CloneNotSupportedException";
	case ArrayStoreException: return "java/lang/ArrayStoreException";

#if JDK_VERSION < 140
	case NoSuchFieldException:	return "java/lang/NoSuchFieldException";
	case NoSuchMethodException:	return "java/lang/NoSuchMethodException";
#endif /* JDK 1.4 */

#if JDK_VERSION >= 140
	case PrivilegedActionException: return "java/security/PrivilegedActionException";
#endif /* JDK 1.4 */

	case JavaException:			 return NULL;
	}

	return NULL;
}


void whereAreWe( Frame* baseFrame )
{
	Frame* fp;
	char buffer[255];
	char *buf;
	int len;

	systemOutPrint( "Where are we:\n", 14 );

	if( !baseFrame )
	{
		systemOutPrint( "    null stack pointer!\n", 24 );
		return;
	}

	/* Add each frames */
	for( fp = baseFrame; fp; fp = (fp->next ? fp->next : fp->init) )
	{
		len = sprintf( buffer, "    at " );
		buf = buffer + len;

		if( !fp->mi->cf )
		{
			systemOutPrint( buffer, strlen( buffer ) );
			systemOutPrint( "\n", 1 );
			continue;
		}

		len = sprintClassName( buf, fp->mi->cf, fp->mi->cf->this_class ); buf += len;
		len = sprintf( buf, "." ); buf += len;

		if( CHECKCONSTANT( fp->mi->cf, fp->mi->name_index, CONSTANT_Utf8 ) )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->name_index ); buf += len;
		}

		len = sprintf( buf, "(" ); buf += len;

		if( fp->mi->cf->sfa )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->cf->sfa->sourcefile_index ); buf += len;
		}

		if( fp->mi->lnta )
		{
			int pc = fp->pc - fp->min_pc;
			int best = -1;
			int bestn = -1;
			int n;

			for( n = 0; n < fp->mi->lnta->line_number_table_length; n++ )
			{
				if( (pc >= fp->mi->lnta->line_number_table[n]->start_pc)
					&& (fp->mi->lnta->line_number_table[n]->start_pc > best) )
				{
					best = fp->mi->lnta->line_number_table[n]->start_pc;
					bestn = n;
				}
			}

			if( (bestn >= 0) && (bestn < fp->mi->lnta->line_number_table_length) )
			{
				len = sprintf( buf, ":%d", fp->mi->lnta->line_number_table[bestn]->line_number ); buf += len;
			}
		}
		
		/* extra position info giving exact pc of exception */
		len = sprintf( buf, "#%d)\n", (int) (fp->pc - fp->min_pc) ); buf += len;

		systemOutPrint( buffer, strlen( buffer ) );
	}
}


/**
 * Create a multi-dimensioned array (matrix).  The number of dimensions is
 * passed in 'dimensions' and actual dimensions are passed in the array
 * 'lengths'.  Type type of the leaf arrays is passed in 'type'.
 * e.g. dimensions = 2; lengths = [3,2] creates [[1,2][3,4][5,6]].
 */
ObjRef* createMultiObjrefArray( int dimensions, int32* lengths, u1 type )
{
	ObjRef *objref;
	int n;
	
	/*printf( "createMultiObjrefArray( dimension = %d, length[0] = %d, size = %d, type = %d )\n",
		dimension, length[0], size, type );*/

	objref = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );

	/*printf( "objref = 0x%08x\n", objref );*/

	objref->cf = NULL;
	objref->type = (dimensions == 1) ? type : T_OBJREF;
	objref->data = NULL;
	objref->length = lengths[0];

	/* ### Should be T_ARRAY and should check constant pool entry */

	if( lengths[0] > 0 )
	{
		if( dimensions == 1 )
		{
			int size = sizeOfJavaType( type );
			objref->data = javavm_malloc( size * lengths[0] );
			memset( objref->data, 0, size * lengths[0] );
			addToList( objref, &arrayPool );
		}
		else
		{
			ObjRef** elements;

			objref->data = javavm_malloc( sizeof( ObjRef* ) * lengths[0] );
			memset( objref->data, 0, sizeof( ObjRef* ) * lengths[0] );
			addToList( objref, &arrayPool );

			elements = (ObjRef**) objref->data;

			for( n = 0; n < lengths[0]; n++ )
				elements[n] = createMultiObjrefArray( dimensions-1, &lengths[1], type );
		}
	}

	return objref;
}


/* add an element to a object reference list */
void addToList( ObjRef* objref, ObjRefList** pool )
{
	ObjRefList* list = (ObjRefList*) javavm_malloc( sizeof( ObjRefList ) );
	list->obj = objref;
	list->next = *pool;
	*pool = list;
}


/* add a block of memory to the list of generic blocks to free */
void addMemoryBlock( void* block, size_t size )
{
	MemoryBlockList* list = (MemoryBlockList*) javavm_malloc( sizeof(MemoryBlockList) );
	list->block = block;
	list->block_size = size;
	list->next  = memoryBlockPool;
	memoryBlockPool = list;
}


int findFreeFD()
{
	int32 n;

	for( n = 3; (n < 255) && openfilesv[n]; n++ );

	if( n == 255 )
	{
		printf( "ERROR: Out of file descriptors\n" );
		return -1;
	}

	if( n >= openfilesc )
		openfilesc = n + 1;

	return (int) n;
}


/**
 * Lookup the offset of the start of a field in a Java object instance.  If the
 * field cannot be found, returns -1.
 */
int32 getFieldOffset( ClassFile* cf, const char* name, const char* desc )
{
	u2 n;

	/* Valid ClassFile? - We are sometimes called with the result of getClass(). */
	if( !cf )
		return -1;

	/* Look for a matching field name */
	for( n = 0; n < cf->fields_count; n++ )
	{
		if( cf->fields[n] && checkSimpleName( cf, cf->fields[n]->name_index, name ) )
			break;
	}

	/* Did we find anything? */
	if( n >= cf->fields_count )
		return -1;

	/* Verify that the types match */
	if( !checkSimpleName( cf, cf->fields[n]->descriptor_index, desc ) )
	{
		printf( "ERROR: getFieldOffset(): Type of field %s does not match %s\n", name, desc );
		return -1;
	}

	return cf->fields[n]->offset;
}
