/******************************************************************************
 *
 * Filename: zipload.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  9th April 2003
 *
 * Purpose:  ZIP file Class loader.
 *
 * (c) Copyright R.H.Thornburrow, 1998-2003
 *
 * Based on Paul Katz's public domain specification for the PKZIP file format.
 * No proprietry code has been referenced in developing this module.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#ifndef USES_UNZIPPED_CLASSES_ONLY


#include <stdio.h>
#include <string.h>

#include "config.h"
#include "classfile.h"
#include "loader.h"
#include "util.h"
#include "globals.h"
#include "zipfile.h"

#ifdef USE_ZIP_HASHTABLE
#include "hashtable.h"
#include "memory.h"
#include "jstring.h"
#endif /* USE_ZIP_HASHTABLE */

#include "zipload.h"


/**
 * Find a given file in the specified zipfile and move the file pointer to the
 * start of it.  If we are using caches, then if the cache hasn't been
 * populated, we populate it.  If we aren't using caches a (slow) sequential
 * search is used.  Returns 0 for success, 1 otherwise.
 */
int findZippedFile( FILE* f, Classpath_Entry* zipfile,
				    const char* filename, int* found, long* length )
{
	u4 zipsig;  /* Either ziplocsig or zipcensig */
	u4 zipofst; /* Relative offset of local header from the start of archive */
	int32 zipecenn, n;
	ZipCentralDir record;
	ZipLocalFile local;

#ifdef USE_ZIP_HASHTABLE

	ZipCacheEntry* cached;

	/* If hashtable has not been populated, populate it */
	if( !zipfile->cache )
	{
		/* (1a) Locate the Central Directory (first time only). */ 
		if( (zipecenn = findCentralDirectory( f )) < 0 )
			return 1;

		zipfile->cache = Hashtable_new( 101 );

		/* Allocate memory for all cache entries in one go for speed */
		zipfile->index = javavm_malloc( zipecenn * sizeof(ZipCacheEntry) );

		/* Remember the size of the index */
		zipfile->zipecenn = zipecenn;

		/* (2b) Start reading the Central Directory and cache it (first time only). */

		for( n = 0; n < zipecenn; n++ )
		{
			/* Read double word to check if we have a Central Directory */
			readDWord( &zipsig, f );

			if( zipsig != CENTRAL_DIR_MAGIC /* central dir */ )
			{
				printf( "ZIPLOAD: Not a valid ZIP file (Bad Central Directory magic).\n" );
				return 1;
			}

			/* Read the remainder of the structure */
			if( readCentralDirectory( &record, f ) != 0 )
				return 1;

			/* Fill in the index entry */
			zipfile->index[n].zipofst = record.zipofst;

			/* Add a new entry to the hashtable */
			Hashtable_put( zipfile->cache, strdup(record.zipcfn), &zipfile->index[n] );
		}

		if( verbose )
			printf( "[Opened %s]\n", zipfile->path );
	}

	/* (3a) Use the hashtable to look up the file we want. */

	/* Offset is stored directly in the hashtable */
	cached = (ZipCacheEntry*) Hashtable_get( zipfile->cache, filename );

	if( !cached )
	{
		/* We signal we didn't find the class by not setting found to true. */
		if( verbose ) {
			printf( "ZIPLOAD: Archive \"%s\" does not contain \"%s\".\n", zipfile->path, filename );
		}
		return 1;
	}

	zipofst = cached->zipofst;

#else /* USE_ZIP_HASHTABLE */

	/* (1b) Locate the Central Directory. */

	if( (zipecenn = findCentralDirectory( f )) < 0 )
		return 1;

	/* (2b) Start reading the Central Directory to find the file we want. */

	for( n = 0; n < zipecenn; n++ )
	{
		/* Read double word to check if we have a Central Directory */
		readDWord( &zipsig, f );

		if( zipsig != CENTRAL_DIR_MAGIC /* central dir */ )
		{
			printf( "ZIPLOAD: Not a valid ZIP file (Bad Central Directory magic).\n" );
			return 1;
		}

		/* Read the remainder of the structure */
		if( readCentralDirectory( &record, f ) != 0 )
			return 1;

		/* (3b) If file matches, we're finished, else keep looking */

		/* Is this the file we are looking for? */
		if( strcmp( filename, record.zipcfn ) == 0 )
		{
			/* Done searching so break out of loop */
			break;
		}
	}

	if( n >= zipecenn )
	{
		/* We signal we didn't find the class by not setting found to true. */
		if( verbose ) {
			printf( "ZIPLOAD: Archive \"%s\" does not contain \"%s\".\n", zipfile->path, filename );
		}
		return 1;
	}

	zipofst = record.zipofst;

#endif /* USE_ZIP_HASHTABLE */

	/* Signal fact that class was found */
	*found = 1;

	if( verbose & 2 )
	{
#ifdef USE_ZIP_HASHTABLE
		printf( "ZIPLOAD: Found class \"%s\" (offset %d)\n", filename, zipofst );
#else
		printf( "ZIPLOAD: Found class \"%s\" (File #%d; offset %d)\n", filename, n, zipofst );
#endif /* USE_ZIP_HASHTABLE */
	}

	/* (4) Find the Local File entry corresponding to the Central Dir entry. */

	/* Move to start of Local File entry */
	if( fseek( f, zipofst, SEEK_SET ) != 0 )
	{
		printf( "ZIPLOAD: Not a valid ZIP file (Can't move to start of Local File).\n" );
		return 1;
	}

	/* Read double word to check if we have a Local File */
	readDWord( &zipsig, f );

	if( zipsig != LOCAL_FILE_MAGIC /* local file */ )
	{
		printf( "ZIPLOAD: Not a valid ZIP file (Bad Local File magic (2)).\n" );
		return 1;
	}

	/* Read the remainder of the structure */
	if( readLocalFile( &local, f ) != 0 )
	{
		return 1;
	}

	/* Pass back length of file */
	*length = local.zipsize;

	/* Only handle uncompressed data */
	if( (local.zipmthd != zm_none) || (local.zipsize != local.zipuncmp) )
	{
		printf( "ZIPLOAD: Exploding of files not yet implemented!\n" );
		return 1;
	}

	return 0;
}


/**
 * Find a given file in the specified zipfile.  The classname is used for error
 * reporting only.  This is the wrapper which opens the zipfile.
 */
ClassFile* readZippedClassFile(
	Classpath_Entry* zipfile, const char* filename, const char* classname,
	Stack* stp, int* found )

{
	long length;
	ClassFile *cf;
	FILE *f;

	/* Open the zip file */
	if( !(f = fopen( zipfile->path, "rb" )) )
	{
		if( verbose ) {
			printf( "ZIPLOAD: Failed to open ZIP file \"%s\".\n", zipfile->path );
		}

		/* We signal we didn't find the class by not setting found to true. */
		return NULL;
	}

	/* Move file pointer to start of file within the zip file */
	if( findZippedFile( f, zipfile, filename, found, &length ) != 0 )
	{
		fclose( f );
		return NULL;
	}

	cf = readClassFileInternal( f, filename, classname, stp );

	fclose( f );

	/* Did load fail? Note that we ignore errors if ZIP file is corrupt! */
	if( (cf == NULL) && (*found == 1) )
	{
		/* Has an exception already been set */
		if( (stp->fp ? stp->fp->ex : stp->ex) == NULL )
		{
			/* Default exception is ClassFormatError */
			stp->retcode = newCompleteException( stp, ClassFormatError, classname );
		}

		return NULL;
	}

	if( verbose & 2 ) {
		printf( "ZIPLOAD: Finished loading %s from %s\n", classname, zipfile->path );
	}

	return cf;
}


/**
 * Find a given resource file in the specified zipfile and return a handle on
 * the file with the file pointer set to the start of the resource.
 */
FILE* fopenZippedResource( Classpath_Entry* zipfile, const char* filename, long* length )
{
	int found;
	FILE *f;

	/* Open the zip file */
	if( !(f = fopen( zipfile->path, "rb" )) )
	{
		if( verbose ) {
			printf( "ZIPLOAD: Failed to open ZIP file \"%s\".\n", zipfile->path );
		}
		return NULL;
	}

	/* Move file pointer to start of file within the zip file */
	if( findZippedFile( f, zipfile, filename, &found, length ) != 0 )
	{
		fclose( f );
		return NULL;
	}

	/* Just return the file pointer as we are now at the start of the file */
	return f;
}


#ifdef USE_ZIP_HASHTABLE
/**
 * Clean-up function for memory allocated to zip caches.
 */
void freeZipCaches()
{
	int n;
	for( n = 0; n < classpathc; n++ )
	{
		if( (classpathv[n].kind == cpt_Archive) && classpathv[n].cache )
		{
			/* Free the keys */
			Hashtable_delete( classpathv[n].cache, 0, 1 , 0, NULL );
			classpathv[n].cache = NULL;
			javavm_free( classpathv[n].index, classpathv[n].zipecenn * sizeof(ZipCacheEntry) );
			classpathv[n].index = NULL;
		}
	}
}
#endif /* USE_ZIP_HASHTABLE */

#endif /* USES_UNZIPPED_CLASSES_ONLY */
