/******************************************************************************
 *
 * Filename: main.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Main program for virtual machine.
 *
 * (c) Copyright R.H.Thornburrow, 1998-2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>    /* for file stat stuff */

#define CRTMEMCHECK 0

#ifdef WIN32

#include <sys/timeb.h>
#define STAT _stat
#define IS_DIRECTORY(buf)	((buf.st_mode & _S_IFDIR) ? 1 : 0)
#define IS_FILE(buf)		((buf.st_mode & _S_IFREG) ? 1 : 0)

#if CRTMEMCHECK
#include <crtdbg.h>
_CrtMemState crts1, crts2, crts3;
#endif

#else

#include <sys/time.h>
#define STAT stat
#define IS_DIRECTORY(buf)	((buf.st_mode & S_IFDIR) ? 1 : 0)
#define IS_FILE(buf)		((buf.st_mode & S_IFREG) ? 1 : 0)

#endif /* WIN32 */


#include <time.h>
#include <signal.h>

#include "../engine/config.h"

#ifdef PLATFORM_CYGWIN32
#include <sys/cygwin.h>
#ifndef MAX_PATH
#define MAX_PATH (260)
#endif /* MAX_PATH */
#endif /* CYGWIN */

#include "../engine/classfile.h"
#include "../engine/loader.h"
#include "../engine/zipload.h"
#include "../engine/linker.h"
#include "../engine/disasm.h"
#include "../engine/engine.h"
#include "../engine/native.h"
#include "../engine/util.h"
#include "../engine/instructions.h"
#include "../engine/memory.h"
#include "../engine/classpath.h"
#include "../engine/openfiles.h"
#include "../engine/globals.h"
#include "../engine/jstring.h"


int verbose = 0;

#ifdef _DEBUG
int method_debug = 0;
int field_debug = 0;
int arithmetic_debug = 0;
int instruction_count_dump = 0;
#endif /* _DEBUG */

#ifndef OPTIMIZED
int instruction_debug = 0;
int instruction_count = 0;
#endif /* OPTIMIZED */

#if JDK_VERSION >= 140
int assertions_enabled = 0;
#endif /* JDK 1.4 */

Stack stack;
ObjRef* threads[1];

#ifdef USE_ZIP_HASHTABLE
#define CACHE_NULL   , NULL
#else
#define CACHE_NULL
#endif /* USE_ZIP_HASHTABLE */

#ifdef USES_ABSOLUTE_FILENAMES
#define CP_ROOT ABSOLUTE_FILENAME_ROOT
#else
#define CP_ROOT "."
#endif /*USE_ZIP_HASHTABLE */

#ifdef USES_UNZIPPED_CLASSES_ONLY
/* Number of entries in classpath */
int classpathc = 1;
/* Classpath for platforms which do NOT support ZIP files */
Classpath_Entry classpathv[] = {
	{ CP_ROOT, cpt_Directory CACHE_NULL }
};
#else /* USES_UNZIPPED_CLASSES_ONLY */

#if USES_ENVIRONMENT_VARIABLES

/* Number of entries in classpath */
#if JDK_VERSION < 140
int classpathc = 3;
#else
int classpathc = 4;
#endif /* JDK 1.4 */

/* Classpath for platforms which support ZIP files and Environment Variables */
Classpath_Entry classpathv[] = {
	{ CP_ROOT, cpt_Directory, 0 CACHE_NULL },
	{ "$(TJVM_HOME)" FILE_SEPARATOR "lib", cpt_Unknown, 1 CACHE_NULL },
#if JDK_VERSION < 140
	{ "$(TJVM_HOME)" FILE_SEPARATOR "lib" FILE_SEPARATOR "classes.zip", cpt_Unknown, 1 CACHE_NULL }
#else
	{ "$(TJVM_HOME)" FILE_SEPARATOR "lib" FILE_SEPARATOR "rt.jar", cpt_Unknown, 1 CACHE_NULL },
	{ "$(TJVM_HOME)" FILE_SEPARATOR "lib" FILE_SEPARATOR "tools.jar", cpt_Unknown, 1 CACHE_NULL }
#endif /* JDK 1.4 */
};

#else /* USES_ENVIRONMENT_VARIABLES */

/* Number of entries in classpath */
#if JDK_VERSION < 140
int classpathc = 2;
#else
int classpathc = 3;
#endif /* JDK 1.4 */

/* Classpath for platforms which support ZIP files, but NOT Environment variables */
Classpath_Entry classpathv[] = {
	{ CP_ROOT, cpt_Directory, 0 CACHE_NULL },
#if JDK_VERSION < 140
	{ ABSOLUTE_FILENAME_ROOT "classes.zip", cpt_Unknown, 1 CACHE_NULL }
#else
	{ ABSOLUTE_FILENAME_ROOT "rt.jar", cpt_Unknown, 1 CACHE_NULL },
	{ ABSOLUTE_FILENAME_ROOT "tools.jar", cpt_Unknown, 1 CACHE_NULL }
#endif /* JDK 1.4 */
};
#endif /* USES_ENVIRONMENT_VARIABLES */
#endif /* USES_UNZIPPED_CLASSES_ONLY */

#if JDK_VERSION >= 140
extern void closeAllJavaUtilZipZipFiles();
#endif /* JDK 1.4 */

/* fwd ref */
void dumpStack( Stack* stp );


#ifdef SUPPORTS_SIGNALS
/**
 * Signal handler.  Called in response to a segmentation fault.  Dumps the Java
 * stack using dumpStack() to give an idea of what went wrong.
 */
void signal_handler( int sig )
{
	printf( "SIGSEGV detected - full thread dump:\n" );
	printf( "    \"main\":\n" );

	dumpStack( &stack );

	exit( 0 );
}
#endif /* SUPPORTS_SIGNALS */


/**
 * Method for dumping the Java stack.  Used by the signal handler and the
 * unhandled exception trap (called when Java generates an exception and
 * doesn't catch it itself.
 */
void dumpStack( Stack* stp )
{
	Frame *fp;
	char buffer[255];
	char *buf;
	int len;

	/* Add each stack frame */
	for( fp = stp->fp; fp; fp = (fp->next ? fp->next : fp->init) )
	{
		len = sprintf( buffer, "        at " ); buf = buffer + len;

		if( !fp->mi->cf )
		{
			printf( "%s\n", buffer );
			continue;
		}

		len = sprintClassName( buf, fp->mi->cf, fp->mi->cf->this_class ); buf += len;
		len = sprintf( buf, "." ); buf += len;

		if( CHECKCONSTANT( fp->mi->cf, fp->mi->name_index, CONSTANT_Utf8 ) )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->name_index ); buf += len;
		}

		len = sprintf( buf, "(" ); buf += len;

		if( fp->mi->cf->sfa )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->cf->sfa->sourcefile_index ); buf += len;
		}

		if( fp->mi->lnta )
		{
			int pc = fp->pc - fp->min_pc;
			int best = -1;
			int bestn = -1;
			int n;

			for( n = 0; n < fp->mi->lnta->line_number_table_length; n++ )
			{
				if( (pc >= fp->mi->lnta->line_number_table[n]->start_pc)
					&& (fp->mi->lnta->line_number_table[n]->start_pc > best) )
				{
					best = fp->mi->lnta->line_number_table[n]->start_pc;
					bestn = n;
				}
			}

			if( (bestn >= 0) && (bestn < fp->mi->lnta->line_number_table_length) )
			{
				len = sprintf( buf, ":%d", fp->mi->lnta->line_number_table[bestn]->line_number ); buf += len;
			}
		}
		
		/* extra position info giving exact pc of exception */
		len = sprintf( buf, "#%d", (int) (fp->pc - fp->min_pc) ); buf += len;

		printf( "%s)\n", buffer );
	}
}


/**
 * Report a Java exception to the user.  If it is a Java exception, then print
 * the class name and any message, followed by the stack trace held in the
 * exception object, otherwise print its kind and message and call dumpStack().
 */
void reportException( Stack* stp )
{
	Exception *ex;

	if( stp->fp && stp->fp->ex )
		ex = stp->fp->ex;
	else
		ex = stp->ex;

	printf( "Unhandled exception caught by JavaVM:\n" );

	if( !ex )
	{
		printf( "\tUnknown Exception\n" );
		return;
	}

	if( ex->kind == JavaException )
	{
		/* Make sure this is a valid object */
		if( ex->ex && ex->ex->cf )
		{
			char buf[1024];
			sprintClassName( buf, ex->ex->cf, ex->ex->cf->this_class );
			printf( "%s", buf );

			/* Is there a message associated with the Throwable (Exception) */
			if( ex->ex->data && ex->ex->data[1] )
			{
				ObjRef* string = (ObjRef*) ex->ex->data[1];
				char mess[1024];

				/* Extract the message and print it out */
				if( extractFromJavaLangString( mess, 1024, string ) )
					printf( ": %s", mess );
				else
					printf( ": <message corrupt!>" );
			}

			printf( "\n" );

			/* stack trace */
			if( ex->ex->data && ex->ex->data[0] )
#if JDK_VERSION < 140
			{
				printf( "%s", (char*) ex->ex->data[0] );
			}
#else
			{
				StackTrace *st = (StackTrace*) ex->ex->data[0];
				StackTrace_Entry *ste;
				char buffer[512];
				int n;

				for( n = 0; n < st->entriesc; n++ )
				{
					char *buf = buffer;

					/* Get the stack trace entry */
					ste = &st->entriesv[n];

					buf += sprintf( buf, "\tat " );

					/* Class Name */
					buf += sprintClassName( buf, ste->cf, ste->cf->this_class );
					buf += sprintf( buf, "." );

					/* Method Name */
					buf += sprintSimpleName( buf, ste->mi->cf, ste->mi->name_index );
					buf += sprintf( buf, "(" );

					/* Do we have a Source File? */
					if( ste->mi->cf->sfa )
						buf += sprintSimpleName( buf, ste->mi->cf, ste->mi->cf->sfa->sourcefile_index );

					/* Do we have a Line Number? */
					if( ste->linenum >= 0 )
						buf += sprintf( buf, ":%d", ste->linenum );

					/* extra position info giving exact pc of exception */
					if( verbose )
						buf += sprintf( buf, "#%d", ste->pc );

					buf += sprintf( buf, ")\n" );

					/* Output the line */
					printf( buffer );
				}
			}
#endif /* JDK 1.4 */
		}
		else
		{
			printf( "\tUnknown Java Exception\n" );
			dumpStack( stp );
		}

		return;
	}

	if( exceptionKindToString( ex->kind ) )
		printf( "\t%s", exceptionKindToString( ex->kind ) );
	else
		printf( "\tUnknown Exception" );

	if( ex->message )
		printf( ": %s", ex->message );

	printf( "\n" );
	dumpStack( stp );
}


/**
 * Expands any environment variables in a pathname.  Uses functions in jstring
 * module.
 */
const char* expandPath( const char* path )
{
	char expanded[255];
	int len, dollar, clparen;
	char *buf = expanded;
	int prev = 0;

	if( (dollar = indexOf(path, '$', 0)) == -1 )
		return path;

	len = strlen( path );

	for( ; dollar != -1; dollar = indexOf(path, '$', prev) )
	{
		/* If some path between potential vars we copy it into the buffer */
		if( dollar - prev > 1 ) {
			strncpy( buf, &path[prev], dollar - prev ); buf += (dollar - prev);
		}

		/* Must have room for $(), ( must be next after dollar, and must have closing ) */
		if( dollar < (len - 3)
			&& (path[dollar+1] == '(')
			&& ((clparen = indexOf( path, ')', dollar )) != -1) )
		{
			char *value, varname[255];

			/* Extract the variable name */
			strncpy( varname, &path[dollar+2], clparen - dollar - 2 );
			varname[clparen - dollar - 2] = '\0';

			/* Lookup the value and copy it into the buffer */
			if( (value = getenv( varname )) != NULL ) {
#ifdef PLATFORM_CYGWIN32
				char posix[MAX_PATH];
				/* Under Cygwin all env vars must converted to POSIX paths.
				   Note: This has no effect if they are already POSIX paths. */
				cygwin_conv_to_posix_path( value, posix );
				value = posix;
#endif /* CYGWIN */
				strcpy( buf, value ); buf += strlen( value );
			} else {
				printf( "WARNING: Environment variable \"%s\" in classpath is unset.\n", varname );
			}

			prev = clparen + 1;
		}
		else
		{
			/* Copy the $ as it is not part of an env var */
			strncpy( buf, "$", 1 ); buf += 1;

			/* Force copy from this non-var $ to the next */
			prev = dollar + 1;
		}
	}

	/* Copy last part of path to buffer (dollar is -1) - also copies '\0' */
	strcpy( buf, &path[prev] );

	return strdup( expanded );
}


/**
 * Validate the contents of the classpathv array.  If the type of an entry is
 * not yet known, we use stat() to determine its type.
 */
void validateClasspath()
{
#ifndef USES_UNZIPPED_CLASSES_ONLY
	struct STAT buf;
	int n;

	for( n = 0; n < classpathc; n++ )
	{
		if( classpathv[n].kind == cpt_Unknown )
		{
			/* Expand any environment variables */
			classpathv[n].path = expandPath( classpathv[n].path );

			/* Check whether entry is a directory or a file (archive) */
			if( STAT( classpathv[n].path, &buf ) == 0 )
			{
				if( IS_DIRECTORY(buf) )
					classpathv[n].kind = cpt_Directory;
				else if( IS_FILE(buf) )
					classpathv[n].kind = cpt_Archive;

				if( verbose )
					printf( "Added \"%s\" to classpath - kind is %s\n", classpathv[n].path,
						((classpathv[n].kind == cpt_Directory) ? "directory"
						: ((classpathv[n].kind = cpt_Archive) ? "archive" : "unknown")) );
			}
			else
			{
				classpathv[n].kind = cpt_Invalid;

				if( verbose )
					printf( "Removed \"%s\" from classpath - does not exist!\n", classpathv[n].path );
			}
		} /* if */
	} /* for */
#endif /* USES_UNZIPPED_CLASSES_ONLY */
}


/**
 * Initialize the VM ready for use.  This involves the following steps:
 * (1) Load java.lang.System and execute System.initializeSystemClass()
 * (2) Load java.lang.ThreadGroup and create and instance using the default
 *     constructor
 * (3) Load java.lang.Thread and create and instance called "main"
 */
int vmInit()
{
	ClassFile *systemClass, *threadClass, *threadGroupClass;
	ObjRef* systemThreadGroup = NULL;
	int n, retcode = 0;

	/* init file descriptor list to 0 */
	for( n = 0; n < 255; n++ )
		openfilesv[n] = 0;

	/* init primitive class objects to null */
	for( n = 0; n < T_LAST; n++ )
		primitiveClass[n] = NULL;


	/* create "system" ThreadGroup object */
	threadGroupClass = load( "java/lang/ThreadGroup", &stack );

	if( !stack.ex && threadGroupClass )
	{
		/* create new ThreadGroup object */
		systemThreadGroup = newObjectRefAndObject( threadGroupClass );

		/* call constructor ThreadGroup.<init>() */
		retcode = forceExecuteJavaMethod( systemThreadGroup, &stack, "<init>", "()V" );

		while( !retcode )
			retcode = executeInstruction( &stack );
	}

	if( !threadGroupClass || !systemThreadGroup || stack.ex || (retcode == 2) )
	{
		reportException( &stack );
		return 2;
	}


	/* Create "main" thread object */
	threadClass = load( "java/lang/Thread", &stack );

	if( !stack.ex && threadClass )
	{
		/* create a new Thread object */
		threads[0] = newJavaLangThread( "main", systemThreadGroup );
	}

	if( !threadClass || stack.ex || (retcode == 2) )
	{
		reportException( &stack );
		return 2;
	}


	/* load and init java.lang.System */
	systemClass = load( "java/lang/System", &stack );

	if( !stack.ex && systemClass )
	{
		retcode = forceExecuteStaticJavaMethod( systemClass, &stack, "initializeSystemClass", "()V" );

		while( !retcode )
			retcode = executeInstruction( &stack );
	}

	if( !systemClass || stack.ex || (retcode == 2) )
	{
		reportException( &stack );
		return 2;
	}

	return 0;
}


/**
 * Display the version of Taurus JVM we are running.
 */
void showVersion()
{
	printf( "Taurus Java Virtual Machine %s v%s %s\n", JAVAVM_PLATFORM, JAVAVM_VERSION, JAVAVM_RELEASE );
#ifdef OPTIMIZED
	printf( "Optimized build. " );
#endif
	printf( "(C) Copyright 1998-2003, Taurus Software\n" );
}


/**
 * Display the version of Java that the VM is built for
 */
void showJavaVersion()
{
	printf( "java version \"" JAVA_SPEC_VERSION "\"\n" );
}


/**
 * Display a usage message with an optional error message.  Uses showVersion()
 * above.
 */
void showUsage( const char* msg )
{
	showVersion();

	if( msg )
		printf( "\n%s\n", msg );

	printf( "\n" );
	printf( "usage: javavm [-options] <class>\n\n" );
	printf( "where options include:\n" );
	printf( "    -[h]elp             this message\n" );
#ifdef _DEBUG
	printf( "    -[d]isasm <level>   debug mode\n" );
#endif
#ifndef OPTIMIZED
	printf( "    -[i]trace           instruction trace\n" );
#endif
#ifdef _DEBUG
	printf( "    -[m]trace           method trace\n" );
	printf( "    -[f]trace           field trace\n" );
	printf( "    -[c]ount            instruction count dump\n" );
	printf( "    -[a]trace           arithmetic trace\n" );
#endif
#if JDK_VERSION >= 140
	printf( "    -ea:<class> or -enableassertions:<class>\n" );
#endif /* JDK 1.4 */
	printf( "    -[v]erbose          verbose mode\n" );
	printf( "    -version            prints version number\n" );
}


/**
 * The main program.  It does the following:
 * (1) Validates the arguments
 * (2) if debug build and -(d)isasm specified dump the class using
 *     dissassemble() in disasm.c
 * (3) Initialise the VM using vmInit()
 * (4) If ok, load the class specified on the command line
 * (5) If ok, call execute() in engine.c
 * (6) Optionally show statistics
 * (7) Free any memory we used
 */
int main( int argc, char* argv[] )
{
#ifdef _DEBUG
	int level = 0;
#endif
	ClassFile *ptr = NULL;
	int retcode = 0;
#ifdef WIN32
	struct _timeb startbuf, endbuf;
#else
	/*struct timeb startbuf, endbuf;*/
	struct timeval startbuf, endbuf;
#endif

#if CRTMEMCHECK
    _CrtSetReportMode( _CRT_WARN, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_WARN, _CRTDBG_FILE_STDOUT );
	_CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_ERROR, _CRTDBG_FILE_STDOUT );
	_CrtSetReportMode( _CRT_ASSERT, _CRTDBG_MODE_FILE );
	_CrtSetReportFile( _CRT_ASSERT, _CRTDBG_FILE_STDOUT );

   _CrtMemCheckpoint( &crts1 );
#endif

	if( argc <= 1 )
	{
		showUsage( "Please specify a class file." );
		return 1;
	}

	argv++; argc--;
	while( *argv && ((*argv)[0] == '-') )
	{
		switch( (*argv)[1] )
		{
#ifdef _DEBUG
		case 'a':
			if( (strcmp( *argv, "-a" ) == 0) || (strcmp( *argv, "-atrace" ) == 0) ) {
				arithmetic_debug = 1;
			} else {
				showUsage( "Invalid option." );
				return 1;
			}
			break;
		case 'c':
			if( (strcmp( *argv, "-c" ) == 0) || (strcmp( *argv, "-count" ) == 0) ) {
				instruction_count_dump = 1;
			} else
			{
				showUsage( "Invalid option." );
				return 1;
			}
			break;
		case 'd':
			if( (strcmp( *argv, "-d" ) == 0)
				|| (strcmp( *argv, "-disasm" ) == 0) )
			{
				argv++; argc--;

				if( *argv && (*argv[0] != '-') )
				{
					level = atoi( *argv );
				}
				else
				{
					printf( "usage: javavm [-disasm <level>] <class>\n\n" );
					printf( "where level is a combination of:\n" );
					printf( "    1 = method code\n" );
					printf( "    2 = constant pool table\n" );
					printf( "    4 = class attributes\n" );
					printf( "    8 = method attributes\n" );
					return 1;
				}
			}
			else
			{
				showUsage( "Invalid option." );
				return 1;
			}
			break;
#endif /* _DEBUG */
#if JDK_VERSION >= 140
		case 'e':
			if( (strncmp( *argv, "-ea:", 3 ) == 0)
				|| strncmp( *argv, "-enableassertions:", 18 ) == 0)
			{
				/* ### Should really do something with the class name after the ':' */
				assertions_enabled = 1;
			}
			else
			{
				showUsage( "Invalid option." );
				return 1;
			}
			break;
#endif /* JDK 1.4 */
#ifndef OPTIMIZED
		case 'i':
			if( (strcmp( *argv, "-i" ) == 0) || (strcmp( *argv, "-itrace" ) == 0) ) {
				instruction_debug = 1;
			} else {
				showUsage( "Invalid option." );
				return 1;
			}
			break;
#endif /* OPTIMIZED */
#ifdef _DEBUG
		case 'f':
			if( (strcmp( *argv, "-f" ) == 0) || (strcmp( *argv, "-ftrace" ) == 0) ) {
				field_debug = 1;
			} else {
				showUsage( "Invalid option." );
				return 1;
			}
			break;
		case 'm':
			if( (strcmp( *argv, "-m" ) == 0) || (strcmp( *argv, "-mtrace" ) == 0) ) {
				method_debug = 1;
			} else {
				showUsage( "Invalid option." );
				return 1;
			}
			break;
#endif /* _DEBUG */
		case 'v':
			if( (strcmp( *argv, "-v" ) == 0) || (strcmp( *argv, "-verbose" ) == 0) ) {
				verbose++;
			} else if( strcmp( *argv, "-version" ) == 0 ) {
				showJavaVersion();
				showVersion();
				return 0;
			} else {
				showUsage( "Invalid option." );
				return 1;
			}
			break;
		case 'h':
			showUsage( NULL );
			return 0;
		default:
			showUsage( "Invalid option." );
			return 1;
		}
		argv++; argc--;
	}

	if( !*argv )
	{
		showUsage( "Please specify a class file." );
		return 1;
	}

	validateClasspath();

	initNativeMethods();

#ifdef _DEBUG
	if( level > 0 )
	{
		char zipfname[255], hostfname[255];
		char *zbuf = zipfname, *hbuf = hostfname;
		Stack stack;
		int n, found = 0;

		/** Create an host dependant filename: Add ".class" and replace '/'s and '$'s */

		/* Copy class name */
		strcpy( hbuf, *argv ); hbuf += strlen( *argv );
#if USES_FILE_EXTENSIONS
		/* Append ".class" onto filename to give host filename */
		strcpy( hbuf, CLASS_FILE_EXTN ); hbuf += strlen( CLASS_FILE_EXTN );
#endif /* USES_FILE_EXTENSIONS */
		*hbuf = '\0';
	
#ifdef USES_BACKSLASHED_FILENAMES
		/* Replace '/'s with '\'s */
		for( hbuf = hostfname; *hbuf; hbuf++ )
		{
			if( *hbuf == '/' ) *hbuf = '\\';
		}
#endif /* USES_BACKSLASHED_FILENAMES */
	
#ifdef USES_NONDOLLAR_FILENAMES
		/* Replace '$'s with '_'s */
		for( hbuf = hostfname; *hbuf; hbuf++ )
		{
			if( *hbuf == '$' ) *hbuf = '_';
		}
#endif /* USES_NONDOLLAR_FILENAMES */

		/** Create a zip format filename: Add ".class" and leave '/'s alone */

		/* Copy class name */
		strcpy( zbuf, *argv ); zbuf += strlen( *argv );
#if USES_ZIP_FILE_EXTENSIONS
		/* Append ".class" onto filename to give zip filename */
		strcpy( zbuf, ZIP_CLASS_FILE_EXTN ); zbuf += strlen( ZIP_CLASS_FILE_EXTN );
#endif /* USES_ZIP_FILE_EXTENSIONS */
		*zbuf = '\0';

		stack.retcode = 0;

		/* Implementation of CLASSPATH */
		for( n = 0; !found && (n < classpathc); n++ )
		{
			ptr = NULL;

			if( classpathv[n].kind == cpt_Directory )
			{
				/* Prepend classpath entry onto host filename */
				char prehfname[512];
				char* pbuf = prehfname;
				strcpy( pbuf, classpathv[n].path ); pbuf += strlen( classpathv[n].path );
				strcpy( pbuf, FILE_SEPARATOR ); pbuf += strlen( FILE_SEPARATOR );
				strcpy( pbuf, hostfname ); pbuf += strlen( hostfname );
				*pbuf = '\0';

				/* Try normal class loader - using prepended host filename  */
				ptr = readClassFile( prehfname, *argv, &stack, &found );
			}
#ifndef USES_UNZIPPED_CLASSES_ONLY
			else if( classpathv[n].kind == cpt_Archive )
			{
				/* Try ZIP loader - using raw filename */
				ptr = readZippedClassFile( &classpathv[n], zipfname, *argv, &stack, &found );
			}
#endif /* USES_UNZIPPED_CLASSES_ONLY */

			if( !ptr && found )
				break; /* load failed - exception will be on stack */
		}
		/* End of CLASSPATH implementation */

		if( ptr && found )
			disassemble( ptr, level );
		else
			printf( "Unable to load class \"%s\"\n", *argv );

		/* Free any memory we might have used */
		freeClassFiles();
		freeNativeMethods();

#ifdef USE_ZIP_HASHTABLE
		freeZipCaches();
#endif /* USE_ZIP_HASHTABLE */

		printf( "\nPress return to exit.\n" );
		fgetc( stdin );
		return 0;
	}
#endif /* _DEBUG */

#ifdef WIN32
	_ftime( &startbuf );
#else
	gettimeofday( &startbuf, NULL );
#endif

#ifdef SUPPORTS_SIGNALS
	if( signal( SIGSEGV, signal_handler ) == SIG_ERR )
	{
		printf( "Failed to attach signal handler\n" );
		return 1;
	}
#endif /* SUPPORTS_SIGNALS */


	stack.ex = NULL;
	stack.fp = NULL;
	stack.tail = NULL;

	if( vmInit() == 2 )
		return -1;

	if( !stack.ex )
		ptr = load( *argv, &stack );

	if( !ptr )
	{
		reportException( &stack );
		return -1;
	}

	argv++; argc--;
	retcode = execute( ptr, &stack, argc, argv );

#ifdef WIN32
	_ftime( &endbuf );
#else
	gettimeofday( &endbuf, NULL );
#endif

#ifndef OPTIMIZED
	if( 1 )
	{
		double elapsed, ips;
#ifdef WIN32
		double starttime = startbuf.time;
		double endtime = endbuf.time;

		starttime *= 1000;
		starttime += startbuf.millitm;
		endtime *= 1000;
		endtime += endbuf.millitm;
		elapsed = (endtime - starttime) / 1000;
#else
		double starttime = startbuf.tv_sec;
		double endtime = endbuf.tv_sec;

		starttime *= 1000000;
		starttime += startbuf.tv_usec;
		endtime *= 1000000;
		endtime += endbuf.tv_usec;
		elapsed = (endtime - starttime) / 1000000;
#endif /* WIN32 */
		ips = instruction_count / elapsed;
		printf( "Executed %d instructions in %.3f seconds (%.2f instructions/sec)\n", instruction_count, elapsed, ips );
	}
#endif /* OPTMIZED */

	if( verbose > 1 )
	{
		if( objectPool )
		{
			ObjRefList *item, *prev, *next;
			ClassFile* type;
			int objCount = 0;

			for( item = objectPool; item; item = item->next )
				objCount++;

			printf( "Allocated total of %d objects:\n", objCount );

			while( objectPool )
			{
				type = objectPool->obj->cf;
				objCount = 1;

				prev = objectPool;
				item = objectPool->next;

				while( item )
				{
					if( item->obj->cf == type )
					{
						objCount++;
						prev->next = item->next;
						javavm_free( item, sizeof(ObjRefList) );
						item = prev->next;
						continue;
					}
					
					prev = item;
					item = item->next;
				} /* item */

				printf( "\t%d %s\n", objCount, type->name );

				next = objectPool->next;
				javavm_free( objectPool, sizeof(ObjRefList) );
				objectPool = next;
			} /* objectPool */
		} /* objectPool */

		if( arrayPool )
		{
			ObjRefList *item, *prev, *next;
			ClassFile* ctype;
			int btype;
			int objCount = 0;

			for( item = arrayPool; item; item = item->next )
				objCount++;

			printf( "Allocated total of %d arrays:\n", objCount );

			while( arrayPool )
			{
				ctype = arrayPool->obj->cf;
				btype = arrayPool->obj->type;
				objCount = 1;

				prev = arrayPool;
				item = arrayPool->next;

				while( item )
				{
					if( (item->obj->cf == ctype) && (item->obj->type == btype) )
					{
						objCount++;
						prev->next = item->next;
						javavm_free( item, sizeof(ObjRefList) );
						item = prev->next;
						continue;
					}
					
					prev = item;
					item = item->next;
				} // item

				if( ctype )
				{
					printf( "\t%d %s[]\n", objCount, ctype->name );
				}
				else
				{
					switch( btype )
					{
					case T_BOOLEAN: printf( "\t%d boolean[]\n", objCount ); break;
					case T_CHAR:	printf( "\t%d char[]\n",    objCount ); break;
					case T_FLOAT:	printf( "\t%d float[]\n",   objCount ); break;
					case T_DOUBLE:	printf( "\t%d double[]\n",  objCount ); break;
					case T_BYTE:	printf( "\t%d byte[]\n",    objCount ); break;
					case T_SHORT:	printf( "\t%d short[]\n",   objCount ); break;
					case T_INT:		printf( "\t%d int[]\n",     objCount); break;
					case T_LONG:	printf( "\t%d long[]\n",    objCount ); break;
					}
				}

				next = arrayPool->next;
				javavm_free( arrayPool, sizeof(ObjRefList) );
				arrayPool = next;
			} /* while */
		} /* if */
	} /* verbose */

#ifdef _DEBUG
	if( instruction_count_dump )
	{
		int inst;

		printf( "Instruction counts:\n" );

		for( inst = 0; inst < 256; inst++ )
		{
			if( (instructionTable[inst].name != NULL)
				&& (instructionTable[inst].count == 0)
				/*(instructionTable[inst].count > 0)
				&& (instructionTable[inst].count <= 5)*/ )
				printf( "\t%s\t\t%d\n", instructionTable[inst].name, ((int) instructionTable[inst].count) );
		}
	}
#endif

	if( retcode == 2 /* exception */ )
		reportException( &stack );

#if JDK_VERSION >= 140
	closeAllJavaUtilZipZipFiles();
#endif /* JDK 1.4 */

	freeObjectPool();
	freeArrayPool();
	freeMemoryBlockPool();
	freeClassFiles();
	freeNativeMethods();

#ifdef USE_ZIP_HASHTABLE
	freeZipCaches();
#endif /* USE_ZIP_HASHTABLE */

#ifndef OPTIMIZED
	if( 1 )
	{
		printf( "Memory total %d kbytes, peak %d kbytes, used %d kbytes\n",
			(int) (bytes_total / 1024), (int) (bytes_peak / 1024), (int) (bytes_used / 1024) );
	}
#endif /* OPTIMIZED */

	/*fgetc( stdin );*/

#if CRTMEMCHECK
	_CrtMemCheckpoint( &crts2 );
	if ( _CrtMemDifference( &crts3, &crts1, &crts2 ) )
		_CrtMemDumpStatistics( &crts3 );
	_CrtCheckMemory();
#endif

	return ((retcode == 2) ? -1 : 0);
}
