/******************************************************************************
 *
 * Filename: java_io_FileInputStream.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for java.io.FileInputStream.
 *
 * (c) Copyright R.H.Thornburrow, 2002-3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#ifdef WIN32
#include <conio.h>
#define GETCHAR()  _getch()
#define PUTCHAR(x) _putch(x)
#else
#include <errno.h>
#define GETCHAR()  getchar()
#define PUTCHAR(x) ;
#endif

#include "../engine/config.h"

#include "java_io_FileInputStream.h"

#include "../engine/util.h"
#include "../engine/openfiles.h"
#include "../engine/memory.h"
#include "../engine/globals.h"

#include "java_io_FileDescriptor.h"

 
/******************************************************************************
 *
 * class java.io.FileInputStream
 *
 *
 * The system dependent file descriptor. The value 'fd' is an index into
 * openfilesv.  A value of NULL for openfilesv[n] indicates the file is not
 * open.
 *
 * class FileInputStream
 * {
 *		private FileDescriptor fd;
 * };
 *
 *****************************************************************************/


/* JDK 1.1.7A & 1.4
+    private native void open(String) throws java.io.FileNotFoundException;
+    public native int read() throws java.io.IOException;
+    private native int readBytes(byte[], int, int) throws java.io.IOException;
+    public native long skip(long) throws java.io.IOException;
+    public native int available() throws java.io.IOException;
+#   private native void close() throws java.io.IOException;
 1.4 private native void close0() throws java.io.IOException;
-1.4 private static native void initIDs();
*/


#if JDK_VERSION >= 140
/**
 * private static native void initIDs();
 */
void Java_java_io_FileInputStream_initIDs( JNIEnv* env, jclass staticself )
{
	/* If openfilesv[] has not been initialized, call FileDescriptor.initIDs() */
	if( !openfilesv[0] )
		Java_java_io_FileDescriptor_initIDs( env, NULL );
}
#endif /* JDK_VERSION 1.4.0 */


/**
 * Returns the number of bytes that can be read from this file without
 * blocking.
 *
 * @exception  IOException  if an error occurs.
 * @since      JDK1.0
 *
 * int available() throws IOException;
 */
jint Java_java_io_FileInputStream_available( JNIEnv* env, jobject self )
{
	/* Get the FileDescriptor (first field) */
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	jint pos, end, avail, fd;

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return 0;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return 0;
	}

	switch( fd )
	{
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Read operation not allowed on out or err stream" );
		return 0;
	case 0: /* stdin */
	default:
		if( openfilesv[fd]->kind == oft_Archive )
		{
			/* length holds number of byte remaining */
			avail = openfilesv[fd]->available;
		}
		else
		{
			/* find the end and calculate the difference */
			pos = ftell( openfilesv[fd]->f );
			fseek( openfilesv[fd]->f, 0, SEEK_END );
			end = ftell( openfilesv[fd]->f );
			avail = end - pos;
			fseek( openfilesv[fd]->f, pos, SEEK_SET );
		}
		/* DEBUG
		if( verbose )
			printf( "FileInputStream.available(): %d bytes available in file %d.\n", avail, fd );
		DEBUG */
		return avail;
	}
}


/**
 * Opens the file named 'name' for reading.
 *
 * void open( String name ) throws IOException;
 */
void Java_java_io_FileInputStream_open( JNIEnv* env, jobject self, jobject str )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, str ) )
	{
		int index = findFreeFD();
		FILE* f;

		/* Try to open the file */
		if( !(f = fopen( filename, "rb" )) )
		{
			char buf[1024];
#ifdef WIN32
			sprintf( buf, "Failed to open file for reading \"%s\"", filename );
#else
			sprintf( buf, "Failed to open file for reading \"%s\" (errno = %d)", filename, errno );
#endif
			env->retcode = newCompleteException( env, IOException, buf );
			return;
		}

		/* Fill in the OpenFile_Entry */
		openfilesv[index] = javavm_malloc( sizeof(OpenFile_Entry) );
		openfilesv[index]->f         = f;
		openfilesv[index]->kind      = oft_Normal;
		openfilesv[index]->available = -1;

		/* Fill in 'fd' field of the FileDescriptor object */
		self->data[0] = (u4) newJavaIoFileDescriptor( index );

		if( verbose )
			printf( "FileInputStream.open(): Opened file %d as \"%s\".\n", index, filename );
		return;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
	return;
}


/**
 * Reads a single byte from this file, blocking if no input is yet available.
 * Returns -1 to indicate EOF.
 *
 * @exception  IOException  if an error occurs.
 * @since      JDK1.0
 *
 * int read() throws IOException;
 */
jint Java_java_io_FileInputStream_read( JNIEnv* env, jobject self )
{
	/* Get the FileDescriptor (first field) */
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	jint c, fd;

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return 0;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return 0;
	}

	switch( fd )
	{
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Read operation not allowed on out or err stream" );
		return 0;
	case 0: /* stdin */
	default:
		if( openfilesv[fd]->kind == oft_Archive )
		{
			if( openfilesv[fd]->available <= 0 )	return EOF;
			c = fgetc( openfilesv[fd]->f );
			openfilesv[fd]->available--;
		}
		else
			c = fgetc( openfilesv[fd]->f ); /* != EOF */
		break;
	}

	/* DEBUG
	if( verbose )
		printf( "FileInputStream.read( %d ): Read 1 byte.\n", fd );
	DEBUG */

	return c;
}


/**
 * Reads 'len' bytes from the file into the array 'b', starting at position
 * 'off'.  Returns the actual number of bytes read, or -1 if we are at EOF.
 *
 * @exception IOException  if an error occurs. 
 *
 * int readBytes( byte b[], int off, int len ) throws IOException;
 */
jint Java_java_io_FileInputStream_readBytes( JNIEnv* env, jobject self, jobject array, jint off, jint len )
{
	/* Get the FileDescriptor (first field) */
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	jint fd;

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad file descriptor" );
		return 0;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return 0;
	}

	/* Validate the array and check off + len is inside array bounds */
	if( array && (array->cf == NULL) && (array->type == T_BYTE) && (off + len <= (jint) array->length) )
	{
		char* dst = (char*) array->data;
		int c;
		size_t n;

		switch( fd )
		{
		case 1: /* stdout */
		case 2: /* stderr */
			env->retcode = newCompleteException( env, IOException, "Read operation not allowed on out or err stream" );
			return 0;
		case 0: /* stdin */
			do {
				c = GETCHAR();
			} while( ((c > 0) && (c < 32) && (c != '\r')) || (c == 127) );
			if( c == '\r' ) {
				PUTCHAR( '\r' ); PUTCHAR( '\n' );
				dst[off]   = '\r';
				dst[off+1] = '\n';
				return 2;
			} else if( c > 0 ) {
			    PUTCHAR( c );
			    dst[off] = c;
			    return 1;
			} else
			    return 0;
		default:
			if( openfilesv[fd]->kind == oft_Archive )
			{
				if( openfilesv[fd]->available <= 0 )  return EOF;
				if( openfilesv[fd]->available <= len )  len = openfilesv[fd]->available;
				n = fread( &dst[off], 1, len, openfilesv[fd]->f );
				openfilesv[fd]->available -= n;
			}
			else
			{
				n = fread( &dst[off], 1, len, openfilesv[fd]->f );
				if( (n == 0) && feof( openfilesv[fd]->f ) )  return EOF;
			}
			break;
		}

		/* DEBUG
		if( verbose )
			printf( "FileInputStream.readBytes( %d ): Read %d bytes (%d requested).\n", fd, n, len );
		DEBUG */

		return n;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to read bytes from FileInputStream" );
	return 0;
}


/**
 * Skips forward 'n' bytes of the file, if possible.  Returns the actual number
 * of bytes skipped.
 *
 * @exception  IOException  if an error occurs.
 * @since      JDK1.0
 *
 * long skip( long len ) throws IOException;
 */
jlong Java_java_io_FileInputStream_skip( JNIEnv* env, jobject self, jlong len )
{
	/* Get the FileDescriptor (first field) */
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	jint fd, pos, newpos;

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return 0;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode =  newCompleteException( env, IOException, "Bad file descriptor" );	
		return 0;
	}

	switch( fd )
	{
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Read operation not allowed on out or err stream" );
		return 0;
	case 0: /* stdin */
	default:
		if( openfilesv[fd]->kind == oft_Archive )
		{
			if( openfilesv[fd]->available <= 0 )  return EOF;
			if( openfilesv[fd]->available <= len )  len = openfilesv[fd]->available;
			fseek( openfilesv[fd]->f, (long) len, SEEK_CUR );
			openfilesv[fd]->available -= (long) len;
		}
		else
		{
			pos = ftell( openfilesv[fd]->f );
			fseek( openfilesv[fd]->f, (long) len, SEEK_CUR );
			newpos = ftell( openfilesv[fd]->f );
			len = newpos - pos;
		}
		/* DEBUG
		if( verbose )
			printf( "FileInputStream.skip(): Skipped %d bytes in file %d.\n", len, fd );
		DEBUG */
		return len; /* ### assume success! */
	}
}


/**
 * Closes this file releases any system resources. 
 *
 * @exception  IOException  if an error occurs.
 * @since      JDK1.0
 *
 * void close() throws IOException;
 */
#if JDK_VERSION < 140
void Java_java_io_FileInputStream_close( JNIEnv* env, jobject self )
#else
void Java_java_io_FileInputStream_close0( JNIEnv* env, jobject self )
#endif /* JDK 1.4 */
{
	/* Get the FileDescriptor (first field) */
	ObjRef *fileDesc = (ObjRef*) self->data[0];
	jint fd;

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad file descriptor" );
		return;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return;
	}

	switch( fd )
	{
	case 0: /* stdin */
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Close not permitted on in, out or err" );
		break;
	default:
		fclose( openfilesv[fd]->f );
		if( openfilesv[fd]->kind != oft_BuiltIn ) {
			javavm_free( openfilesv[fd], sizeof(OpenFile_Entry) );
			openfilesv[fd] = NULL;
		}
		break;
	}

	if( verbose )
		printf( "FileInputStream.close(): Closed file %d.\n", fd );
}
