/******************************************************************************
 *
 * Filename: java_io_FileOutputStream.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  25th March 2002
 *
 * Purpose:  Native method implementations for java.io.FileOutputStream.
 *
 * (c) Copyright R.H.Thornburrow, 2002-3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef WIN32
#include <errno.h>
#endif

#include "../engine/config.h"

#include "java_io_FileOutputStream.h"

#include "../engine/native-io.h"
#include "../engine/util.h"
#include "../engine/openfiles.h"
#include "../engine/memory.h"

#include "java_io_FileDescriptor.h"


/******************************************************************************
 *
 * class java.io.FileOutputStream
 *
 *
 * The system dependent file descriptor. The value is
 * 1 more than actual file descriptor. This means that
 * the default value 0 indicates that the file is not open.
 *
 * class FileOutputStream
 * {
 *		private FileDescriptor fd;
 * };
 *
 *****************************************************************************/


/* JDK 1.4
+    private native void open(String) throws java.io.FileNotFoundException;
+    private native void openAppend(String) throws java.io.FileNotFoundException;
+    public native void write(int) throws java.io.IOException;
+    private native void writeBytes(byte[], int, int) throws java.io.IOException;
+#   private native void close() throws java.io.IOException;
 1.4 private native void close0() throws java.io.IOException;
-1.4 private static native void initIDs();
*/


#if JDK_VERSION >= 140
/**
 * private static native void initIDs();
 */
void Java_java_io_FileOutputStream_initIDs( JNIEnv* env, jclass staticself )
{
	/* If openfilesv[] has not been initialized, call FileDescriptor.initIDs() */
	if( !openfilesv[0] )
		Java_java_io_FileDescriptor_initIDs( env, NULL );
}
#endif /* JDK_VERSION 1.4.0 */


/**
 * Opens the file named 'name' for writing.
 *
 * private native void open( String name ) throws IOException;
 */
void Java_java_io_FileOutputStream_open( JNIEnv* env, jobject self, jobject name )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, name ) )
	{
		int index = findFreeFD();
		FILE* f;

		/* Try to open the file */
		if( !(f = fopen( filename, "wb" )) )
		{
			char buf[1024];
#ifdef WIN32
			sprintf( buf, "Failed to open file for writing \"%s\"", filename );
#else
			sprintf( buf, "Failed to open file for writing \"%s\" (errno = %d)", filename, errno );
#endif
			env->retcode = newCompleteException( env, IOException, buf );
			return;
		}

		/* Fill in the OpenFile_Entry */
		openfilesv[index] = javavm_malloc( sizeof(OpenFile_Entry) );
		openfilesv[index]->f         = f;
		openfilesv[index]->kind      = oft_Normal;
		openfilesv[index]->available = -1;

		/* Fill in the 'fd' field of the FileDescriptor object */
		self->data[0] = (u4) newJavaIoFileDescriptor( index );
		return;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
}


/**
 * Opens the file named 'name' for writing in append mode.
 *
 * private native void openAppend( String name ) throws IOException;
 */
void Java_java_io_FileOutputStream_openAppend( JNIEnv* env, jobject self, jobject name )
{
	/* ### Implement me! */
	env->retcode = newCompleteException( env, InternalError, "FileOutputStream.openAppend() not implemented!" );
	return;
}


/**
 * Writes a single byte to this file. 
 *
 * @exception  IOException  if an error occurs.
 * @since      JDK1.0
 *
 * public native void write( int b ) throws IOException;
 */
void Java_java_io_FileOutputStream_write( JNIEnv* env, jobject self, jint b )
{
	/* Get the FileDescriptor (first field) */
	ObjRef *fileDesc = (ObjRef*) self->data[0];
	jint fd;
	char buf[2];

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return;
	}

	switch( fd )
	{
	case 0: /* stdin */
		env->retcode = newCompleteException( env, IOException, "Write operation not allowed on input stream" );
		return;
	case 1: /* stdout */
	case 2: /* stderr */
		sprintf( buf, "%c", (u2) b );
		systemOutPrint( buf, 1 );
		break;
	default:
		fputc( b, openfilesv[fd]->f );
		break;
	}
}


/**
 * Writes 'len' bytes to the file from the array 'b', starting at position
 * 'off'.
 *
 * @exception IOException  if an error occurs.
 *
 * private native void writeBytes( byte b[], int off, int len ) throws IOException;
 */
void Java_java_io_FileOutputStream_writeBytes( JNIEnv* env, jobject self, jobject array, jint off, jint len )
{
	/* Get the FileDescriptor (first field) */
	ObjRef *fileDesc = (ObjRef*) self->data[0];
	int32 n;
	jint fd;

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return;
	}

	/* Validate the array and check off + len is inside array bounds */
	if( array && (array->cf == NULL) && (array->type == T_BYTE) && (off + len <= (jint) array->length) )
	{
		const char* src = (const char*) array->data;

		switch( fd )
		{
		case 0: /* stdin */
			env->retcode = newCompleteException( env, IOException, "Write operation not allowed on input stream" );
			return;
		case 1: /* stdout */
		case 2: /* stderr */
			{
			char* buf = (char*) javavm_malloc( len + 1 );
			strncpy( buf, src + off, len );
			buf[ len ] = '\0';
			systemOutPrint( buf, len );
			javavm_free( buf, len+1 );
			}
			break;
		default:
			for( n = 0; n < len; n++ )
				fputc( src[n], openfilesv[fd]->f );
			break;
		}

		return;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to write bytes to FileOutputStream" );
}


/**
 * Closes this file releases any system resources. 
 *
 * @exception  IOException  if an error occurs.
 * @since      JDK1.0
 *
 * public native void close() throws IOException;
 */
#if JDK_VERSION < 140
void Java_java_io_FileOutputStream_close( JNIEnv* env, jobject self )
#else
void Java_java_io_FileOutputStream_close0( JNIEnv* env, jobject self )
#endif /* JDK 1.4 */
{
	/* Get the FileDescriptor (first field) */
	ObjRef *fileDesc = (ObjRef*) self->data[0];
	jint fd;

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return;
	}

	switch( fd )
	{
	case 0: /* stdin */
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Close not permitted on in, out or err" );
		return;
	default:
		fclose( openfilesv[fd]->f );
		if( openfilesv[fd]->kind != oft_BuiltIn ) {
			javavm_free( openfilesv[fd], sizeof(OpenFile_Entry) );
			openfilesv[fd] = NULL;
		}
		break;
	}
}
