/******************************************************************************
 *
 * Filename: java_io_RandomAccessFile.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  21st April 2003
 *
 * Purpose:  Native method implementations for java.io.RandomAccessFile.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#include "../engine/config.h"

#include "java_io_RandomAccessFile.h"

#include "../engine/util.h"
#include "../engine/openfiles.h"

#include "java_io_FileInputStream.h"
#include "java_io_FileOutputStream.h"


/*
-    private native void open(java.lang.String, int) throws java.io.FileNotFoundException;
     public  native int  read() throws java.io.IOException;
-    private native int  readBytes(byte[], int, int) throws java.io.IOException;
     public  native void write(int) throws java.io.IOException;
     private native void writeBytes(byte[], int, int) throws java.io.IOException;
     public  native long getFilePointer() throws java.io.IOException;
+    public  native void seek(long) throws java.io.IOException;
+    public  native long length() throws java.io.IOException;
     public  native void setLength(long) throws java.io.IOException;
-    public  native void close() throws java.io.IOException;
 1.4 private static native void initIDs();
 */

/**
 * Opens a file and returns the file descriptor.  The file is opened in read-
 * write mode if writeable is true, else the file is opened as read-only.
 *
 * @param name       the name of the file
 * @param writeable  the boolean indicating whether file is writeable or not.
 *
 * public native void open( String name, boolean writeable ) throws IOException;
 */
void Java_java_io_RandomAccessFile_open( JNIEnv* env, jobject self, jobject name, jboolean writeable )
{
	/* ### Cheat a little - use the existing file opening code */

	if( writeable ) {
		Java_java_io_FileOutputStream_open( env, self, name );
		return;
	}
	
	Java_java_io_FileInputStream_open( env, self, name );
}


/**
 * Returns the length of this file.
 *
 * @exception  IOException  if an error occurs.
 * @since      JDK1.0
 *
 * public native long length() throws java.io.IOException;
 */
jlong Java_java_io_RandomAccessFile_length( JNIEnv* env, jobject self )
{
	/* Get the FileDescriptor (first field) */
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	jint pos, start, end, length, fd;

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return 0;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return 0;
	}

	switch( fd )
	{
	case 0: /* stdin */
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Length not allowed on in, out or err stream" );
		return 0;
	default:
		/* find the start then end and calculate the difference */
		pos = ftell( openfilesv[fd]->f );
		fseek( openfilesv[fd]->f, 0, SEEK_SET );
		start = ftell( openfilesv[fd]->f );
		fseek( openfilesv[fd]->f, 0, SEEK_END );
		end = ftell( openfilesv[fd]->f );
		length = end - start;
		fseek( openfilesv[fd]->f, pos, SEEK_SET );
		return length;
	}
}


/**
 * Reads 'len' bytes from the file into the array 'b', starting at position
 * 'off'.  Returns the actual number of bytes read, or -1 if we are at EOF.
 *
 * @exception IOException  if an error occurs. 
 *
 * private native int readBytes( byte[] b, int off, int len ) throws java.io.IOException;
 */
int Java_java_io_RandomAccessFile_readBytes( JNIEnv* env, jobject self, jobject b, jint off, jint len )
{
	/* ### Cheat a little - use the existing file reading code */
	
	return Java_java_io_FileInputStream_readBytes( env, self, b, off, len );
}


/**
 * Moves the file pointer to the specified offset, 'pos', from the beginning of
 * the file.  If the offset is beyond the end of the file, the file size is not
 * changed.
 *
 * @exception IOException  if an error occurs.
 * @since     JDK1.0
 *
 * public native void seek( long pos ) throws java.io.IOException;
 */
void Java_java_io_RandomAccessFile_seek( JNIEnv* env, jobject self, jlong pos )
{
	/* Get the FileDescriptor (first field) */
	ObjRef* fileDesc = (ObjRef*) self->data[0];
	jint fd;

	/* Make sure fileDesc is a FileDescriptor object */
	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return;
	}

	/* Get the field 'fd' from the FileDescriptor (first field) */
	fd = fileDesc->data[0];

	/* Validate 'fd', plus make sure it is an open file */
	if( (fd < 0) || (fd >= openfilesc) || !openfilesv[fd] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return;
	}

	switch( fd )
	{
	case 0: /* stdin */
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Seek not allowed on in, out or err stream" );
		break;
	default:
		/* Set file pointer relative to start */
		fseek( openfilesv[fd]->f, (long) pos, SEEK_SET );
		break;
	}
}

/**
 * Closes this file and releases any system resources associated with it. 
 *
 * @exception  IOException  if an error occurs.
 * @since      JDK1.0
 *
 * public native void close() throws java.io.IOException;
 */
void Java_java_io_RandomAccessFile_close( JNIEnv* env, jobject self )
{
	/* ### Cheat a little - use the existing file close code */

#if JDK_VERSION < 140
	Java_java_io_FileInputStream_close( env, self );
#else
	Java_java_io_FileInputStream_close0( env, self );
#endif /* JDK 1.4 */
}
