/******************************************************************************
 *
 * Filename: java_lang_Class.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for java.lang.Class.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "../engine/config.h"

#include "java_lang_Class.h"

#include "../engine/util.h"
#include "../engine/disasm.h"
#include "../engine/hashtable.h"
#include "../engine/globals.h"
#include "../engine/memory.h"

#include "java_lang_String.h"


/* JDK 1.1.7A & 1.4.0
-1.4 private static native void registerNatives();
+#   public static native java.lang.Class forName(java.lang.String);
+1.4 private static native java.lang.Class forName0(String,boolean,ClassLoader) throws ClassNotFoundException;
+#   public native java.lang.Object newInstance();
     public native boolean isInstance(java.lang.Object);
+    public native boolean isAssignableFrom(java.lang.Class);
+    public native boolean isInterface();
-    public native boolean isArray();
+    public native boolean isPrimitive();
+    public native java.lang.String getName();
+#   public native java.lang.ClassLoader getClassLoader();
+1.4 native java.lang.ClassLoader getClassLoader0();
+    public native java.lang.Class getSuperclass();
+    public native java.lang.Class getInterfaces()[];
     public native java.lang.Class getComponentType();
+    public native int getModifiers();
     public native java.lang.Object getSigners()[];
     native void setSigners(java.lang.Object[]);
+    static native java.lang.Class getPrimitiveClass(java.lang.String);
+#   private native java.lang.reflect.Field getFields0(int)[];
+1.4 private native java.lang.reflect.Field getDeclaredFields0(boolean)[];
 #   private native java.lang.reflect.Method getMethods0(int)[];
 1.4 private native java.lang.reflect.Method getDeclaredMethods0(boolean)[];
+#   private native java.lang.reflect.Constructor getConstructors0(int)[];
+1.4 private native java.lang.reflect.Constructor getDeclaredConstructors0(boolean)[];
+#   private native java.lang.reflect.Field getField0(java.lang.String, int);
 #   private native java.lang.reflect.Method getMethod0(java.lang.String, java.lang.Class[], int);
+#   private native java.lang.reflect.Constructor getConstructor0(java.lang.Class[], int);
 1.4 private native java.lang.Class getDeclaredClasses0()[];
+1.4 private static native boolean desiredAssertionStatus0(java.lang.Class);
 1.4 public native java.lang.Class getDeclaringClass();
 1.4 private native java.security.ProtectionDomain getProtectionDomain0();
 1.4 native void setProtectionDomain0(java.security.ProtectionDomain);

    #   - not in JDK 1.4
	1.4 - in JDK1.4 (certainly JDK1.2 or later)
	+   - implemented
	-   - dummied
*/


/* Mapping from primitive type to Java Class */
jclass primitiveClass[T_LAST];

/* Mapping from array descriptor to Java Class object */
Hashtable *arrayClasses = NULL;

/**
 * Internal function.  Gets the Class object representing the primitive type
 * 'kind'.
 */
jclass getClassForPrimitiveType( u1 kind )
{
	/* Make sure we have a valid primitive type (ObjRef is not primitive) */
	if( (kind < T_BOOLEAN) || (kind > T_LONG) )
	{
		printf( "WARNING: getClassForPrimitiveType( %d ) - invalid type\n", kind );
		return NULL;
	}

	/* If the Class object has not been created yet, create it */
	if( !primitiveClass[kind] )
		primitiveClass[kind] = newJavaLangClass( NULL );

	/* Return the one and only instance of Class for that type */
	return primitiveClass[kind];
}


/**
 * Internal function.  Reverse lookup of the primitive type represented by a
 * given Class object.
 */
u1 getPrimitiveTypeForClass( jclass clazz )
{
	ClassFile* cf = (ClassFile*) clazz->data[clazz->cf->object_size];	/* last field is cf */
	u1 n;

	/* Non-null indicates a real class */
	if( cf )
		return 0;

	/* Check all our primitive class objects */
	for( n = T_BOOLEAN; n <= T_LONG; n++ )
	{
		/* If we match, it's primitive */
		if( primitiveClass[n] && (primitiveClass[n] == clazz) )
			return n;
	}

	/* Passed end, so not valid */
	return 0;
}


/**
 * Internal function.  Gets the Class object representing the array type with
 * the descriptor 'desc'.
 */
jclass getClassForArrayType( const char* desc )
{
	jclass arrayClass;

	if( !arrayClasses )
		arrayClasses = Hashtable_new( 16 );

	arrayClass = Hashtable_get( arrayClasses, desc );

	if( !arrayClass )
	{
		arrayClass = newJavaLangClass( NULL );
		Hashtable_put( arrayClasses, strdup(desc), arrayClass );
	}

	return arrayClass;
}


/**
 * Internal function.  Returns 1 if the give Class object represents an array.
 */
int isArrayType( jclass clazz )
{
	ClassFile* cf = (ClassFile*) clazz->data[clazz->cf->object_size];	/* last field is cf */

	/* Non-null indicates a real class */
	if( cf )
		return 0;

	/* If hashtable hasn't been created yet, we can't be an array class */
	if( !arrayClasses )
		return 0;

	/* Does hashtable contain this class? */
	return Hashtable_findValue( arrayClasses, clazz ) ? 1 : 0;
}


#if JDK_VERSION >= 140
/**
 * Returns whether assertions are enabled for 'aClass' (this is done with
 * -ea:<class> on the command line).  Assertions are a new feature in JDK1.4.
 *
 * Assertions require no more than for this method to return true in order for
 * them to be used.  At present we don't maintain a list of classes which have
 * assertions enabled.  This method would normally return true/false on a class
 * by class basis.
 *
 * private static native boolean desiredAssertionStatus0( java.lang.Class aClass );
 *
 * @since JDK1.4
 */
jboolean Java_java_lang_Class_desiredAssertionStatus0( JNIEnv* env, jclass staticself, jclass aClass )
{
	/*ClassFile* cf = (ClassFile*) aClass->data[aClass->cf->object_size];*/    /* last field is cf */
	return assertions_enabled ? JNI_TRUE : JNI_FALSE;
}
#endif /* JDK 1.4 */


/**
 * Returns the Class object associated with the class with the given string
 * as is fully qualified name.  Loads and links the class and returns the Class
 * object representing the class.  Throws a ClassNotFoundException if not
 * successful.
 *
 * @since      JDK1.0
 *
 * public static native Class forName(String className)
 *	throws ClassNotFoundException;
 */
#if JDK_VERSION < 140
jclass Java_java_lang_Class_forName( JNIEnv* env, jclass staticself, jstring str )
#else
jclass Java_java_lang_Class_forName0(
	JNIEnv* env, jclass staticself, jstring str, jboolean aBoolean, jobject aClassLoader )
#endif /* JDK_VERSION 1.4.0 */
{
	char classname[255];

	if( extractFromJavaLangString( classname, 255, str ) )
	{
		ClassFile* cf;
		u4 n;

		/* Convert class name to slash separated name */
		for( n = 0; n < strlen( classname ); n++ )
		{
			if( classname[n] == '.' )
				classname[n] = '/';
		}

		/* Array types are special - they aren't real classes */
		if( classname[0] == '[' )
			return getClassForArrayType( classname );

		/* Locate the ClassFile for the class, loading if necessary */
		if( !( cf = getClass( env, classname ) ) )
		{
			if( verbose )
				printf( "Class.forName() failed to load class \"%s\"\n", classname );

			/* all exceptions translate into ClassNotFoundException's */
			env->retcode = newCompleteException( env, ClassNotFoundException, classname );
			return NULL;
		}

		/* Return the Class object */
		return cf->classobj;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
	return NULL;
}


/**
 * Returns the class loader that created this class or interface or null if the
 * class was not created by a class loader.
 *
 * @since   JDK1.0
 *
 * public native java.lang.ClassLoader getClassLoader()
 */
#if JDK_VERSION < 140
jobject Java_java_lang_Class_getClassLoader( JNIEnv* env, jobject self )
#else
jobject Java_java_lang_Class_getClassLoader0( JNIEnv* env, jobject self )
#endif /* JDK 1.4 */
{
	/* We are always the system class loader so return null */
	return NULL;
}


/**
 * Internal method to split a method descriptor into an array of Class objects,
 * one for each parameter type.  Returns null if an error occurs.
 */
jobject descriptorToClassArray( JNIEnv* env, ClassFile* cf, u2 index )
{
	ObjRef *paramClasses[32], *ret;
	u2 desclen, n, prev, params, array;
	char desc[512], param[512];
	u1 kind = 0;

	/* Get the method descriptor (i.e. its parameters) */
	if( !CHECKCONSTANT( cf, index, CONSTANT_Utf8 ) )
		return NULL;
	
	/* Convert utf8 descriptor to ASCII */
	desclen = sprintUtf8( desc, (CONSTANT_Utf8_info*) cf->constant_pool[index] );

	/* Check descriptor starts with a bracket */
	if( !desclen || (desc[0] != '(') )
		return NULL;

	/* Work through the arguments as listed in the descriptor.  'params'
	   will hold the number of arguments, 'paramClasses' will hold the
	   java.lang.Class objects representing those types */
	params = 0; array = 0;
	
	/* End of previous param descriptor */
	prev = 0;

	for( n = 1; (n < desclen) && (desc[n] != ')'); n++ )
	{
		switch( desc[n] )
		{
		case 'Z': kind = T_BOOLEAN; break;
		case 'C': kind = T_CHAR;    break;
		case 'F': kind = T_FLOAT;   break;
		case 'D': kind = T_DOUBLE;  break;
		case 'B': kind = T_BYTE;    break;
		case 'S': kind = T_SHORT;   break;
		case 'I': kind = T_INT;     break;
		case 'J': kind = T_LONG;    break;
		case 'L': kind = T_OBJREF;  break;
		case '[': array++; continue;
		}

		/* Is it an array type? */
		if( !array )
		{
			/* Is it an object type? */
			if( kind == T_OBJREF )
			{
				ClassFile *paramcf;
				u2 m = 0;

				/* Copy the class name into a buffer */
				for( n++; (n < desclen) && (desc[n] != ';'); n++ )
					param[m++] = desc[n];

				param[m] = '\0';

				/* Fetch the ClassFile and then its Class object */
				paramcf = getClass( env, param );
				paramClasses[params++] = paramcf ? paramcf->classobj : NULL;
			}
			else  /* Must be a primitive type */
			{				
				/* Fetch the Class object for the type */
				paramClasses[params++] = getClassForPrimitiveType( kind );
			}
		}
		else
		{
			/* Is it an object type? */
			if( kind == T_OBJREF )
			{
				/* Find end of object array type descriptor */
				for( n++; (n < desclen) && (desc[n] != ';'); n++ );
			}

			/* Copy the array type descriptor into a buffer */
			strncpy( param, &desc[prev+1], n - prev );
			param[n - prev] = '\0';

			/* Fetch the Class object for the array type */
			paramClasses[params++] = getClassForArrayType( param );
		}

		array = 0;
		prev = n;
	}

	/* Create an array of the correct size to return */
	ret = newObjectRefAndArray( T_OBJREF, params );

	/* Copy the Class objects from paramClasses into the return array */
	for( n = 0; n < params; n++ )
		ret->data[n] = (u4) paramClasses[n];

	return ret;
}


/**
 * Returns all the constructors to this class.  If there are no constructors,
 * then a zero length array must be returned.
 *
 * @since  JDK1.4?  (certainly JDK1.2 or later)
 *
 * private native java.lang.reflect.Constructor[] getDeclaredConstructors0( boolean publicOnly );
 */
#if JDK_VERSION < 140
jobject Java_java_lang_Class_getConstructors0( JNIEnv* env, jobject self, jint privateFlag )
#else
jobject Java_java_lang_Class_getDeclaredConstructors0( JNIEnv* env, jobject self, jboolean publicOnly )
#endif /* JDK 1.4 */
{
	ClassFile* cf = (ClassFile*) self->data[self->cf->object_size];    /* last field is cf */

	ObjRef *ret, *parameterTypes, *exceptionTypes, *constructor;
	method_info* constrs[32];
#if JDK_VERSION < 140
	int32 clazzOffset, paramTypesOffset, excepTypesOffset;
	jboolean publicOnly = privateFlag ? JNI_FALSE : JNI_TRUE;
#else
	u2 argc;
	u1 argvt[5];
	u4 argv[5];
#endif /* JDK 1.4 */
	u2 index, count = 0;
	ClassFile* constrcf;
	int retcode;

	/* Load java.lang.reflect.Constructor */
	if( !(constrcf = getClass( env, "java/lang/reflect/Constructor" )) )
		return NULL;  /* exception will be on stack */

#if JDK_VERSION < 140
	clazzOffset = getFieldOffset( constrcf, "clazz", "Ljava/lang/Class;" );
	paramTypesOffset = getFieldOffset( constrcf, "parameterTypes", "[Ljava/lang/Class;" );
	excepTypesOffset = getFieldOffset( constrcf, "exceptionTypes", "[Ljava/lang/Class;" );

	if( (clazzOffset == -1) || (paramTypesOffset == -1) || (excepTypesOffset == -1) )
	{
		env->retcode = newCompleteException( env, InternalError, "java.lang.reflect.Constructor does not have correct fields" );
		return NULL;
	}
#endif /* !JDK 1.4 */

	/* Find all methods with the name "<init>" (i.e. constructors) */
	for( index = 0; (index < cf->methods_count) && (count < 32); index++ )
	{
		/* Are we including only public members? */
		if( publicOnly )
		{
			/* Is this a public constructor? */
			if( cf->methods[index] && ((cf->methods[index]->access_flags & ACC_PUBLIC) != 0)
				&& checkSimpleName( cf, cf->methods[index]->name_index, "<init>" ) )
				constrs[count++] = cf->methods[index];
		}
		else
		{
			/* Is this a constructor */
			if( cf->methods[index] && checkSimpleName( cf, cf->methods[index]->name_index, "<init>" ) )
				constrs[count++] = cf->methods[index];
		}
	}
	
	/* If we didn't reach the end, then count == 32 */
	if( index < cf->methods_count )
	{
		char buf[256];
		sprintClassName( buf, cf, cf->this_class );
		printf( "WARNING: java.lang.Class.getDeclaredConstructors0() for class %s has more than 32 constructors.\n", buf );
	}

	/* Create an array to hold the Constructor objects */
	ret = newObjectRefAndArray( T_OBJREF, count );
	addToList( ret, &arrayPool );

	/* For each method create a Constructor object and store it in the array */
	for( index = 0; index < count; index++ )
	{
		/* Setup the parameterTypes array */
		if( !(parameterTypes = descriptorToClassArray( env, cf, constrs[index]->descriptor_index )) )
		{
			env->retcode = newCompleteException( env, InternalError, "Error processing descriptor for constructor" );
			return NULL;
		}

		addToList( parameterTypes, &arrayPool );

		/* Create a new Constructor object */
		constructor = newObjectRefAndObject( constrcf );
		addToList( constructor, &objectPool );

		/* TODO: Fill in the exceptionTypes array */
		exceptionTypes = NULL;

#if JDK_VERSION < 140
		forceExecuteJavaMethod( constructor, env, "<init>", "()V" );
#else
		/* Setup the arguments to Constructor.<init>() */
		argc = 5;
		argvt[0] = T_OBJREF; argv[0] = (u4) self;							/* clazz */
		argvt[1] = T_OBJREF; argv[1] = (u4) parameterTypes;					/* parameterTypes */
		argvt[2] = T_OBJREF; argv[2] = (u4) exceptionTypes;					/* exceptionTypes */
		argvt[3] = T_INT;    argv[3] = (u4) constrs[index]->access_flags;	/* modifiers */
		argvt[4] = T_INT;    argv[4] = (u4) 0;								/* slot */

		/* Create a new stack frame for <init>() */
		forceExecuteJavaMethodWithArgs( constructor, env, "<init>", "(Ljava/lang/Class;[Ljava/lang/Class;[Ljava/lang/Class;II)V", argc, argv, argvt );
#endif /* JDK 1.4 */

		/* Make execution stop after <init>() completes */
		env->fp->stopOnPop = 1;
		retcode = 0;

		/* Execute <init>() */
		while( !retcode )
			retcode = executeInstruction( env );

		/* Make sure we stopped because of stop-on-pop - exception will be on stack */
		if( retcode != 1 )
			return NULL;

#if JDK_VERSION < 140
		constructor->data[clazzOffset] = (u4) self;
		constructor->data[paramTypesOffset] = (u4) parameterTypes;
		constructor->data[excepTypesOffset] = (u4) exceptionTypes;
#endif /* !JDK 1.4 */

		/* Store the new Constructor object in the array to be returned */
		ret->data[index] = (u4) constructor;
	}

	return ret;
}


#if JDK_VERSION < 140
/**
 * ??Returns the constructor with the given argument types or raises a
 * NoSuchMethodException??
 *
 * @since  JDK1.1?  (not in JDK 1.4)
 *
 * private native java.lang.reflect.Constructor[] getConstructors( Class[] paramTypes, boolean publicOnly );
 */
jobject Java_java_lang_Class_getConstructor0( JNIEnv* env, jobject self, jobject paramTypes, jint privateFlag )
{
	ClassFile* cf = (ClassFile*) self->data[self->cf->object_size];    /* last field is cf */

	ObjRef *exceptionTypes, *constructor;
	int32 clazzOffset, paramTypesOffset, excepTypesOffset;
	jboolean publicOnly = privateFlag ? JNI_FALSE : JNI_TRUE;
	u2 n, index;
	ClassFile* constrcf;
	int retcode, found;

	/* Load java.lang.reflect.Constructor */
	if( !(constrcf = getClass( env, "java/lang/reflect/Constructor" )) )
		return NULL;  /* exception will be on stack */

	clazzOffset = getFieldOffset( constrcf, "clazz", "Ljava/lang/Class;" );
	paramTypesOffset = getFieldOffset( constrcf, "parameterTypes", "[Ljava/lang/Class;" );
	excepTypesOffset = getFieldOffset( constrcf, "exceptionTypes", "[Ljava/lang/Class;" );

	if( (clazzOffset == -1) || (paramTypesOffset == -1) || (excepTypesOffset == -1) )
	{
		env->retcode = newCompleteException( env, InternalError, "java.lang.reflect.Constructor does not have correct fields" );
		return NULL;
	}

	/* Find all methods with the name "<init>" (i.e. constructors) */
	for( index = 0; index < cf->methods_count; index++ )
	{
		if( !cf->methods[index] )
			continue;

		/* If only getting public fields, skip non-public ones */
		if( publicOnly && ((cf->methods[index]->access_flags & ACC_PUBLIC) == 0) )
			continue;

		/* Is this a constructor? */
		if( checkSimpleName( cf, cf->methods[index]->name_index, "<init>" ) )
		{
			/* Convert args of this constructor into a Class[] */
			jobject argTypes = descriptorToClassArray( env, cf, cf->methods[index]->descriptor_index );
			
			/* If conversion fails, abort */
			if( !argTypes )
			{
				env->retcode = newCompleteException( env, InternalError, "Error processing descriptor for constructor" );
				return NULL;
			}

			/* Number of arguments must be equal */
			if( argTypes->length != paramTypes->length )
			{
				javavm_free( argTypes, argTypes->length * sizeof(ObjRef*) );
				argTypes = NULL;
				continue;
			}

			/* Assume matched and disprove it */
			found = 1;

			/* Match each argument type in turn - only one instance of each Class, so use == */
			for( n = 0; n < paramTypes->length; n++ )
			{
				if( argTypes->data[n] != paramTypes->data[n] )
				{
					found = 0;
					break;
				}
			}

			/* Done with arg types array - free it */
			javavm_free( argTypes, argTypes->length * sizeof(ObjRef*) );
			argTypes = NULL;

			/* Has a match been found - if so, stop */
			if( found )
				break;
		}
	}
	
	/* Did we find a match? */
	if( index >= cf->methods_count )
	{
		env->retcode = newCompleteException( env, NoSuchMethodException, "No matching constructor" );
		return NULL;
	}

	/* Create a new Constructor object */
	constructor = newObjectRefAndObject( constrcf );
	addToList( constructor, &objectPool );

	/* TODO: Fill in the exceptionTypes array */
	exceptionTypes = NULL;

	forceExecuteJavaMethod( constructor, env, "<init>", "()V" );

	/* Make execution stop after <init>() completes */
	env->fp->stopOnPop = 1;
	retcode = 0;

	/* Execute <init>() */
	while( !retcode )
		retcode = executeInstruction( env );

	/* Make sure we stopped because of stop-on-pop - exception will be on stack */
	if( retcode != 1 )
		return NULL;

	constructor->data[clazzOffset] = (u4) self;
	constructor->data[paramTypesOffset] = (u4) paramTypes;
	constructor->data[excepTypesOffset] = (u4) exceptionTypes;

	return constructor;
}
#endif /* !JDK 1.4 */


/**
 * Internal method to convert a field descriptor into a Class objects.  Returns
 * null if an error occurs.
 */
jclass descriptorToClass( JNIEnv* env, ClassFile* cf, u2 index )
{
	u2 desclen, n, array;
	char desc[512];
	u1 kind = 0;

	/* Get the descriptor */
	if( !CHECKCONSTANT( cf, index, CONSTANT_Utf8 ) )
		return NULL;
	
	/* Convert utf8 descriptor to ASCII */
	desclen = sprintUtf8( desc, (CONSTANT_Utf8_info*) cf->constant_pool[index] );

	/* Check descriptor starts with a bracket */
	if( !desclen )
		return NULL;

	/* Work through the descriptor counting any array brackets and converting
	   the type to an int */
	array = 0;

	for( n = 0; n < desclen; n++ )
	{
		switch( desc[n] )
		{
		case 'Z': kind = T_BOOLEAN; break;
		case 'C': kind = T_CHAR;    break;
		case 'F': kind = T_FLOAT;   break;
		case 'D': kind = T_DOUBLE;  break;
		case 'B': kind = T_BYTE;    break;
		case 'S': kind = T_SHORT;   break;
		case 'I': kind = T_INT;     break;
		case 'J': kind = T_LONG;    break;
		case 'L': kind = T_OBJREF;  break;
		case '[': array++; continue;
		}

		break;	/* If we get to here we have set kind, so stop */
	}

	/* Is it an array type? */
	if( array )
		return getClassForArrayType( desc );

	/* Is it an object type? */
	if( kind == T_OBJREF )
	{
		char classname[255];
		ClassFile *desccf;
		u2 m = 0;

		for( n++; (n < desclen) && (desc[n] != ';'); n++ )
			classname[m++] = desc[n];

		classname[m] = '\0';

		desccf = getClass( env, classname );
		return desccf ? desccf->classobj : NULL;
	}
	
	/* Must be a primitive type */
	return getClassForPrimitiveType( kind );
}


/**
 * Returns the fields of this class.  If there are no fields, then a zero
 * length array must be returned.
 *
 * @since  JDK1.4?  (certainly JDK1.2 or later)
 *
 * private native java.lang.reflect.Field[] getDeclaredFields0( boolean publicOnly );
 */
#if JDK_VERSION < 140
jobject Java_java_lang_Class_getFields0( JNIEnv* env, jobject self, jint declaredFlag )
#else
jobject Java_java_lang_Class_getDeclaredFields0( JNIEnv* env, jobject self, jboolean publicOnly )
#endif /* JDK 1.4 */
{
	ClassFile* cf = (ClassFile*) self->data[self->cf->object_size];    /* last field is cf */

	ObjRef *ret, *field, *fieldName, *fieldType;
	field_info *f;
#if JDK_VERSION < 140
	int32 clazzOffset, nameOffset, typeOffset;
	jboolean publicOnly = declaredFlag ? JNI_FALSE : JNI_TRUE;
#else
	u2 argc;
	u1 argvt[5];
	u4 argv[5];
#endif /* JDK 1.4 */
	u2 index, count = 0;
	ClassFile* fieldcf;
	int retcode;

	/* Load java.lang.reflect.Constructor */
	if( !(fieldcf = getClass( env, "java/lang/reflect/Field" )) )
		return NULL;  /* exception will be on stack */

#if JDK_VERSION < 140
	clazzOffset = getFieldOffset( fieldcf, "clazz", "Ljava/lang/Class;" );
	nameOffset  = getFieldOffset( fieldcf, "name", "Ljava/lang/String;" );
	typeOffset  = getFieldOffset( fieldcf, "type", "Ljava/lang/Class;" );

	if( (clazzOffset == -1) || (nameOffset == -1) || (typeOffset == -1) )
	{
		env->retcode = newCompleteException( env, InternalError, "java.lang.reflect.Field does not have correct fields" );
		return NULL;
	}
#endif /* !JDK 1.4 */

#if JDK_VERSION < 140
	/* Loop through super-classes */
	for( ; cf; cf = cf->super )
	{
#endif /* !JDK 1.4 */
		/* If only getting public fields must count how many */
		if( publicOnly )
		{
			for( index = 0; index < cf->fields_count; index++ )
			{
				/* Is this a public field? */
				if( (cf->fields[index]->access_flags & ACC_PUBLIC) != 0 )
					count++;
			}
		}
		else
			count += cf->fields_count;

#if JDK_VERSION < 140
		if( declaredFlag )
			break;
	}
#endif /* !JDK 1.4 */

	/* Create an array to hold the Constructor objects */
	ret = newObjectRefAndArray( T_OBJREF, count );
	addToList( ret, &arrayPool );

	/* Reset the counter, which is now used as the index into the return array */
	count = 0;
	cf = (ClassFile*) self->data[self->cf->object_size];    /* last field is cf */

#if JDK_VERSION < 140
	/* Loop through super-classes */
	for( ; cf; cf = cf->super )
	{
#endif /* !JDK 1.4 */
		/* For each field create a Field object and store it in the array */
		for( index = 0; index < cf->fields_count; index++ )
		{
			/* Lookup the field_info entry */
			f = cf->fields[index];

			/* If only getting public fields, skip non-public ones */
			if( publicOnly && ((cf->fields[index]->access_flags & ACC_PUBLIC) == 0) )
				continue;

			/* Get the descriptor */
			if( !CHECKCONSTANT( cf, f->name_index, CONSTANT_Utf8 ) )
			{
				env->retcode = newCompleteException( env, InternalError, "Error processing name for field" );
				return NULL;
			}

			/* Get the field name */
			fieldName = newJavaLangStringFromUtf8( (CONSTANT_Utf8_info*) cf->constant_pool[f->name_index] );
			addToList( fieldName, &objectPool );

			/* Ensure that objref == will work by using String.intern() */
			fieldName = Java_java_lang_String_intern( env, fieldName );

			/* Convert the field descriptor into a Class object */
			if( !(fieldType = descriptorToClass( env, cf, f->descriptor_index )) )
			{
				env->retcode = newCompleteException( env, InternalError, "Error processing descriptor for field" );
				return NULL;
			}

			/* Create a new Field object */
			field = newObjectRefAndObject( fieldcf );
			addToList( field, &objectPool );

#if JDK_VERSION < 140
			/* Create a new stack frame for <init>() */
			forceExecuteJavaMethod( field, env, "<init>", "()V" );
#else
			/* Setup the arguments to Field.<init>() */
			argc = 5;
			argvt[0] = T_OBJREF; argv[0] = (u4) cf->classobj;		/* clazz */
			argvt[1] = T_OBJREF; argv[1] = (u4) fieldName;			/* name */
			argvt[2] = T_OBJREF; argv[2] = (u4) fieldType;			/* type */
			argvt[3] = T_INT;    argv[3] = (u4) f->access_flags;	/* modifiers */
			argvt[4] = T_INT;    argv[4] = (u4) 0;					/* slot */

			/* Create a new stack frame for <init>(Class,String,Class,int,int) */
			forceExecuteJavaMethodWithArgs( field, env, "<init>", "(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;II)V", argc, argv, argvt );
#endif /* JDK 1.4 */

			/* Make execution stop after <init>() completes */
			env->fp->stopOnPop = 1;
			retcode = 0;

			/* Execute <init>() */
			while( !retcode )
				retcode = executeInstruction( env );

			/* Make sure we stopped because of stop-on-pop - exception will be on stack */
			if( retcode != 1 )
				return NULL;

#if JDK_VERSION < 140
			field->data[clazzOffset] = (u4) cf->classobj;
			field->data[nameOffset]  = (u4) fieldName;
			field->data[typeOffset]  = (u4) fieldType;
#endif /* !JDK 1.4 */

			/* Store the new Constructor object in the array to be returned */
			ret->data[count++] = (u4) field;
		}

#if JDK_VERSION < 140
		if( declaredFlag )
			break;
	}
#endif /* !JDK 1.4 */

	return ret;
}


#if JDK_VERSION < 140
/**
 * ??Returns a single field of this class.  If the field does not exist,
 * returns null.??
 *
 * @since  JDK1.1?  (not in JDK 1.4)
 *
 * private native java.lang.reflect.Field getField0( String name, int arg2 );
 */
jobject Java_java_lang_Class_getField0( JNIEnv* env, jobject self, jobject fieldName, jint declaredFlag )
{
	ClassFile* cf = (ClassFile*) self->data[self->cf->object_size];    /* last field is cf */

	ObjRef *field, *fieldType;
	field_info *f = NULL;
	int32 clazzOffset, nameOffset, typeOffset;
	jboolean publicOnly = declaredFlag ? JNI_FALSE : JNI_TRUE;
	u2 index;
	char name[255];
	ClassFile* fieldcf;
	int retcode, found = 0;

	if( !extractFromJavaLangString( name, 255, fieldName ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Unable to extract field name" );
		return NULL;
	}

	/* Load java.lang.reflect.Constructor */
	if( !(fieldcf = getClass( env, "java/lang/reflect/Field" )) )
		return NULL;  /* exception will be on stack */

	clazzOffset = getFieldOffset( fieldcf, "clazz", "Ljava/lang/Class;" );
	nameOffset  = getFieldOffset( fieldcf, "name", "Ljava/lang/String;" );
	typeOffset  = getFieldOffset( fieldcf, "type", "Ljava/lang/Class;" );

	if( (clazzOffset == -1) || (nameOffset == -1) || (typeOffset == -1) )
	{
		env->retcode = newCompleteException( env, InternalError, "java.lang.reflect.Field does not have correct fields" );
		return NULL;
	}

	for( ; cf; cf = cf->super )
	{
		/* For each field check its name to see if it matched the one we are looking for */
		for( index = 0; index < cf->fields_count; index++ )
		{
			/* Lookup the field_info entry */
			f = cf->fields[index];

			/* If only getting public fields, skip non-public ones */
			if( publicOnly && ((cf->fields[index]->access_flags & ACC_PUBLIC) == 0) )
				continue;

			/* Is this the field we want? */
			if( checkSimpleName( cf, f->name_index, name ) )
			{
				found = 1;
				break;
			}
		}

		if( found || declaredFlag )
			break;
	}

	/* Did we find anything? If not, raise a NoSuchFieldException */
	if( !found )
	{
		env->retcode = newCompleteException( env, NoSuchFieldException, name );
		return NULL;
	}


	/* Convert the field descriptor into a Class object */
	if( !(fieldType = descriptorToClass( env, cf, f->descriptor_index )) )
	{
		env->retcode = newCompleteException( env, InternalError, "Error processing descriptor for field" );
		return NULL;
	}

	/* Create a new Field object */
	field = newObjectRefAndObject( fieldcf );
	addToList( field, &objectPool );

	/* Create a new stack frame for <init>() */
	forceExecuteJavaMethod( field, env, "<init>", "()V" );

	/* Make execution stop after <init>() completes */
	env->fp->stopOnPop = 1;
	retcode = 0;

	/* Execute <init>() */
	while( !retcode )
		retcode = executeInstruction( env );

	/* Make sure we stopped because of stop-on-pop - exception will be on stack */
	if( retcode != 1 )
		return NULL;

	field->data[clazzOffset] = (u4) cf->classobj;
	field->data[nameOffset]  = (u4) fieldName;
	field->data[typeOffset]  = (u4) fieldType;

	return field;
}
#endif /* !JDK 1.4 */


/**
 * Determines the parent interfaces of this class. 
 * 
 * If this is a Java class, then it returns an array of all the interfaces
 * implemented by the class. The order is that in which they were declared in
 * the implements clause.
 * 
 * If this is an interface, then it returns an array of all interfaces extended
 * by the interface. The order is that in which they were declared in the
 * extends clause.
 * 
 * If the class or interface implements no interfaces, an array of length 0 is
 * returned. 
 *
 * @since   JDK1.0
 *
 * public native Class[] getInterfaces()
 */
jobject Java_java_lang_Class_getInterfaces( JNIEnv *env, jobject self )
{
	ClassFile* cf = (ClassFile*) self->data[self->cf->object_size];    /* last field is cf */
	ObjRef* array = newObjectRefAndArray( T_OBJREF, cf->interfaces_count );
	u2 n;

	addToList( array, &arrayPool );

	for( n = 0; n < cf->interfaces_count; n++ )
		array->data[n] = (u4) cf->interfaces[n]->cf->classobj;

	return array;
}


/**
 * Returns the modifiers (public, protected, private, final, and interface) for
 * this class.
 *
 * This is a non-static version of java.io.ObjectStreamClass.getClassAccess()
 * and/or the JDK 1.4 method sun.reflect.Reflection.getClassAccessFlags().
 *
 * @see    table 4.1 of The Java Virtual Machine Specification.
 *
 * @since  JDK1.1
 *
 * public native int getModifiers();
 */
jint Java_java_lang_Class_getModifiers( JNIEnv *env, jobject self )
{
	ClassFile* cf = (ClassFile*) self->data[self->cf->object_size];    /* last field is cf */
	return cf ? cf->access_flags : 0;
}


/**
 * Returns the fully-qualified name of the type (class, interface,
 * array, or primitive) represented by this Class object, as a String.
 *
 * @since   JDK1.0
 *
 * public native String getName();
 */
jstring Java_java_lang_Class_getName( JNIEnv* env, jobject self )
{
	ObjRef* str;
	ClassFile* cf;
	const char* key;
	char buf[255];
	int len;

	cf = (ClassFile*) self->data[self->cf->object_size];	/* last field is cf */

	/* Is it a primitive or array type? */
	if( !cf )
	{
		/* Do a reverse lookup of primitive types */
		u1 kind = getPrimitiveTypeForClass( self );

		/* Convert type index to String */
		switch( kind )
		{
		case T_BOOLEAN: str = newJavaLangStringFromBytes( "boolean", 7 ); break;
		case T_CHAR:	str = newJavaLangStringFromBytes( "char", 4 ); break;
		case T_FLOAT:	str = newJavaLangStringFromBytes( "float", 5 ); break;
		case T_DOUBLE:	str = newJavaLangStringFromBytes( "double", 6 ); break;
		case T_BYTE:	str = newJavaLangStringFromBytes( "byte", 4 ); break;
		case T_SHORT:	str = newJavaLangStringFromBytes( "short", 5 ); break;
		case T_INT:		str = newJavaLangStringFromBytes( "int", 3 ); break;
		case T_LONG:	str = newJavaLangStringFromBytes( "long", 4 ); break;
		default:
			/* Do a reverse lookup of the array Class'es */
			if( (key = Hashtable_findValue( arrayClasses, self )) )
				str = newJavaLangStringFromBytes( key, strlen(key) );
			else
				str = newJavaLangStringFromBytes( "<unknown>", 8 );
			break;
		}

		addToList( str, &objectPool );
		return str;
	}

	/* Just return the name of the class */
	len = sprintClassName( buf, cf, cf->this_class );
	str = newJavaLangStringFromBytes( buf, len );
	addToList( str, &objectPool );
	return str;
}


/*
 * Return the Virtual Machine's Class object for the named primitive type.
 *
 * static native java.lang.Class getPrimitiveClass(String name);
 */
jclass Java_java_lang_Class_getPrimitiveClass( JNIEnv* env, jclass cls, jstring name )
{
	char classname[255];

	if( extractFromJavaLangString( classname, 255, name ) )
	{
		if( strcmp( classname, "boolean" ) == 0 )
			return getClassForPrimitiveType( T_BOOLEAN );
		else if( strcmp( classname, "char" ) == 0 )
			return getClassForPrimitiveType( T_CHAR );
		else if( strcmp( classname, "float" ) == 0 )
			return getClassForPrimitiveType( T_FLOAT );
		else if( strcmp( classname, "double" ) == 0 )
			return getClassForPrimitiveType( T_DOUBLE );
		else if( strcmp( classname, "byte" ) == 0 )
			return getClassForPrimitiveType( T_BYTE );
		else if( strcmp( classname, "short" ) == 0 )
			return getClassForPrimitiveType( T_SHORT );
		else if( strcmp( classname, "int" ) == 0 )
			return getClassForPrimitiveType( T_INT );
		else if( strcmp( classname, "long" ) == 0 )
			return getClassForPrimitiveType( T_LONG );

		printf( "ERROR: Class.getPrimitiveClass( \"%s\" ) - unrecognised type name!\n", classname );
	}

	env->retcode = newCompleteException( env, InternalError, "Class.getPrimitiveClass() failed" );
	return NULL;
}


/**
 * If this is anything other than the class Object, then the superclass is
 * returned, otherwise null is returned.
 *
 * @since   JDK1.0
 *
 * public native java.lang.Class getSuperclass();
 */
jclass Java_java_lang_Class_getSuperclass( JNIEnv* env, jobject self )
{
	ClassFile* cf = (ClassFile*) self->data[self->cf->object_size];	/* last field is cf */

	if( !cf || !cf->super || (cf->access_flags & ACC_INTERFACE) )
		return NULL;

	return cf->super->classobj;
}


/**
 * If this Class object represents an array type, returns true, otherwise
 * returns false.
 *
 * @since   JDK1.1
 *
 * public native boolean isArray();
 */
jboolean Java_java_lang_Class_isArray( JNIEnv* env, jobject self )
{
	/* Is this an array type? */
	return isArrayType( self ) ? JNI_TRUE : JNI_FALSE;
}


/**
 * If this is a class or interface, then it determines if this class is either
 * the same as, or is a super-class or super-interface of, the class cls.
 * If this is a primitive type then cls must be of the same type.
 *
 * @exception NullPointerException if the specified Class parameter is null.
 * @since   JDK1.1
 *
 * public native boolean isAssignableFrom( java.lang.Class cls )
 */
jboolean Java_java_lang_Class_isAssignableFrom( JNIEnv* env, jobject self, jclass cls )
{
	ClassFile* selfcf = (ClassFile*) self->data[self->cf->object_size];	/* last field is cf */
	ClassFile* clscf  = (ClassFile*) cls->data[cls->cf->object_size];	/* last field is cf */

	if( subclassOf( clscf, selfcf ) )
		return JNI_TRUE;

	return JNI_FALSE;
}


/**
 * Returns true if this Class object represents an interface type, false
 * otherwise.
 *
 * @since   JDK1.0
 *
 * public native boolean isInterface();
 */
jboolean Java_java_lang_Class_isInterface( JNIEnv* env, jobject self )
{
	ClassFile* cf = (ClassFile*) self->data[self->cf->object_size];	/* last field is cf */
	return (cf && ((cf->access_flags & ACC_INTERFACE) != 0)) ? JNI_TRUE :  JNI_FALSE;
}


/**
 * If this Class object represents an primitive type, returns true, otherwise
 * returns false.
 *
 * @since   JDK1.1
 *
 * public native boolean isPrimitive();
 */
jboolean Java_java_lang_Class_isPrimitive( JNIEnv* env, jobject self )
{
	ClassFile* cf = (ClassFile*) self->data[self->cf->object_size];	/* last field is cf */
	int n;

	/* Non-null cf indicates a real class */
	if( cf )
		return JNI_FALSE;

	/* Check all our primitive class objects */
	for( n = 0; n < T_LAST; n++ )
	{
		/* If we match, it's primitive */
		if( primitiveClass[n] && (primitiveClass[n] == self) )
			return JNI_TRUE;
	}

	/* No match */
	return JNI_FALSE;
}


#if JDK_VERSION >= 140
/**
 * private static native void registerNatives();
 */
void Java_java_lang_Class_registerNatives( JNIEnv* env, jclass staticself )
{
	/* Do Nothing! */
}
#endif /* JDK_VERSION 1.4.0 */
