/******************************************************************************
 *
 * Filename: java_lang_ClassLoader.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for java.lang.ClassLoader.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "../engine/config.h"

#include "java_lang_ClassLoader.h"

#include "../engine/util.h"
#include "../engine/openfiles.h"
#include "../engine/globals.h"
#include "../engine/zipload.h"
#include "../engine/classpath.h"
#include "../engine/memory.h"

#if JDK_VERSION >= 140
#include "../engine/loader.h"
#endif /* JDK 1.4 */


/* JDK 1.4 & JDK 1.1.7A
-#  private native void init();
+#  private native java.lang.Class findSystemClass0(java.lang.String);
+#  private static native java.io.InputStream getSystemResourceAsStream0(java.lang.String);
 #  private static native java.lang.String getSystemResourceAsName0(java.lang.String);
-1.4 private static native void registerNatives();
 #   private native java.lang.Class defineClass0(String, byte[], int, int);
 1.4 private native java.lang.Class defineClass0(String, byte[], int, int, java.security.ProtectionDomain);
     private native void resolveClass0(java.lang.Class);
 1.4 private native java.lang.Class findBootstrapClass(java.lang.String)
		 throws java.lang.ClassNotFoundException;
 1.4 protected final native java.lang.Class findLoadedClass(java.lang.String);
 1.4 private static native java.lang.AssertionStatusDirectives retrieveDirectives();
*/


#if JDK_VERSION < 140
/**
 * Initializes the Class loader.
 *
 * void init();
 */
void Java_java_lang_ClassLoader_init( JNIEnv* env, jobject self )
{
	/* ### doesn't need to do anything? */
}


/**
 * Searches the classpath and returns a stream onto the specified resource.
 * Returns null if the resource cannot be found.
 *
 * @since   JDK1.1
 *
 * private static native java.io.InputStream getSystemResourceAsStream0( java.lang.String name );
 */
jobject Java_java_lang_ClassLoader_getSystemResourceAsStream0( JNIEnv* env, jclass cls, jobject str )
{
	char filename[255];
	ObjRef* objref;
	ClassFile* cf;

	if( extractFromJavaLangString( filename, 255, str ) )
	{
		int n, index = findFreeFD();
		OpenFile_Type kind = oft_Normal;
		long length = -1;
		FILE* f = NULL;

		if( !(cf = getClass( env, "java/io/FileInputStream" )) )
			return NULL; /* getClass creates an exception for us */

		/* Implementation of CLASSPATH */
		for( n = 0; !f && (n < classpathc); n++ )
		{
			if( classpathv[n].kind == cpt_Directory )
			{
				/* Prepend classpath entry onto filename */
				char prehfname[512];
				char* pbuf = prehfname;
				strcpy( pbuf, classpathv[n].path ); pbuf += strlen( classpathv[n].path );
				strcpy( pbuf, FILE_SEPARATOR ); pbuf += strlen( FILE_SEPARATOR );
				strcpy( pbuf, filename ); pbuf += strlen( filename );
				*pbuf = '\0';

				/* Try normal directory */
				f = fopen( prehfname, "rb" );
			}
#ifndef USES_UNZIPPED_CLASSES_ONLY
			else if( classpathv[n].kind == cpt_Archive )
			{
				/* Try an archive */
				f = fopenZippedResource( &classpathv[n], filename, &length );

				if( f )
					kind = oft_Archive;
			}
#endif /* USES_UNZIPPED_CLASSES_ONLY */
		}

		if( !f )
		{
			if( verbose )
				printf( "WARNING: Failed to load system resource \"%s\"\n", filename );

			return NULL;
		}

		/* Fill in the OpenFile_Entry */
		openfilesv[index] = javavm_malloc( sizeof(OpenFile_Entry) );
		openfilesv[index]->f         = f;
		openfilesv[index]->kind      = kind;
		openfilesv[index]->available = length;

		/* Create a new FileInputStream object */
		objref = newObjectRefAndObject( cf );
		addToList( objref, &objectPool );

		/* Fill in the FileDescriptor object */
		objref->data[0] = (u4) newJavaIoFileDescriptor( index );
		addToList( (ObjRef*) objref->data[0], &objectPool );

		if( verbose )
			printf( "Opened system resource \"%s\"\n", filename );

		return objref;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
	return NULL;
}
#endif /* JDK 1.1 */


/**
 * Searched the classpath for the named class and loads it without using a
 * ClassLoader.  Raises a ClassNotFoundException if there is no such definition
 * or a NoClassDefFoundError if the class cannot be found.
 *
 * @since      JDK1.0
 *
 * java.lang.Class findSystemClass0( java.lang.String name )
 *     throws ClassNotFoundException;
 */
#if JDK_VERSION < 140
jclass Java_java_lang_ClassLoader_findSystemClass0( JNIEnv* env, jobject self, jobject str )
#else
/* !!!Should search boot classpath only? !!! */
jclass Java_java_lang_ClassLoader_findBootstrapClass( JNIEnv* env, jobject self, jobject str )
#endif /* JDK 1.4 */
{
	char classname[255];

	if( extractFromJavaLangString( classname, 255, str ) )
	{
		ClassFile* cf;
		u4 n;

		for( n = 0; n < strlen( classname ); n++ )
		{
			if( classname[n] == '.' )
				classname[n] = '/';
		}

		if( !(cf = getClass( env, classname )) )
		{
			/* all exceptions get translated into ClassNotFoundExceptions */
			env->retcode = newCompleteException( env, ClassNotFoundException, classname );
			return NULL;
		}

		return cf->classobj;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
	return NULL;
}


#if JDK_VERSION >= 140
/**
 * ???Returns a reference to the already loaded Class 'name'.  If the class
 * hasn't been loaded, returns NULL.???
 *
 * protected final native java.lang.Class findLoadedClass( java.lang.String name );
 */
jclass Java_java_lang_ClassLoader_findLoadedClass( JNIEnv* env, jobject self, jobject name )
{
	char classname[255];

	if( extractFromJavaLangString( classname, 255, name ) )
	{
		ClassFile* cf;
		u4 n;

		/* Convert class name to file name */
		for( n = 0; n < strlen( classname ); n++ )
		{
			if( classname[n] == '.' )
				classname[n] = '/';
		}

		/* Look for class and return Class object if it class is loaded */
		cf = checkLoaded( classname );
		return cf ? cf->classobj : NULL;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
	return NULL;
}


/**
 * private static native void registerNatives();
 */
void Java_java_lang_ClassLoader_registerNatives( JNIEnv* env, jclass staticself )
{
	/* Do Nothing! */
}
#endif /* JDK_VERSION 1.4.0 */


/******************************************************************************
 *
 * java.lang.ClassLoader$NativeLibrary
 *
 *****************************************************************************/

/* JDK 1.4 (doesn't exist in JDK 1.1.7A)
-1.4 native void load(java.lang.String);
 1.4 native long find(java.lang.String);
 1.4 native void unload();
*/


/**
 * !!!Pretends to load the named native library - in reality we statically link
 * against it!!!
 *
 * native void load(java.lang.String);
 */
void Java_java_lang_ClassLoader_NativeLibrary_load( JNIEnv* env, jobject self, jobject str )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, str ) )
	{
		int32 handleOffset;

		printf( "WARNING: Attempt to load native library \"%s\"\n", filename );

		handleOffset = getFieldOffset( self->cf, "handle", "J" );

		if( handleOffset < 0 )
		{
			env->retcode = newCompleteException( env, InternalError, "java.lang.ClassLoader$NativeLibrary does not have a 'handle' field" );
			return;
		}

		self->data[handleOffset] = 1;
		return;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
}
