/******************************************************************************
 *
 * Filename: java_lang_System.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  25th March 2002
 *
 * Purpose:  Native method implementations for java.lang.System.
 *
 * (c) Copyright R.H.Thornburrow, 2002-3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <sys/timeb.h>   /* for ftime() stuff */
#include <direct.h>		 /* for _getcwd() */
#define GETCWD  _getcwd
#else
#include <sys/time.h>	 /* for gettimeofday() */
#include <unistd.h>		 /* for _getcwd() */
#define GETCWD  getcwd
#endif

#include "../engine/config.h"
#include "../engine/disasm.h"
#include "../engine/util.h"
#include "../engine/classpath.h"
#include "../engine/globals.h"

#include "java_lang_System.h"

#include "java_lang_Object.h"


/* JDK 1.4
-1.4 private static native void registerNatives();
+    private static native void setIn0(java.io.InputStream);
+    private static native void setOut0(java.io.PrintStream);
+    private static native void setErr0(java.io.PrintStream);
+    public static native long currentTimeMillis();
+    public static native void arraycopy(java.lang.Object, int, java.lang.Object, int, int);
+1.1 public static native int identityHashCode(java.lang.Object);
+    private static native java.util.Properties initProperties(java.util.Properties);
+    public static native java.lang.String mapLibraryName(java.lang.String);
*/

/**
 * Copies a portion of one array to the specified position in another array.
 * 'length' elements of the array 'src', starting at index 'src_position' are
 * copied into the array 'dst' starting at index 'dst_position'.
 * 
 * If 'src' and 'dst' are the same array, copying is done as though the
 * elements were copied via a temporary array.
 *
 * An ArrayStoreException is throw if any of the following apply: 
 * 
 *  o Either 'src' or 'dst' is not an array
 *  o 'src' and 'dst' are arrays of differing primitive types
 *  o Either 'src' or 'dst' is an array of a reference type and the other is an
 *    array of a primitive type 
 * 
 * An ArrayIndexOutOfBoundsException is thrown if any of the following apply:
 *
 *  o 'src_position', 'dst_position' or 'length' are negative
 *  o 'src_position + length' is greater than 'src.length'
 *  o 'dst_postion + length' is greater than 'dst.length'
 * 
 * Finally, if during copying, any element of the source array cannot be
 * converted to the type of the destination array, an ArrayStoreException will
 * again be thrown, but this time the destination array may have been modified.
 *
 * @since      JDK1.0
 *
 * public static native void arraycopy(Object src, int src_position,
 *                                     Object dst, int dst_position,
 *                                     int length);
 */
void Java_java_lang_System_arraycopy( JNIEnv* env, jclass staticself,
									  jobject src, jint src_position,
									  jobject dst, jint dst_position, jint length)
{
	int size = 0;

	/* Array types must be either same primitive types or both reference types */
	if( src->type != dst->type )
	{
#ifdef _DEBUG
		char stbuf[32], dtbuf[32];
		sprintTypeName( stbuf, src->type );
		sprintTypeName( dtbuf, dst->type );
		printf( "WARNING: Cannot copy array of %s to array of %s\n", stbuf, dtbuf );
#endif
		env->retcode = newCompleteException( env, ArrayStoreException, "Cannot copy between different array types" );
		return;
	}

	/* Find the size of each array element */
	size = (int) sizeOfJavaType( src->type );

	/* If array has an invalid type code bail out now */
	if( !size )
	{
#ifdef _DEBUG
		{
			char buf[32];
			sprintTypeName( buf, src->type );
			printf( "ERROR: attempting to copy array of %s\n", buf );
		}
#endif
		env->retcode = newCompleteException( env, InternalError, "Bad type in arraycopy" );
		return;
	}

	/* If we are not asked to copy any elements, then we've finished */
	if( length == 0 )
		return;

	/* Verify that the position indices and length are valid and that both
	   arrays are long enough */
	if( (src_position >= 0) && (dst_position >= 0) && (length > 0)
		&& (src_position + length <= (jint) src->length)
		&& (dst_position + length <= (jint) dst->length) )
	{
		u1* srcdata = (u1*) src->data;
		u1* dstdata = (u1*) dst->data;
		u1* srcpos = &srcdata[src_position*size];
		u1* dstpos = &dstdata[dst_position*size];
		int32 m;
		/*char sbuf[STRING_MAX], dbuf[STRING_MAX], rbuf[STRING_MAX];
		int32 n;
		
		if( verbose ) {
			int32 srclen = ( src->length < STRING_MAX ) ? src->length : (STRING_MAX - 1);
			int32 dstlen = ( dst->length < STRING_MAX ) ? dst->length : (STRING_MAX - 1);
			if( src->type == T_CHAR ) {
				w2cstrncpy( sbuf, (const u2*) srcdata, srclen ); sbuf[srclen] = '\0';
				w2cstrncpy( dbuf, (const u2*) dstdata, dstlen ); dbuf[dstlen] = '\0';
			} else if( src->type == T_BYTE ) {
				sbuf[srclen]='\0'; for( n = 0; n < srclen; n++ )
				{ if( srcdata[n] < 32 ) sbuf[n] = '.'; else sbuf[n] = srcdata[n]; }
				dbuf[dstlen]='\0'; for( n = 0; n < dstlen; n++ )
				{ if( dstdata[n] < 32 ) dbuf[n] = '.'; else dbuf[n] = dstdata[n]; }
			}
		}*/

		/*if( ((srcpos > dstpos) && (srcpos < &dstpos[length*size]))
				|| ((&srcpos[length*size] > dstpos) && (&srcpos[length*size] < &dstpos[length*size])) )
			printf( "WARNING: System.arrayCopy(): Buffers overlap!!!\n" );*/

		/* If same array and destination if lower than source do a forwards
		   copy to allow for overlap, otherwise do a backwards copy, which is
		   faster, to cope with overlap the other way */
		if( (src == dst) && (dst_position < src_position) )
		{
			/* This does a forwards memcpy */
			for( m = 0; m < length*size; m++ )
				dstpos[m] = srcpos[m];
		}
		else
		{
			/* This does a backwards memcpy to cope with overlapping buffers */
			for( m = length*size - 1; m >= 0; m-- )
				dstpos[m] = srcpos[m];
		}

		/*
		if( verbose ) {
			int32 dstlen = ( dst->length < STRING_MAX ) ? dst->length : (STRING_MAX - 1);
			if( src->type == T_CHAR ) {
				w2cstrncpy( rbuf, (const u2*) dstdata, dstlen ); rbuf[dstlen] = '\0';
				printf( "Copied %d bytes from \"%s\" (pos %d, len %d), into \"%s\" (pos %d, len %d) giving \"%s\"\n", length, sbuf, src_position, src->length, dbuf, dst_position, dst->length, rbuf );
			} else if( src->type == T_BYTE ) {
				rbuf[dstlen]='\0'; for( n = 0; n < dstlen; n++ )
				{ if( dstdata[n] < 32 ) rbuf[n] = '.'; else rbuf[n] = dstdata[n]; }
				printf( "Copied %d bytes from \"%s\" (pos %d, len %d), into \"%s\" (pos %d, len %d) giving \"%s\"\n", length, sbuf, src_position, src->length, dbuf, dst_position, dst->length, rbuf );
			}
		}*/

		/*
		if( src->type == T_OBJREF )
		{
			for( n = 0; n < length; n++ )
				printf( "Copied src[%d] (0x%08x) as dst[%d] (0x%08x)\n", src_position+n, src->data[src_position+n], dst_position+n, dst->data[dst_position+n] );
		}*/

		return;
	}

	env->retcode = newCompleteException( env, ArrayIndexOutOfBoundsException, "" );
}


/**
 * Returns the current time (from the system clock) in milliseconds.  This is
 * measured from midnight, January 1, 1970 UTC.
 *
 * @since   JDK1.0
 *
 * public static native long currentTimeMillis();
 */
jlong Java_java_lang_System_currentTimeMillis( JNIEnv* env, jclass staticself )
{
	jlong secs, millis, ret;

#ifdef WIN32
	struct _timeb tbuf;

	_ftime( &tbuf );
	secs = tbuf.time;
	millis = tbuf.millitm;
#else
	struct timeval tbuf;

	gettimeofday( &tbuf, NULL );
	secs = tbuf.tv_sec;
	millis = tbuf.tv_usec / 1000;
#endif

    ret = (secs * 1000) + millis;

#ifdef GCC_OPTIMIZER_CLUDGE
    printf( "System.currentTimeMillis(): %.0f secs + %.0f millis = %.0f\n",
            (double) secs, (double) millis, (double) ret );
#endif /* GCC_OPTIMIZER_CLUDGE */

	return ret;
}


/**
 * Non-overridable version of Object.hashCode().  If a null reference is
 * passed, 0 is returned.
 *
 * @since   JDK1.1
 *
 * public static native int identityHashCode( java.lang.Object obj );
 */
jint Java_java_lang_System_identityHashCode( JNIEnv* env, jclass staticself, jobject obj )
{
	/* hash of null is 0 */
	if( !obj )
		return 0;

	/* Re-use Object.hashCode() */
	return Java_java_lang_Object_hashCode( env, obj );
}


/* Internal method */
int internal_System_addProperty( Frame* fp, ObjRef* props, const char* property, const char* value )
{
	Stack stack;
	int retcode;
	ObjRef *propObj, *valObj, *dummy;

	stack.ex = NULL;
	stack.fp = fp;
	stack.tail = NULL;

	propObj = newJavaLangStringFromBytes( property, strlen( property ) );
	addToList( propObj, &objectPool );
	PUSH_OBJREF( fp, propObj );

	valObj = newJavaLangStringFromBytes( value, strlen( value ) );
	addToList( valObj, &objectPool );
	PUSH_OBJREF( fp, valObj );

	retcode = forceExecuteVirtualJavaMethod( props, &stack, "put", "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;", 1 );

	while( !retcode )
		retcode = executeInstruction( &stack );

	if( retcode == 2 )
	{
		fp->ex = stack.ex ? stack.ex : (stack.fp ? stack.fp->ex : NULL );
		return retcode;
	}

	POP_OBJREF( fp, dummy );
	return 0;
}


/*
 * java.version					 = Java version number
 * java.vendor					 = Java vendor-specific string
 * java.vendor.url				 = Java vendor URL
 * java.runtime.name             = Java Runtime Classes name (1.4)
 * java.runtime.version			 = Java Runtime Classed version (1.4)
 * java.specification.name		 = Java Platform API Specification name
 * java.specification.vendor	 = Java Platform API Specification vendor-specific string
 * java.specification.version	 = Java Platform API Specification version number
 * java.vm.name					 = Java Virtual Machine name
 * java.vm.specification.name	 = Java Virtual Machine Specification name
 * java.vm.specification.vendor  = Java Virtual Machine Specification vendor-specific string
 * java.vm.specification.version = Java Virtual Machine Specification version number
 * java.vm.vendor				 = Java Virtual Machine vendor-specific string
 * java.vm.version				 = Java Virtual Machine version number
 * java.home					 = Java installation directory
 * java.class.version			 = Java class format version number
 * java.class.path				 = Java class path
 * os.name						 = Operating system name
 * os.arch						 = Operating system architecture
 * os.version					 = Operating system version</td></tr>
 * file.separator				 = File separator ("/" on UNIX)
 * path.separator				 = Path separator (":" on UNIX)
 * line.separator				 = Line separator ("\n" on UNIX)
 * user.name					 = User's account name
 * user.home					 = User's home directory
 * user.dir						 = User's current working directory
 * user.country					 = User's locale (GB, US, etc.) (1.4)
 * user.timezone				 = User's timezone (GMT, PST, etc.) (1.4)
 */

typedef enum
{
        pk_text, pk_envvar, pk_pwd, pk_classpath, pk_libpath
} PropKind;

typedef struct _SystemProperty
{
	const char* name;
    PropKind kind;
	const char* value;
	const char* defval;
} SystemProperty;

SystemProperty systemProperties[] =
{
	{ "java.version",                    pk_text, JAVA_SPEC_VERSION, NULL },
	{ "java.vendor",                     pk_text, "Taurus Software", NULL },
	{ "java.vendor.url",                 pk_text, "http://www.tsac.co.uk", NULL },
#if JDK_VERSION > 120
	{ "java.runtime.name",               pk_text, "Java(TM) 2 Runtime Environment, Standard Edition", NULL },
	{ "java.runtime.version",            pk_text, JAVA_SPEC_VERSION, NULL },
#endif /* JAVA 2 */
	{ "java.specification.name",		 pk_text, "Java Platform API Specification", NULL },
	{ "java.specification.vendor",	     pk_text, "Sun Microsystems Inc.", NULL },
	{ "java.specification.version",	     pk_text, JAVA_SPEC_VERSION, NULL },
	{ "java.vm.name",					 pk_text, "Taurus JavaVM", NULL },
	{ "java.vm.specification.name",	     pk_text, "Java Virtual Machine Specification", NULL },
	{ "java.vm.specification.vendor",    pk_text, "Sun Microsystems Inc.", NULL },
	{ "java.vm.specification.version",   pk_text, "1.0", NULL },
	{ "java.vm.vendor",				     pk_text, "Taurus Software", NULL },
	{ "java.vm.version",				 pk_text, JAVAVM_VERSION, NULL },
#ifdef WIN32
	{ "java.home",           pk_envvar, "TJVM_HOME", "D:\\Projects\\javavm" },
#else
#ifdef USES_ABSOLUTE_FILENAMES
	{ "java.home",           pk_envvar, "TJVM_HOME", ABSOLUTE_FILENAME_ROOT },
#else
	{ "java.home",           pk_envvar, "TJVM_HOME", "/export/Projects/javavm" },
#endif
#endif /* WIN32 */
	{ "java.class.path",     pk_classpath, NULL, NULL },
	{ "java.class.version",  pk_text,   JAVA_CLASS_VERSION, NULL },
#if JDK_VERSION >= 140
	{ "java.library.path",   pk_libpath, NULL, NULL },
	{ "sun.boot.class.path", pk_classpath, "system", NULL },
	{ "sun.boot.library.path", pk_libpath, "system", NULL },
#endif /* JDK 1.4 */
	{ "awt.toolkit",         pk_text,   "sun.awt.windows.WToolkit", NULL },
	{ "file.encoding.pkg",   pk_text,   "sun.io", NULL },
#if JDK_VERSION >= 140
	{ "file.encoding",		 pk_text,	"Cp1252", NULL },
#endif /* JDK 1.4 */
	{ "line.separator",      pk_text,   LINE_SEPARATOR, NULL },
	{ "file.separator",      pk_text,   FILE_SEPARATOR, NULL },
	{ "os.arch",			 pk_text,	OS_ARCH, NULL },
	{ "os.name",			 pk_text,	OS_NAME, NULL },
	{ "path.separator",      pk_text,   PATH_SEPARATOR, NULL },
	{ "user.dir",            pk_pwd,	NULL, NULL },
#ifdef WIN32
	{ "user.home",           pk_envvar, "HOME", "D:\\Users\\Robert" },
#else
#ifdef USES_ABSOLUTE_FILENAMES
	/*{ "user.dir",            pk_envvar, "PWD", "." },*/
	{ "user.home",           pk_envvar, "HOME", ABSOLUTE_FILENAME_ROOT },
#else
	{ "user.home",           pk_envvar, "HOME", "/home/robert" },
#endif
#endif /* WIN32 */
	{ "user.name",           pk_envvar, "USER", "robert" },
#if JDK_VERSION >= 140
	{ "user.country",		 pk_text,	"GB", NULL },
	{ "user.timezone",		 pk_text,	"GMT", NULL },
#endif /* JDK 1.4 */
	{ NULL, pk_text, NULL, NULL }
};


/**
 * Sets up the system properties (read via System.getProperty()) which describe
 * a number of things about the Virtual Machine environment.  These include the
 * name of the platform we are running on and the file and path separators to
 * use (these are all defined in config.h).  Uses the function
 * internal_System_addProperty() to add each property from the above table to
 * the Properties object 'props'.
 *
 * private static native Properties initProperties(Properties props);
 */
jobject Java_java_lang_System_initProperties( JNIEnv* env, jclass staticself, jobject props )
{
	SystemProperty* add;
	env->retcode = 0;

	for( add = &systemProperties[0]; !env->retcode && add->name; add++ )
	{
		switch( add->kind )
		{
		case pk_envvar:
			{
			const char *value;

			if( (value = getenv( add->value )) )
			{
				/*printf( "ENVIRONMENT: %s = \"%s\"\n", add->name, value );*/
				env->retcode = internal_System_addProperty( env->fp, props, add->name, value );
			}
			else
				env->retcode = internal_System_addProperty( env->fp, props, add->name, add->defval );
			}
			break;
		case pk_pwd:
			{
			char value[512];
			GETCWD( value, 512 );
			/*printf( "user.dir = %s\n", value );*/
			env->retcode = internal_System_addProperty( env->fp, props, add->name, strdup(value) );
			}
			break;
		case pk_classpath:
			{
			char value[1024], *buf = value;
			int n, flag = 0;

			for( n = 0; n < classpathc; n++ )
			{
#if JDK_VERSION >= 140
				/* Only include if isSystem and we are building system
				   classpath or !isSystem and we are building user classpath */
				if( classpathv[n].isSystem == (add->value ? 1 : 0) ) {
#endif /* JDK 1.4 */
					if( flag ) buf += sprintf( buf, PATH_SEPARATOR );
					buf += sprintf( buf, classpathv[n].path );
					flag = 1;
#if JDK_VERSION >= 140
				}
#endif /* JDK 1.4 */
			}
			/*printf( "%s = %s\n", add->name, value );*/
			env->retcode = internal_System_addProperty( env->fp, props, add->name, strdup(value) );
			}
			break;
		case pk_libpath:
			{
			char value[1024], *buf = value;
			int n, flag = 0;

			for( n = 0; n < classpathc; n++ )
			{
#if JDK_VERSION >= 140
				/* Only include if isSystem and we are building system
				   classpath or !isSystem and we are building user lib path */
				if( classpathv[n].isSystem == (add->value ? 1 : 0) ) {
#endif /* JDK 1.4 */
					if( classpathv[n].kind == cpt_Directory ) {
						if( flag ) buf += sprintf( buf, PATH_SEPARATOR );
						buf += sprintf( buf, classpathv[n].path );
						flag = 1;
					} /* if directory */
#if JDK_VERSION >= 140
				} /* system/user library path */
#endif /* JDK 1.4 */
			}
			/*printf( "%s = %s\n", add->name, value );*/
			env->retcode = internal_System_addProperty( env->fp, props, add->name, strdup(value) );
			}
			break;
		case pk_text:
		default:
			env->retcode = internal_System_addProperty( env->fp, props, add->name, add->value );
	    }
	}

	return props;
}


#if JDK_VERSION >= 140
/**
 * Takes a library name and maps it to a library filename on the host platform.
 * 
 * For most platforms this is as simple as appending ".dll" or ".so" to 'name'.
 *
 * public static native java.lang.String mapLibraryName( String name );
 */
jobject Java_java_lang_System_mapLibraryName( JNIEnv* env, jclass staticself, jobject name )
{
	char library[255];

	if( extractFromJavaLangString( library, 255, name ) )
	{
		jobject str;
		char buf[512];

		sprintf( buf, "%s.%s", library, LIBRARY_EXTENSION );
		str = newJavaLangStringFromBytes( buf, strlen( buf ) );
		addToList( str, &objectPool );

		/*printf( "System.mapLibraryName( \"%s\" ) returning \"%s\".\n", library, buf );*/

		return str;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to extract library name" );
	return NULL;
}


/**
 * private static native void registerNatives();
 */
void Java_java_lang_System_registerNatives( JNIEnv* env, jclass staticself )
{
	/* Do Nothing! */
}
#endif /* JDK_VERSION 1.4.0 */


/**
 * Initializes the InputStream object 'System.in' to be 'in'.  This is a final
 * field and so cannot be modified outside the constructor, but we need to
 * write to it from initSystemClass(), hence the existance of this method.
 *
 * private static native void setIn0(InputStream in);
 */
void Java_java_lang_System_setIn0( JNIEnv* env, jclass staticself, jobject in )
{
	ClassFile* cf = (ClassFile*) staticself->data[staticself->cf->object_size];	/* last field is cf */
	cf->static_data[0] = (u4) in;
}


/**
 * Initializes the PrintStream object 'System.out' to be 'out'.  This is a
 * final field and so cannot be modified outside the constructor, but we need
 * to write to it from initSystemClass(), hence the existance of this method.
 *
 * private static native void setOut0(PrintStream out);
 */
void Java_java_lang_System_setOut0( JNIEnv* env, jclass staticself, jobject out )
{
	ClassFile* cf = (ClassFile*) staticself->data[staticself->cf->object_size];	/* last field is cf */
	cf->static_data[1] = (u4) out;
}

/**
 * Initializes the PrintStream object 'System.err' to be 'err'.  This is a
 * final field and so cannot be modified outside the constructor, but we need
 * to write to it from initSystemClass(), hence the existance of this method.
 *
 * private static native void setErr0(PrintStream err);
 */
void Java_java_lang_System_setErr0( JNIEnv* env, jclass staticself, jobject err )
{
	ClassFile* cf = (ClassFile*) staticself->data[staticself->cf->object_size];	/* last field is cf */
	cf->static_data[2] = (u4) err;
}
