/******************************************************************************
 *
 * Filename: java_lang_reflect_Constructor.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  17th July 2003
 *
 * Purpose:  Native method implementations for java.lang.reflect.Constructor.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#include "../engine/config.h"

#include "java_lang_reflect_Constructor.h"

#include "../engine/engine.h"
#include "../engine/util.h"
#include "../engine/disasm.h"
#include "../engine/memory.h"

#include "java_lang_Class.h"


/* JDK 1.1.7A
 #   public native int getModifiers();
+#   public native java.lang.Object newInstance( java.lang.Object[] args );
*/

/* Fields are:
    java.lang.Class		clazz;
    int					slot;
    java.lang.Class[]	parameterTypes;
    java.lang.Class[]	exceptionTypes;
*/


#if JDK_VERSION < 140
/**
 * Create a new instance of the object for which this is a constructor.
 * The supplied 'args' should be uses as the arguments to the constructor call.
 *
 * public native java.lang.Object newInstance( java.lang.Object[] args );
 */
jobject Java_java_lang_reflect_Constructor_newInstance(	JNIEnv* env, jobject self, jobject args )
{
	jclass clazz;
	jobject parameterTypes;
	int32 clazzOffset, paramTypesOffset;
	ClassFile* cf;
	ObjRef* instance;
	u2 n, index;
	int retcode = 0, found, argc;
	char desc[255];
	u4 argv[32];
	u1 kind, argvt[32];

	/* Lookup the position of clazz and parameterTypes within the Constructor object */
	clazzOffset      = getFieldOffset( self->cf, "clazz", "Ljava/lang/Class;" );
	paramTypesOffset = getFieldOffset( self->cf, "parameterTypes", "[Ljava/lang/Class;" );

	if( (clazzOffset < 0) || (paramTypesOffset < 0) )
	{
		env->retcode = newCompleteException( env, InternalError, "java.lang.reflect.Constructor has unexpected fields" );
		return NULL;
	}

	/* Extract clazz - the Class to which the constructor belongs */
	clazz = (jclass) self->data[clazzOffset];

	/* Extract parameterTypes - the types of the constructors arguments */
	parameterTypes = (jobject) self->data[paramTypesOffset];

	/* Extract the ClassFile from the Class object, clazz */
	cf = (ClassFile*) clazz->data[clazz->cf->object_size];	/* last field is cf */

	/* Find all methods of cf with the name "<init>" (i.e. constructors) */
	for( index = 0; index < cf->methods_count; index++ )
	{
		if( cf->methods[index] && checkSimpleName( cf, cf->methods[index]->name_index, "<init>" ) )
		{
			/* Convert args of this constructor into a Class[] */
			jobject argTypes = descriptorToClassArray( env, cf, cf->methods[index]->descriptor_index );
			
			/* If conversion fails, abort */
			if( !argTypes )
			{
				env->retcode = newCompleteException( env, InternalError, "Error processing descriptor for constructor" );
				return NULL;
			}

			/* Number of arguments must be equal */
			if( argTypes->length != parameterTypes->length )
			{
				javavm_free( argTypes, argTypes->length * sizeof(ObjRef*) );
				argTypes = NULL;
				continue;
			}

			/* Assume matched and disprove it */
			found = 1;

			/* Match each argument type in turn - only one instance of each Class, so use == */
			for( n = 0; n < parameterTypes->length; n++ )
			{
				if( argTypes->data[n] != parameterTypes->data[n] )
				{
					found = 0;
					break;
				}
			}

			/* Done with arg types array - free it */
			javavm_free( argTypes, argTypes->length * sizeof(ObjRef*) );
			argTypes = NULL;

			/* Has a match been found - if so, stop */
			if( found )
				break;
		}
	}

	/* Did we find a match? */
	if( index >= cf->methods_count )
	{
		env->retcode = newCompleteException( env, InternalError, "No matching constructor" );
		return NULL;
	}

	/* DEBUG
	{
	char buf[1024];
	sprintMethodInfo( buf, cf->methods[index] );
	printf( "Matching constructor is %s at index %d\n", buf, cf->methods[index]->index );
	}
	DEBUG */

	/* Allow null argument array to mean no arguments */
	argc = args ? args->length : 0;

	/* Make sure caller has passed the right number of argument values */
	if( (u4) argc != parameterTypes->length )
	{
		env->retcode = newCompleteException( env, InternalError, "Incorrect number of arguments" );
		return NULL;
	}

	/* Create a new object of type clazz (cf is clazz's ClassFile) */
	instance = newObjectRefAndObject( cf );

	/* Build up the list of arguments to pass to forceExecuteJavaMethodWithArgs() */

	for( n = 0; n < argc; n++ )
	{
		/* Get the parameter type */
		ObjRef *paramClass = (ObjRef*) parameterTypes->data[n];
		ClassFile *paramcf = paramClass ? (ClassFile*) paramClass->data[paramClass->cf->object_size] : NULL;

		/* Is the arg a primitive or an array? */
		if( !paramcf )
		{
			/* Is the arg an array type? */
			if( isArrayType(paramClass) )
			{
				/* ### TODO: Implement checking of array types */

				/* Just pass the array as though it is an object */
				argv[n] = args->data[n];
				argvt[n] = T_OBJREF;
			}
			/* Is is a primitive type? */
			else if( (kind = getPrimitiveTypeForClass(paramClass)) != 0 )
			{
				/* Get the argument */
				ObjRef *arg = args ? (ObjRef*) args->data[n] : NULL;
				int32 offset = -1;

				/* Find the offset of the value field in the wrapper */
				switch( kind )
				{
				case T_BOOLEAN: offset = getFieldOffset( getClass( env, "java/lang/Boolean" ),   "value", "Z" ); break;
				case T_CHAR:    offset = getFieldOffset( getClass( env, "java/lang/Character" ), "value", "C" ); break;
				case T_FLOAT:   offset = getFieldOffset( getClass( env, "java/lang/Float" ),     "value", "F" ); break;
				case T_DOUBLE:  offset = getFieldOffset( getClass( env, "java/lang/Double" ),    "value", "D" ); break;
				case T_BYTE:    offset = getFieldOffset( getClass( env, "java/lang/Byte" ),      "value", "B" ); break;
				case T_SHORT:   offset = getFieldOffset( getClass( env, "java/lang/Short" ),     "value", "S" ); break;
				case T_INT:     offset = getFieldOffset( getClass( env, "java/lang/Integer" ),   "value", "I" ); break;
				case T_LONG:    offset = getFieldOffset( getClass( env, "java/lang/Long" ),      "value", "J" ); break;
				}

				/* If no value field, then bail out */
				if( (offset < 0) || !arg->data )
				{
					env->retcode = newCompleteException( env, InternalError, "Unable to convert wrapper to primitive" );
					return NULL;
				}

				/* ### TODO: Implement conversion of long/double args from objects to primitive types */
				if( (kind == T_DOUBLE) || (kind == T_LONG) )
				{
					env->retcode = newCompleteException( env, InternalError, "Constructor with long/double args not yet implemented" );
					return NULL;
				}

				/* Unwrap the value */
				argv[n] = arg->data[offset];
				argvt[n] = kind;
			}
			else  /* Not object, array or primitive! */
			{
				env->retcode = newCompleteException( env, InternalError, "Unknown Class for constructor parameter" );
				return NULL;
			}
		}
		else  /* object type */
		{
			/* Get the argument type */
			ClassFile *argcf = args ? (args->data[n] ? ((ObjRef*) args->data[n])->cf : NULL) : NULL;

			/* Do the arg and param types match? */
			if( argcf != paramcf )
			{
				env->retcode = newCompleteException( env, InternalError, "Argument/Parameter type mismatch in constructor call" );
				return NULL;
			}

			/* Must be a reference type (but not an array) */
			argv[n] = args->data[n];
			argvt[n] = T_OBJREF;
		}
	}

	/* Get the descriptor for the constructor we have chosen */
	if( !CHECKCONSTANT( cf, cf->methods[index]->descriptor_index, CONSTANT_Utf8 ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Cannot extract descriptor" );
		return NULL;
	}
	
	/* Convert utf8 descriptor to ASCII */
	sprintUtf8( desc, (CONSTANT_Utf8_info*) cf->constant_pool[cf->methods[index]->descriptor_index] );

	/* Execute the <init>() method on instance */
	forceExecuteJavaMethodWithArgs( instance, env, "<init>", desc, argc, argv, argvt );

	/* Make execution stops after <init>() completes */
	env->fp->stopOnPop = 1;

	/* Execute <init>() */
	while( !retcode )
		retcode = executeInstruction( env );

	/* Make sure we stop-on-pop'ed */
	if( retcode != 1 )
		return NULL;	/* Exception on stack */

	/* Return the newly created object */
	return instance;
}
#endif /* !JDK 1.4 */
