/******************************************************************************
 *
 * Filename: java_security_AccessController.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  12th April 2003
 *
 * Purpose:  Native method implementations for java.security.Privileged.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "../engine/config.h"

#include "java_security_AccessController.h"

#include "../engine/disasm.h"
#include "../engine/engine.h"
#include "../engine/util.h"
#include "../engine/native.h"
#include "../engine/globals.h"
#include "../engine/memory.h"


/* JDK 1.4 (class does not exist in JDK 1.1.7)
+1.2 public static native Object doPrivileged(java.security.PrivilegedAction);
+1.2 public static native Object doPrivileged(
		 java.security.PrivilegedAction, java.security.AccessControlContext);
+1.2 public static native Object doPrivileged(java.security.PrivilegedExceptionAction)
			 throws java.security.PrivilegedActionException;
+1.2 public static native Object doPrivileged(
		 java.security.PrivilegedExceptionAction, java.security.AccessControlContext)
			 throws java.security.PrivilegedActionException;
-1.2 private static native java.security.AccessControlContext getStackAccessControlContext();
 1.2 static native java.security.AccessControlContext getInheritedAccessControlContext();
*/

#if JDK_VERSION >= 140
/**
 * Executes the run() method of action as privileged code.  Checks only the
 * privileges of the immediate caller.
 *
 * We don't understand privileges yet, so we execute everything as privileged.
 * As we need the return value from run(), we use stopOnPop() to halt execution
 * when run() completes and have our own execution loop.
 *
 * private static native Object doPrivileged( java.security.PrivilegedAction action );
 */
jobject Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedAction_2(
	JNIEnv* env, jclass staticself, jobject action )
{
	ObjRef* ret;
	int retcode = 0;

	ClassFile* privAction = getClass( env, "java/security/PrivilegedAction" );

	if( !privAction ) {
		env->retcode = newCompleteException( env, NoClassDefFoundError, "java/security/PrivilegedAction" );
		return NULL;
	}

	/* Is 'action' a subclass of java.security.PrivilegedAction */
	if( !subclassOf( action->cf, privAction ) )
	{
		char buf[256], msg[512];
		sprintClassName( buf, action->cf, action->cf->this_class );
		sprintf( msg, "%s is not a subclass of java/security/PrivilegedAction.\n", buf );
		env->retcode = newCompleteException( env, InternalError, msg );
		return NULL;
	}

	/* ### Should check privileges here! ### */

	/* Execute the run() method on 'action' */
	if( forceExecuteJavaMethod( action, env, "run", "()Ljava/lang/Object;" ) != 0 )
		return NULL;	/* Exception will be on stack */

	/* Make execution stops after run() completes */
	env->fp->stopOnPop = 1;

	/* Execute run() */
	while( !retcode )
		retcode = executeInstruction( env );

	/* Stack will be empty if we get an exception or we don't stop-on-pop */
	if( (retcode != 1) || env->ex || (env->fp && env->fp->ex) )
		return NULL;

	/* run() returns an Object and pop and return it */
	POP_OBJREF( env->fp, ret );

	/*if( ret )
	{
		char buf[256];
		sprintClassName( buf, ret->cf, ret->cf->this_class );
		printf( "java.security.AccessController.doPrivileged(): returned a %s.\n", buf );
	}
	else
		printf( "java.security.AccessController.doPrivileged(): returned null.\n" );*/

	return ret;
}


/**
 * Executes the run() method of action as privileged code.  Checks only the
 * privileges of the context.
 *
 * We don't understand privileges yet, so we execute everything as privileged.
 * As we need the return value from run(), we use stopOnPop() to halt execution
 * when run() completes and have our own execution loop.
 *
 * public static native Object doPrivileged(
 *     java.security.PrivilegedAction action,
 *     java.security.AccessControlContext context );
 */
jobject Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedAction_2Ljava_security_AccessControlContext_2(
	JNIEnv *env, jclass staticself, jobject action, jobject context )
{
	/* Re-use our exising code, as at the moment we do nothing different! */
	return Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedAction_2( env, staticself, action );
}


/**
 * Internal function to create a PrivilegedActionException.  The jobject
 * parameter is the Exception object to be wrapped.
 */
int newPrivilegedActionException( JNIEnv* env, jobject wrapped )
{
	Frame* fp = env->fp;
	ClassFile* excf;
	const char* exclass = exceptionKindToString( PrivilegedActionException );

	if( !fp )
	{
		if( verbose )
		{
			char wbuf[256];
			sprintClassName( wbuf, wrapped->cf, wrapped->cf->this_class );
			printf( "Exception: %s: %s\n", (exclass ? exclass : "Unknown Exception"), wbuf );
			printf( "\tStack empty.\n" );;
		}

		env->ex = (Exception*) javavm_malloc( sizeof( Exception ) );
		env->ex->kind = PrivilegedActionException;
		env->ex->message = NULL;
		env->ex->ex = NULL;
		return 2;
	}

	if( verbose )
	{
		char wbuf[256];
		sprintClassName( wbuf, wrapped->cf, wrapped->cf->this_class );
		printf( "JavaException: %s: %s\n", (exclass ? exclass : "Unknown Exception"), wbuf );
		whereAreWe( fp );
	}

	if( exclass )
	{
		if( (excf = getClass( env, exclass )) )
		{
			u4 argv[1]; u1 argvt[1] = { T_OBJREF }; int argc = 1;
		
			fp->ex = (Exception*) javavm_malloc( sizeof( Exception ) );
			fp->ex->kind = JavaException;
			fp->ex->message = NULL;
			fp->ex->ex = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
			fp->ex->ex->cf = excf;
			fp->ex->ex->type = 0;
			fp->ex->ex->data = (u4*) javavm_malloc( 4 * excf->object_size );
			fp->ex->ex->length = 1;
			memset( fp->ex->ex->data, 0, 4 * excf->object_size );

			/* reset stack pointer to clear stack */
			if( fp->mi && fp->mi->ca )
			{
				fp->sp = fp->local + fp->mi->ca->max_locals;
				fp->tsp = fp->tlocal + fp->mi->ca->max_locals;
			}

			/* Call <init> with wrapped Exception *without* using the stack */
			argv[0] = (u4) wrapped;
			return forceExecuteJavaMethodWithArgs( fp->ex->ex, env, "<init>", "(Ljava/lang/Exception;)V", argc, argv, argvt );
		}
	}

	printf( "WARNING: Could not load java.security.PrivilegedActionException\n" );

	fp->ex = (Exception*) javavm_malloc( sizeof( Exception ) );
	fp->ex->kind = PrivilegedActionException;
	fp->ex->message = NULL;
	fp->ex->ex = NULL;

	return 2;
}


/**
 * Executes the run() method of action as privileged code.  Checks only the
 * privileges of the immediate caller.  This method differs in that if an
 * exception occurs, we trap it and throw a PrivilegedActionException instead.
 *
 * We don't understand privileges yet, so we execute everything as privileged.
 * As we need the return value from run(), we use stopOnPop() to halt execution
 * when run() completes and have our own execution loop.
 *
 * private static native Object doPrivileged( java.security.PrivilegedExceptionAction action )
 *     throws java.security.PrivilegedActionException;
 */
jobject Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedExceptionAction_2(
	JNIEnv* env, jclass staticself, jobject action )
{
	ObjRef* ret;
	int retcode = 0;

	ClassFile* privExAction = getClass( env, "java/security/PrivilegedExceptionAction" );

	if( !privExAction ) {
		env->retcode = newCompleteException( env, NoClassDefFoundError, "java/security/PrivilegedExceptionAction" );
		return NULL;
	}

	/* Is 'action' a subclass of java.security.PrivilegedExceptionAction */
	if( !subclassOf( action->cf, privExAction ) )
	{
		char buf[256], msg[512];
		sprintClassName( buf, action->cf, action->cf->this_class );
		sprintf( msg, "%s is not a subclass of java/security/PrivilegedExceptionAction.\n", buf );
		env->retcode = newCompleteException( env, InternalError, msg );
		return NULL;
	}

	/* ### Should check privileges here! ### */

	/* Execute the run() method on 'action' */
	if( forceExecuteJavaMethod( action, env, "run", "()Ljava/lang/Object;" ) != 0 )
		return NULL;	/* Exception on stack */

	/* Make execution stops after run() completes */
	env->fp->stopOnPop = 1;

	/* Execute run() */
	while( !retcode )
		retcode = executeInstruction( env );

	/* Stack will be empty if we get an exception or we don't stop-on-pop */
	if( (retcode != 1) && (retcode != 2) )
		return NULL;

	/* Was an exception raised? If so, convert it into a PrivilegedActionException */
	if( (retcode == 2) || env->ex || (env->fp && env->fp->ex) )
	{
		env->retcode = newPrivilegedActionException( env, (env->ex ? env->ex->ex : (env->fp ? env->fp->ex->ex : NULL)) );
		return NULL;
	}

	/* run() returns an Object and pop and return it */
	POP_OBJREF( env->fp, ret );

	/*if( ret )
	{
		char buf[256];
		sprintClassName( buf, ret->cf, ret->cf->this_class );
		printf( "java.security.AccessController.doPrivileged(): returned a %s.\n", buf );
	}
	else
		printf( "java.security.AccessController.doPrivileged(): returned null.\n" );*/

	return ret;
}


/**
 * Executes the run() method of action as privileged code.  Checks only the
 * privileges of the context.  This method differs in that if an
 * exception occurs, we trap it and throw a PrivilegedActionException instead.
 *
 * We don't understand privileges yet, so we execute everything as privileged.
 * As we need the return value from run(), we use stopOnPop() to halt execution
 * when run() completes and have our own execution loop.
 *
 * public static native Object doPrivileged(
 *     java.security.PrivilegedExceptionAction action,
 *     java.security.AccessControlContext context )
 *         throws java.security.PrivilegedActionException;
 */
jobject Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedExceptionAction_2Ljava_security_AccessControlContext_2(
	JNIEnv *env, jclass staticself, jobject action, jobject context )
{
	/* Re-use our exising code, as at the moment we do nothing different! */
	return Java_java_security_AccessController_doPrivileged__Ljava_security_PrivilegedExceptionAction_2( env, staticself, action );
}


/**
 * private static native java.security.AccessControlContext getStackAccessControlContext();
 */
jobject Java_java_security_AccessController_getStackAccessControlContext( JNIEnv* env, jclass staticself )
{
	printf( "WARNING: java.security.AccessController.getStackAccessControlContext() dummied.\n" );
	return NULL;
}
#endif /* JDK 1.4 */
