/******************************************************************************
 *
 * Filename: java_util_zip_ZipFile.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  10th May 2003
 *
 * Purpose:  Native method implementations for java.util.zip.ZipFile.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>    /* for file stat stuff */

#include "../engine/config.h"

#include "java_util_zip_ZipFile.h"

#include "../engine/util.h"
#include "../engine/memory.h"
#include "../engine/zipfile.h"
#include "../engine/hashtable.h"


/* JDK1.4 (There are no native methods in JDK1.1.7A)
-    private static native void initIDs();
+1.3 private static native long open(java.lang.String, int, long);
+1.2 private static native int getTotal(long);
+    private static native long getEntry(long, java.lang.String);
+    private static native void freeEntry(long, long);
+    private static native int getMethod(long);
+    private static native long getNextEntry(long, int);
+    private static native void close(long);
+    private static native int read(long, long, int, byte[], int, int);
+    private static native int getCSize(long);
+    private static native int getSize(long);
+    private static native java.lang.String getZipMessage(long);
*/


#if JDK_VERSION >= 140

OpenZipFile* openzipsv[MAX_ZIPFILES];
int openzipsc = 0;

OpenZipEntry* openentriesv[256];
int openentriesc = 0;


/**
 * Internal function to cache the contents of the Central Directory of an
 * already open zip file.  Returns non-zero if there is a problem reading the
 * zip file.
 */
int populateCache( OpenZipFile* zipfile )
{
	u4 zipsig;  /* Either ziplocsig or zipcensig */
	int32 n;
	ZipCentralDir record;

	/* Locate the Central Directory (first time only). */ 
	if( (zipfile->entries = findCentralDirectory( zipfile->f )) < 0 )
		return 1;

	/* Create a new hashtable */
	zipfile->cache = Hashtable_new( 101 );

	/* Allocate a block of memory for all the cache entries, this is also used
	   to for the sequential index */
	zipfile->index = (ZipCacheEntry*) javavm_malloc( zipfile->entries * sizeof(ZipCacheEntry) );

	/* Start reading the Central Directory and cache it. */
	for( n = 0; n < zipfile->entries; n++ )
	{
		/* Read double word to check if we have a Central Directory */
		readDWord( &zipsig, zipfile->f );

		if( zipsig != CENTRAL_DIR_MAGIC )
			return 1;

		/* Read the remainder of the structure */
		if( readCentralDirectory( &record, zipfile->f ) != 0 )
			return 1;

		/* Fill in the index entry */
		zipfile->index[n].zipofst = record.zipofst;

		/* Add a new entry to the hashtable */
		Hashtable_put( zipfile->cache, strdup(record.zipcfn), &zipfile->index[n] );
	}

	return 0;
}


/**
 * Internal function, called by main() to free any memory held in open ZipFiles
 * or ZipEntry's.  In particular this frees the cache and index.  Note that the
 * zipfile itself is kept closed, so we never end up with unclosed files.
 */
void closeAllJavaUtilZipZipFiles()
{
	int n;

	for( n = 0; n < openentriesc; n++ )
	{
		javavm_free( openentriesv[n], sizeof(OpenZipEntry) );
		openentriesv[n] = NULL;
	}

	openentriesc = 0;

	for( n = 0; n < openzipsc; n++ )
	{
		if( openzipsv[n] )
		{
			/*printf( "closeAllJavaUtilZipZipFiles() freeing zipfile %d\n", n );*/

			/* Free the directory cache keys, if we populated it */
			if( openzipsv[n]->cache )
				Hashtable_delete( openzipsv[n]->cache, 0, 1, 0, NULL );

			/* Free the index of cache entries (which are also the values from the hashtable) */
			if( openzipsv[n]->index )
				javavm_free( openzipsv[n]->index, openzipsv[n]->entries * sizeof(ZipCacheEntry) );

			/* Free the file record and null the pointer */
			javavm_free( openzipsv[n], sizeof(OpenZipFile) );
			openzipsv[n] = NULL;
		}
	}

	openzipsc = 0;
}


/**
 * ???Closes the ZipFile specified by 'handle'.???
 *
 * private static native void close( long handle );
 */
void Java_java_util_zip_ZipFile_close( JNIEnv* env, jclass staticself, jlong handle )
{
	/* Convert handle to index */
	int index = (int) handle - 1;

	/*printf( "java.util.zip.ZipFile.close( %d )\n", (int32) handle );*/

	/* Check that index is valid and that it refers to an open file */
	if( (index < 0) || (index >= openzipsc) || !(openzipsv[index]) )
	{
		env->retcode = newCompleteException( env, InternalError, "Invalid ZipFile handle" );
		return;
	}

	/* Free the directory cache keys, if we populated it */
	if( openzipsv[index]->cache )
		Hashtable_delete( openzipsv[index]->cache, 0, 1, 0, NULL );

	/* Free the index of cache entries (which are also the values from the hashtable) */
	if( openzipsv[index]->index )
		javavm_free( openzipsv[index]->index, openzipsv[index]->entries * sizeof(ZipCacheEntry) );

	/* Free the file record and null the pointer */
	javavm_free( openzipsv[index], sizeof(OpenZipFile) );
	openzipsv[index] = NULL;
	
	/* If last entry, decrement open entries count */
	if( index == openzipsc - 1 )
		openzipsc--;
}


/**
 * ???Frees any resources used to hold the zip entry specified by the handle
 * 'hentry' in the zip file associated with the handle 'hzip'???
 *
 * private static native void freeEntry( long hzip, long hentry );
 */
void Java_java_util_zip_ZipFile_freeEntry( JNIEnv* env, jclass staticself, jlong hzip, jlong hentry )
{
	int zipind, entind;

	/* Convert handles to indices */
	zipind = (int) hzip - 1;
	entind = (int) hentry - MAX_ZIPFILES;

	/*printf( "java.util.zip.ZipFile.freeEntry( %d, %d ) returning 0\n", (int32) hzip, (int32) hentry );*/

	/* Check that index is valid and that it refers to an open file */
	if( (zipind < 0) || (zipind >= openzipsc) || !(openzipsv[zipind]) )
	{
		env->retcode = newCompleteException( env, InternalError, "Invalid ZipFile handle" );
		return;
	}

	/* Check that index is valid and that it refers to an open zip entry */
	if( (entind < 0) || (entind >= openentriesc) || !(openentriesv[entind]) )
	{
		env->retcode = newCompleteException( env, InternalError, "Invalid ZipEntry handle" );
		return;
	}

	/* Free the entry record and null the pointer */
	javavm_free( openentriesv[entind], sizeof(OpenZipEntry) );
	openentriesv[entind] = NULL;

	/* If last entry, decrement open entries count */
	if( entind == openentriesc - 1 )
		openentriesc--;
}


/**
 * ???Returns the compressed size of the zip file entry specified by 'hentry'.???
 *
 * private static native int getCSize( long hentry );
 */
jint Java_java_util_zip_ZipFile_getCSize( JNIEnv* env, jclass staticself, jlong hentry )
{
	/* Convert handle to an index into openentriesv */
	int index = (int) hentry - MAX_ZIPFILES;

	/* Check that index is valid and that it refers to an open file */
	if( (index < 0) || (index >= openentriesc) || !openentriesv[index] )
	{
		/* Return a default value */
		/*printf( "java.util.zip.ZipFile.getCSize( %d ) returning 0\n", (int32) hentry );*/
		return 0;
	}

	/*printf( "java.util.zip.ZipFile.getCSize( %d ) returning %d\n", (int32) hentry, openentriesv[index]->local.zipsize );*/

	/* Return the cached zip entry size */
	return openentriesv[index]->local.zipsize;
}


/**
 * ???Returns a handle onto the ZipEntry with the filename 'name' in the
 * (previously opened) ZipFile specified by 'handle'.???
 *
 * private static native long getEntry( long handle, java.lang.String name );
 */
jlong Java_java_util_zip_ZipFile_getEntry( JNIEnv* env, jclass staticself, jlong handle, jobject name )
{
	char entryname[255];
	int index;
	jlong ret;

	/* Convert handle to an index into openzipsv */
	index = (int) handle - 1;

	/* Check that index is valid and that it refers to an open file */
	if( (index < 0) || (index >= openzipsc) || !(openzipsv[index]) ) {
		env->retcode = newCompleteException( env, InternalError, "Invalid ZipFile handle" );
		return 0;
	}

	/* Get hold of the filename */
	if( extractFromJavaLangString( entryname, 255, name ) )
	{
		FILE *f;
		u4 zipsig;  /* Either ziplocsig or zipcensig */
		ZipLocalFile local;

		/* Lookup the filename in the cache */
		ZipCacheEntry* cached = (ZipCacheEntry*) Hashtable_get( openzipsv[index]->cache, entryname );

		/* If not found, return 0 */
		if( !cached ) {
			/*printf( "java.util.zip.ZipFile.getEntry( %d, \"%s\" ) returning 0\n", (int32) handle, entryname );*/
			return 0;
		}

		if( !(f = fopen( openzipsv[index]->filename, "rb" )) ) {
			env->retcode = newCompleteException( env, InternalError, "Unable to re-open zip file" );
			return 0;
		}

		/* Move to start of Local File entry */
		if( fseek( f, cached->zipofst, SEEK_SET ) != 0 ) {
			fclose( f );
			env->retcode = newCompleteException( env, InternalError, "Corrupt zip file" );
			return 0;
		}

		/* Read double word to check if we have a Local File */
		readDWord( &zipsig, f );

		if( zipsig != LOCAL_FILE_MAGIC ) {
			fclose( f );
			env->retcode = newCompleteException( env, InternalError, "Missing ZipEntry" );
			return 0;
		}

		/* Read the remainder of the structure */
		if( readLocalFile( &local, f ) != 0 ) {
			fclose( f );
			env->retcode = newCompleteException( env, InternalError, "Corrupt ZipEntry" );
			return 0;
		}

		fclose( f );

		/* Create an new open zip file record */
		openentriesv[openentriesc] = (OpenZipEntry*) javavm_malloc( sizeof(OpenZipEntry) );
		memcpy( &openentriesv[openentriesc]->local, &local, sizeof(ZipLocalFile) );
		openentriesv[openentriesc]->zipofst = cached->zipofst;

		ret = MAX_ZIPFILES + openentriesc;
		openentriesc++;

		/*printf( "java.util.zip.ZipFile.getEntry( %d, \"%s\" ) returning %d\n", (int32) handle, entryname, ret );*/
		return ret;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to extract zip file name from java.lang.String" );
	return 0;
}


/**
 * ???Returns the compression method of the zip file entry specified by 'hentry'.???
 *
 * private static native int getMethod( long hentry );
 */
jint Java_java_util_zip_ZipFile_getMethod( JNIEnv* env, jclass staticself, jlong hentry )
{
	/* Convert handle to an index into openentriesv */
	int index = (int) hentry - MAX_ZIPFILES;

	/* Check that index is valid and that it refers to an open file */
	if( (index < 0) || (index >= openentriesc) || !openentriesv[index] )
	{
		/* Return a default value */
		/*printf( "java.util.zip.ZipFile.getMethod( %d ) returning 0\n", (int32) hentry );*/
		return 0;
	}

	/*printf( "java.util.zip.ZipFile.getMethod( %d ) returning %d\n", (int32) hentry, openentriesv[index]->local.zipmthd );*/

	/* Return the cached zip entry compression method */
	return openentriesv[index]->local.zipmthd;
}


/**
 * ???Returns a handle onto the given numbered ZipEntry (starting at 0) from
 * the (previously opened) ZipFile specified by 'handle'.???
 *
 * private static native long getNextEntry( long handle, int count );
 */
jlong Java_java_util_zip_ZipFile_getNextEntry( JNIEnv* env, jclass staticself, jlong handle, jint count )
{
	int index;
	FILE *f;
	u4 zipsig;  /* Either ziplocsig or zipcensig */
	ZipLocalFile local;
    ZipCacheEntry* cached;
	jlong ret;

	/* Convert handle to an index into openzipsv */
	index = (int) handle - 1;

	/* Check that index is valid and that it refers to an open file */
	if( (index < 0) || (index >= openzipsc) || !(openzipsv[index]) ) {
		env->retcode = newCompleteException( env, InternalError, "Invalid ZipFile handle" );
		return 0;
	}

	/* Check that index is valid and that it refers to an open file */
	if( (count < 0) || (count >= openzipsv[index]->entries) || !(openzipsv[index]->index) ) {
		env->retcode = newCompleteException( env, InternalError, "Invalid ZipEntry index" );
		return 0;
	}

	/* Lookup the cache entry in the index */
	cached = &openzipsv[index]->index[count];

	if( !(f = fopen( openzipsv[index]->filename, "rb" )) ) {
		env->retcode = newCompleteException( env, InternalError, "Unable to re-open zip file" );
		return 0;
	}

	/* Move to start of Local File entry */
	if( fseek( f, cached->zipofst, SEEK_SET ) != 0 ) {
		fclose( f );
		env->retcode = newCompleteException( env, InternalError, "Corrupt zip file" );
		return 0;
	}

	/* Read double word to check if we have a Local File */
	readDWord( &zipsig, f );

	if( zipsig != LOCAL_FILE_MAGIC ) {
		fclose( f );
		env->retcode = newCompleteException( env, InternalError, "Missing ZipEntry" );
		return 0;
	}

	/* Read the remainder of the structure */
	if( readLocalFile( &local, f ) != 0 ) {
		fclose( f );
		env->retcode = newCompleteException( env, InternalError, "Corrupt ZipEntry" );
		return 0;
	}

	fclose( f );

	/* Create an new open zip file record */
	openentriesv[openentriesc] = (OpenZipEntry*) javavm_malloc( sizeof(OpenZipEntry) );
	memcpy( &openentriesv[openentriesc]->local, &local, sizeof(ZipLocalFile) );
	openentriesv[openentriesc]->zipofst = cached->zipofst;

	ret = MAX_ZIPFILES + openentriesc;
	openentriesc++;

	/*printf( "java.util.zip.ZipFile.getNextEntry( %d, %d ) returning %d\n", (int32) handle, count, ret );*/
	return ret;
}


/**
 * ???Returns the uncompressed size of the zip file entry specified by 'hentry'.???
 *
 * private static native int getSize( long hentry );
 */
jint Java_java_util_zip_ZipFile_getSize( JNIEnv* env, jclass staticself, jlong hentry )
{
	/* Convert handle to an index into openentriesv */
	int index = (int) hentry - MAX_ZIPFILES;

	/* Check that index is valid and that it refers to an open file */
	if( (index < 0) || (index >= openentriesc) || !openentriesv[index] )
	{
		/* Return a default value */
		/*printf( "java.util.zip.ZipFile.getSize( %d ) returning 0\n", (int32) hentry );*/
		return 0;
	}

	/*printf( "java.util.zip.ZipFile.getSize( %d ) returning %d\n", (int32) hentry, openentriesv[index]->local.zipuncmp );*/

	/* Return the cached zip entry uncompressed size */
	return openentriesv[index]->local.zipuncmp;
}


/**
 * ???Returns the total number of entries in the (previously opened) zip file
 * specified by 'handle'.???
 *
 * private static native int getTotal( long handle );
 */
jint Java_java_util_zip_ZipFile_getTotal( JNIEnv* env, jclass staticself, jlong handle )
{
	/* Convert handle to an index into openzipsv */
	int index = (int) handle - 1;

	/* Check that index is valid and that it refers to an open file */
	if( (index < 0) || (index >= openzipsc) || !openzipsv[index] )
	{
		/* Return a default value */
		/*printf( "java.util.zip.ZipFile.getTotal( %d ) returning 0\n", (int32) handle );*/
		return 0;
	}

	/*printf( "java.util.zip.ZipFile.getTotal( %d ) returning %d\n", (int32) handle, openzipsv[index]->size );*/

	/* Return the number of entries */
	return openzipsv[index]->entries;
}


/**
 * ???Gets a message indicating what the last error was.???
 *
 * private static native java.lang.String getZipMessage( long handle );
 */
jobject Java_java_util_zip_ZipFile_getZipMessage( JNIEnv* env, jclass staticself, jlong handle )
{
	printf( "java.util.zip.ZipFile.getZipMessage( %d ) dummied.\n", (int32) handle );
	return NULL;
}


/**
 * private static native void initIDs();
 */
void Java_java_util_zip_ZipFile_initIDs( JNIEnv* stp, jclass staticself )
{
	int n = 0;

	for( n = 0; n < MAX_ZIPFILES; n++ )
		openzipsv[n] = NULL;

	for( n = 0; n < 256; n++ )
		openentriesv[n] = NULL;
}


#define OPEN_READ   1
#define OPEN_DELETE 4

/**
 * ???Opens the zip file with the filename 'file'.  'mode' dictates how the zip
 * file should be opened and is a bitmask of the above #defined values.
 * 'mtime' is the modification timestamp of the zip file.  If the mode bit
 * OPEN_DELETE is set, then the zip file is marked for automatic deletion when
 * it is closed.  Although not specified, it is most likely that this method
 * raises an IOException if the zip file cannot be found.???
 *
 * private static native long open( java.lang.String file, int mode, long mtime );
 */
jlong Java_java_util_zip_ZipFile_open( JNIEnv* env, jclass staticself, jobject file, jint mode, jlong mtime )
{
	char filename[255];

	/* Check that mode is valid and something we understand */
	if( mode != OPEN_READ )
	{
		env->retcode = newCompleteException( env, InternalError, "java.util.zip.ZipFile.open() with mode other than read not yet implemented" );
		return 0;
	}

	/* Get hold of the filename */
	if( extractFromJavaLangString( filename, 255, file ) )
	{
		FILE *f;

		/* Try to open the zip file */
		if( !(f = fopen( filename, "rb" )) )
		{
			char msg[512];
			sprintf( "Unable to open \"%s\"", filename );
			env->retcode = newCompleteException( env, IOException, msg );
			return 0;
		}

		/* Create and new open zip file record */
		openzipsv[openzipsc] = (OpenZipFile*) javavm_malloc( sizeof(OpenZipFile) );
		openzipsv[openzipsc]->filename = strdup(filename);
		openzipsv[openzipsc]->f        = f;
		openzipsv[openzipsc]->entries  = 0;
		openzipsv[openzipsc]->cache    = NULL;
		openzipsv[openzipsc]->index    = NULL;

		/*printf( "java.util.zip.ZipFile.open( \"%s\", %d, %d ) returning %d\n", filename, mode, (int32) mtime, openzipsc+1 );*/

		/* Cache the contents of the Central Directory */
		if( populateCache( openzipsv[openzipsc] ) != 0 )
		{
			char msg[512];
			sprintf( "Bad zip file \"%s\"", filename );
			env->retcode = newCompleteException( env, IOException, msg );
			fclose( f );
			return 0;
		}

		fclose( f );

		return ++openzipsc;	/* returns openzipsc + 1 */
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to extract zip file name from java.lang.String" );
	return 0;
}


/**
 * ???Reads 'len' bytes from the compressed data for the entry with the handle
 * 'hentry' into the array 'b' at offset 'off'.  Subsequent reads, read from
 * where this read left off.
 *
 * private static native int read( long hzip, long hentry, int pos, byte[] b, int off, int len );
 */
jint Java_java_util_zip_ZipFile_read( JNIEnv* env, jclass staticself,
	jlong hzip, jlong hentry, jint pos, jobject array, jint off, jint len )
{
	int zipind, entind;
	FILE* f;
	jint n;
	char* dst;

	/* Convert handles to indices */
	zipind = (int) hzip - 1;
	entind = (int) hentry - MAX_ZIPFILES;

	/* Check that index is valid and that it refers to an open file */
	if( (zipind < 0) || (zipind >= openzipsc) || !(openzipsv[zipind]) )	{
		env->retcode = newCompleteException( env, InternalError, "Invalid ZipFile handle" );
		return 0;
	}

	/* Check that index is valid and that it refers to an open zip entry */
	if( (entind < 0) || (entind >= openentriesc) || !(openentriesv[entind]) ) {
		env->retcode = newCompleteException( env, InternalError, "Invalid ZipEntry handle" );
		return 0;
	}

	/* Validate the array and check off + len is inside array bounds */
	if( !array || (array->cf != NULL) || (array->type != T_BYTE) || (off + len > (jint) array->length) ) {
		env->retcode = newCompleteException( env, InternalError, "Invalid byte[] for read" );
		return 0;
	}

	dst = (char*) array->data;

	if( !(f = fopen( openzipsv[zipind]->filename, "rb" )) )	{
		env->retcode = newCompleteException( env, InternalError, "Unable to re-open zip file" );
		return 0;
	}

	/* Move to start of comressed data, taking into acount how far we have previously read */
	if( fseek( f, openentriesv[entind]->zipofst + 30
		+ openentriesv[entind]->local.zipfnln
		+ openentriesv[entind]->local.zipxtraln
		/* + openentriesv[entind]->local.zipcmntln */
		+ pos, SEEK_SET ) != 0 )
	{
		fclose( f );
		env->retcode = newCompleteException( env, InternalError, "Unable to read data from zip file" );
		return 0;
	}

	n = fread( &dst[off], 1, len, f );


	if( (n == 0) && feof( f ) )
	{
		fclose( f );
		/*printf( "java.util.zip.ZipFile.read( %d, %d, %d, %p[%d], %d, %d ) returning EOF\n",
			(int32) hzip, (int32) hentry, pos, array, array->length, off, len );*/
		return EOF;
	}

	fclose( f );

	/*printf( "java.util.zip.ZipFile.read( %d, %d, %d, %p[%d], %d, %d ) returning %d\n",
		(int32) hzip, (int32) hentry, pos, array, array->length, off, len, n );*/

	return n;
}
#endif /* JDK 1.4 */
